/** \file    configfile.h
    \brief   Reading/writing of INI files
    \author  EV
    \date    2014

    This file defines the class for reading and writing parameters from/to INI text files.
    The file format is identical to MS .ini file: 

    [Section]
    key=value
    ;key=old value -- commented out, lines starting with ; # / are ignored but preserved in the file upon saving

    In addition, several convenience function are provided to read and write entire groups 
    of SMILE parameters related to potential and orbit integration.
*/

#pragma once
#include <string>
#include <vector>

namespace smile{

/** Utility class for handling an INI file.
    Provides high-level interface for reading and writing values of various types, 
    while isolating the details of the implementation in an internal pointer. 
    Currently, there are two such implementations -- based on Qt QSettings class 
    (used in SMILE up to version 2.4), and a simple hand-made class which deals with text files directly. 
    Data values are addressed by a case-insensitive key string which combines section and key name: Section/Key.
*/
class CConfigFile {
public:
    /// open an INI file with the given name; if it exists, read its contents on initialization
    CConfigFile(const std::string& fileName);
    /// alternative way of parsing a list of parameters given in the form "name=value" 
    /// and attribute them to the given section of a non-existent ini file 
    /// (useful for mimicking the ini file with command-line arguments)
    CConfigFile(const std::string& iniSection, const int numParams, const char* const* params);
    /// destroy the class and write unsaved data back to INI file, if it was modified
    ~CConfigFile();

    /// return a string value from the INI file
    std::string getValueString(const std::string& key, const std::string& defaultValue="") const;
    /// overloaded for convenience, return an integer value
    int getValueInt(const std::string& key, int defaultValue=0) const;
    /// overloaded for convenience, return a floating-point value
    double getValueDouble(const std::string& key, double defaultValue=0) const;
    /// overloaded for convenience, return a boolean value
    bool getValueBool(const std::string& key, bool defaultValue=false) const;
    /// set an integer value
    void setValue(const std::string& key, int value);
    /// set an unsigned integer value
    void setValue(const std::string& key, unsigned int value);
    /// set a floating-point value
    void setValue(const std::string& key, double value);
    /// set a floating-point value with a given precision (rather, width of the string representation)
    void setValue(const std::string& key, double value, unsigned int width);
    /// set a string value
    void setValue(const std::string& key, const std::string& value);
    /// set a string value (overloaded)
    void setValue(const std::string& key, const char* value);
    /// set a boolean value
    void setValue(const std::string& key, bool value);
    /// return a list of all sections in the INI file
    void listGroups(std::vector<std::string> *list) const;
private:
    void* pImpl;   ///< implementation-specific data
};

struct CConfigPotential;  // forward declaration
struct CConfigOrbit;      // forward declaration

/** Read the configuration data from the INI file which is contained in one [Potential*] section.
    \param[in] configReader is the instance of class that handles the INI file;
    \param[in,out] ptrConfigPotential is the pointer to the configuration block 
    that will be filled with the values read from the INI file; 
    if no corresponding value exists in the file, then an old one is retained.
    \param[in] potGroup is the index of potential group (the first one is contained in the section [Potential], 
    the second one in [Potential1], etc.). **/
void readConfigPotential (const CConfigFile& configReader, CConfigPotential* ptrConfigPotential, unsigned int potGroup=0);

/** Write the configuration data for one potential component into the INI file.
    \param[in,out] configWriter is the instance of class that handles the INI file (it will be updated);
    \param[in] ptrConfigPotential is the pointer to the configuration block for the potential component;
    \param[in] potGroup is the index of potential section (0th is written to the section [Potential], 
    1st to section [Potential1], etc.) **/
void writeConfigPotential(CConfigFile& configWriter, const CConfigPotential* ptrConfigPotential, unsigned int potGroup=0);

/** Read configuration data for orbit integration (a subset of parameters contained in the [Orbit] section of INI file).
    \param[in] configReader is the instance of class that handles the INI file;
    \param[in,out] ptrConfigOrbit is the pointer to the configuration block 
    that will be filled with the values read from the INI file; 
    if no corresponding value exists in the file, then an old one is retained. **/
void readConfigOrbit (const CConfigFile& configReader, CConfigOrbit* ptrConfigOrbit);

/** Write configuration data for orbit integration to INI file.
    \param[in,out] configWriter is the instance of class that handles the INI file (it will be updated);
    \param[in] ptrConfigOrbit is the pointer to the configuration block for the orbit integration. **/
void writeConfigOrbit(CConfigFile& configWriter, const CConfigOrbit* ptrConfigOrbit);

}; // namespace
