/** \file    iosnapshot.h
    \brief   Input/output of Nbody snapshots in various formats
    \author  EV
    \date    2010-2013

    The base class, smile::CBasicIOSnapshot, is used as the common interface 
    for reading and writing Nbody snapshots to disk. 
    The snapshots are provided by smile::CPointMassSet.
    Derived classes implement the data storage in various formats.
    Helper routines create an instance of the class corresponding to a given 
    format string or to the actual file format.
*/

#pragma once
#include "common.h"

namespace smile{

/// The abstract class implementing reading and writing snapshots; 
/// derived classes take the filename as the argument of the constructor.
class CBasicIOSnapshot {
public:
    virtual ~CBasicIOSnapshot() {};
    /** read a snapshot from the file; 
        \param[in,out] points is a pointer to the CPointMassSet class;
                       if it is NULL then a new instance of this class is created,
                       otherwise the content of this object is replaced by the new snapshot.
        \return success or failure; in the latter case points may contain garbage.
    */
    virtual bool readSnapshot(CPointMassSet<float> * &points)=0;
    virtual bool readSnapshot(CPointMassSet<double>* &points)=0;
    /** write a snapshot to the file; return success or failure. */
    virtual bool writeSnapshot(const CPointMassSet<float> *points)=0;
    virtual bool writeSnapshot(const CPointMassSet<double>*points)=0;
};

/// Text file with three coordinates, possibly three velocities and mass, space or tab-separated.
class CIOSnapshotText: public CBasicIOSnapshot {
public:
    CIOSnapshotText(const std::string &_fileName): fileName(_fileName) {};
    virtual bool readSnapshot(CPointMassSet<float> * &points) { return readSnapshotT(points); }
    virtual bool readSnapshot(CPointMassSet<double>* &points) { return readSnapshotT(points); }
    virtual bool writeSnapshot(const CPointMassSet<float> *points) { return writeSnapshotT(points); }
    virtual bool writeSnapshot(const CPointMassSet<double>*points) { return writeSnapshotT(points); }
private:
    const std::string fileName;
    template<typename NumT> bool readSnapshotT(CPointMassSet<NumT> *&points);
    template<typename NumT> bool writeSnapshotT(const CPointMassSet<NumT> *points);
};

/// NEMO snapshot format.
/// reading is supported only if compiled with UNSIO library; 
/// writing is implemented by builtin routines.
class CIOSnapshotNemo: public CBasicIOSnapshot {
public:
    /// create the class to read or write to the file; 
    /// if writing is intended, may provide a header string and timestamp
    /// and choose whether to append to file if it already exists
    CIOSnapshotNemo(const std::string &_fileName, const std::string &_header="", double _time=0, bool _append=false) :
      fileName(_fileName), header(_header), time(_time), append(_append) {};
    virtual bool readSnapshot(CPointMassSet<float> * &points) { return readSnapshotT(points); }
    virtual bool readSnapshot(CPointMassSet<double>* &points) { return readSnapshotT(points); }
    virtual bool writeSnapshot(const CPointMassSet<float> *points) { return writeSnapshotT(points); }
    virtual bool writeSnapshot(const CPointMassSet<double>*points) { return writeSnapshotT(points); }
private:
    const std::string fileName;
    const std::string header;    ///< header string which will be written to the file
    const double time;           ///< timestamp of the snapshot to write
    const bool append;           ///< whether to append to the end of file or overwrite it
    template<typename NumT> bool readSnapshotT(CPointMassSet<NumT> *&points);
    template<typename NumT> bool writeSnapshotT(const CPointMassSet<NumT> *points);
};

#ifdef HAVE_UNSIO
/// GADGET snapshot format; needs UNSIO library.
class CIOSnapshotGadget: public CBasicIOSnapshot {
public:
    CIOSnapshotGadget(const std::string &_fileName): fileName(_fileName) {};
    virtual bool readSnapshot(CPointMassSet<float> * &points) { return readSnapshotT(points); }
    virtual bool readSnapshot(CPointMassSet<double>* &points) { return readSnapshotT(points); }
    virtual bool writeSnapshot(const CPointMassSet<float> *points) { return writeSnapshotT(points); }
    virtual bool writeSnapshot(const CPointMassSet<double>*points) { return writeSnapshotT(points); }
private:
    const std::string fileName;
    template<typename NumT> bool readSnapshotT(CPointMassSet<NumT> *&points);
    template<typename NumT> bool writeSnapshotT(const CPointMassSet<NumT> *points);
};
#endif

/* ------ Correspondence between file format names and types ------- */
/// list of all available IO snapshot formats, initialized at module start 
/// according to the file format supported at compile time
extern std::vector< std::string > formatsIOSnapshot;

/// creates an instance of appropriate snapshot reader, according to the file format 
/// determined by reading first few bytes, or NULL if a file doesn't exist
CBasicIOSnapshot* createIOSnapshotRead (const std::string &fileName);

/// creates an instance of snapshot writer for a given format name, 
/// or NULL if the format name string is incorrect
CBasicIOSnapshot* createIOSnapshotWrite(const std::string &fileFormat, const std::string &fileName, 
    const std::string& header="", const double time=0, const bool append=false);

};  // namespace
