#include "managepotential.h"
#include <cassert>
#include <fstream>
#include "iosnapshot.h"
#include "massmodel.h"
#include "potential.h"
#include "utils.h"

namespace smile {

const CDensity* createDensity(const CConfigPotential* config)
{
    switch(config->DensityType) 
    {
    case CDensity::PT_DEHNEN: 
        return new CPotentialDehnen(config->mass, config->scalerad, config->q, config->p, config->Gamma); 
    case CDensity::PT_PLUMMER:
        return new CDensityPlummer(config->mass, config->scalerad, config->q, config->p);
    case CDensity::PT_ISOCHRONE:
        return new CDensityIsochrone(config->mass, config->scalerad, config->q, config->p);
    case CDensity::PT_PERFECTELLIPSOID:
        return new CDensityPerfectEllipsoid(config->mass, config->scalerad, config->q, config->p); 
    case CDensity::PT_NFW: {
        double concentration = std::max<double>(config->scalerad2/config->scalerad, 1.0);
        return new CDensityNFW(config->mass, config->scalerad, config->q, config->p, concentration);
    }
    case CDensity::PT_SERSIC:
        return new CDensitySersic(config->mass, config->scalerad, config->q, config->p, config->sersicIndex);
    case CDensity::PT_FERRERS:
        return new CPotentialFerrers(config->mass, config->scalerad, config->q, config->p); 
    case CDensity::PT_MIYAMOTONAGAI:
        return new CPotentialMiyamotoNagai(config->mass, config->scalerad, config->scalerad2);
    case CDensity::PT_EXPDISK:
        return new CDensityExpDisk(config->mass, config->scalerad, config->scalerad2, CDensityExpDisk::ED_SECH2);
    default: return NULL;
    }
}

const CPotential* createPotential(CConfigPotential* config)
{
    const CPotential* potential=NULL;
    switch(config->PotentialType)
    {
    case CDensity::PT_LOG:
        potential=new CPotentialLog(config->mass, config->q, config->p, config->scalerad);
        break;
    case CDensity::PT_HARMONIC:
        potential=new CPotentialHarmonic(config->mass, config->q, config->p);
        break;
    case CDensity::PT_DEHNEN:
        potential=new CPotentialDehnen(config->mass, config->scalerad, config->q, config->p, config->Gamma);
        break;
    case CDensity::PT_MIYAMOTONAGAI:
        potential=new CPotentialMiyamotoNagai(config->mass, config->scalerad, config->scalerad2);
        break;
    case CDensity::PT_FERRERS:
        potential = new CPotentialFerrers(config->mass, config->scalerad, config->q, config->p); 
        break;
    case CDensity::PT_SCALEFREE:
        potential=new CPotentialScaleFree(config->mass, config->scalerad, config->q, config->p, config->Gamma);
        break;
    case CDensity::PT_SCALEFREESH:
        if(config->DensityType==CDensity::PT_COEFS)
            potential=readPotential(config);   // load coefs from a text file
        else
            potential=new CPotentialScaleFreeSH(config->mass, config->scalerad, config->q, config->p, config->Gamma, config->Ncoefs_angular);
        break;
    case CDensity::PT_NB:
        potential=readPotential(config);
        break;
    case CDensity::PT_SPLINE:
    case CDensity::PT_BSE:
    case CDensity::PT_BSECOMPACT:
    case CDensity::PT_CYLSPLINE:
    case CDensity::PT_SPHERICAL:
        // two possible variants: either init from analytic density model or from Nbody/coef/massmodel file
        if( config->DensityType==CDensity::PT_NB || 
            config->DensityType==CDensity::PT_ELLIPSOIDAL || 
            config->DensityType==CDensity::PT_MGE || 
            config->DensityType==CDensity::PT_COEFS )
        {
            potential=readPotential(config);
        } else {
            // create a temporary instance of density model, use it for computing expansion coefs
            const CDensity* densModel = createDensity(config);
            if(densModel!=NULL) 
            {
                if(config->PotentialType==CPotential::PT_BSE)
                    potential = new CPotentialBSE(
                        config->Alpha, config->Ncoefs_radial, config->Ncoefs_angular, densModel);
                else if(config->PotentialType==CPotential::PT_BSECOMPACT)
                    potential = new CPotentialBSECompact(
                        config->Rmax, config->Ncoefs_radial, config->Ncoefs_angular, densModel);
                else if(config->PotentialType==CPotential::PT_SPLINE) {
                    if(config->splineRMin>0 && config->splineRMax>0)
                    {
                        vectord grid;
                        createNonuniformGrid(config->Ncoefs_radial+1, 
                            config->splineRMin, config->splineRMax, true, &grid);
                        potential = new CPotentialSpline(
                            config->Ncoefs_radial, config->Ncoefs_angular, densModel, &grid);
                    } else  // no user-supplied grid extent -- will be assigned by default
                        potential = new CPotentialSpline(
                            config->Ncoefs_radial, config->Ncoefs_angular, densModel);
                }
#ifdef HAVE_INTERP2D
                else if(config->PotentialType==CPotential::PT_CYLSPLINE) {
                    if(config->DensityType==CDensity::PT_DEHNEN || config->DensityType==CDensity::PT_FERRERS ||
                        config->DensityType==CDensity::PT_MIYAMOTONAGAI) { // use potential for initialization, without intermediate CPotentialDirect step
                        potential = new CPotentialCylSpline(
                            config->Ncoefs_radial, config->Ncoefs_vertical, config->Ncoefs_angular, 
                            static_cast<const CPotential*>(densModel), // explicitly type cast to CPotential to call an appropriate constructor
                            config->splineRMin, config->splineRMax, config->splineZMin, config->splineZMax);
                    } else {
                        potential = new CPotentialCylSpline(
                            config->Ncoefs_radial, config->Ncoefs_vertical, config->Ncoefs_angular, densModel, 
                            config->splineRMin, config->splineRMax, config->splineZMin, config->splineZMax);
                    }
                }
#endif
                else if(config->PotentialType==CPotential::PT_SPHERICAL) {
                    const CMassModel* massModel=createMassModel(densModel, config->Ncoefs_radial, config->splineRMin, config->splineRMax);
                    if(massModel) { 
                        potential = new CPotentialSpherical(*massModel);
                        delete massModel;
                    }
                }
                else 
                    my_error(FUNCNAME, "Unknown type of potential expansion");
                delete densModel;
            }
        }
        break;
    default: ;
    }
    if(!potential)
        return NULL;  // signal of error
    config->PotentialType = potential->PotentialType();
    config->SymmetryType  = potential->symmetry();
#ifdef DEBUGPRINT
    my_message(FUNCNAME, 
        "Gamma="+convertToString(potential->getGamma())+", total mass="+convertToString(potential->totalMass()));
#endif
    return potential;
}

/* ------- auxiliary function to create potential of a given type from a set of point masses ------ */
template<typename NumT> 
const CPotential* createPotentialFromPoints(const CConfigPotential *configPotential, const CPointMassSet<NumT> *points)
{
    const CPotential* pot=NULL;
    if(configPotential->PotentialType==CDensity::PT_NB) 
    {
        pot = new CPotentialNB(configPotential->treecodeEps, configPotential->treecodeTheta, *points);
    }
    else if(configPotential->PotentialType==CDensity::PT_SPLINE)
    {
        if(configPotential->splineRMin>0 && configPotential->splineRMax>0) 
        {
            vectord radii;
            createNonuniformGrid(configPotential->Ncoefs_radial+1, configPotential->splineRMin, 
                configPotential->splineRMax, true, &radii);
            pot = new CPotentialSpline(configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, 
                *points, configPotential->SymmetryType, configPotential->splineSmoothFactor, &radii);
        } else
            pot = new CPotentialSpline(configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, 
                *points, configPotential->SymmetryType, configPotential->splineSmoothFactor);
    }
#ifdef HAVE_INTERP2D
    else if(configPotential->PotentialType==CDensity::PT_CYLSPLINE)
        pot = new CPotentialCylSpline(configPotential->Ncoefs_radial, configPotential->Ncoefs_vertical,
            configPotential->Ncoefs_angular, *points, configPotential->SymmetryType, 
            configPotential->splineRMin, configPotential->splineRMax,
            configPotential->splineZMin, configPotential->splineZMax);
#endif
    else if(configPotential->PotentialType==CDensity::PT_BSE)
        pot = new CPotentialBSE(configPotential->Alpha, configPotential->Ncoefs_radial, 
            configPotential->Ncoefs_angular, *points, configPotential->SymmetryType);
    else if(configPotential->PotentialType==CDensity::PT_BSECOMPACT)
        pot = new CPotentialBSECompact(configPotential->Rmax, configPotential->Ncoefs_radial, 
            configPotential->Ncoefs_angular, *points, configPotential->SymmetryType);
    else if(configPotential->PotentialType==CDensity::PT_SPHERICAL) {
        if(configPotential->splineRMin>0 && configPotential->splineRMax>0) 
        {
            vectord radii;
            createNonuniformGrid(configPotential->Ncoefs_radial+1, configPotential->splineRMin, 
                configPotential->splineRMax, true, &radii);
            pot = new CPotentialSpherical(*points, configPotential->splineSmoothFactor, &radii);
        } else
            pot = new CPotentialSpherical(*points, configPotential->splineSmoothFactor);
    }
    return pot;
}
// instantiations
template const CPotential* createPotentialFromPoints(const CConfigPotential *configPotential, const CPointMassSet<float> *points);
template const CPotential* createPotentialFromPoints(const CConfigPotential *configPotential, const CPointMassSet<double> *points);

/* ----- reading potential from a text or snapshot file ---- */

const CPotential* readPotential(CConfigPotential *configPotential)
{
    if( configPotential==NULL || 
      !(configPotential->PotentialType == CPotential::PT_NB || 
        configPotential->PotentialType == CPotential::PT_SPHERICAL ||
        configPotential->PotentialType == CPotential::PT_BSE ||
        configPotential->PotentialType == CPotential::PT_BSECOMPACT ||
        configPotential->PotentialType == CPotential::PT_SPLINE ||
        configPotential->PotentialType == CPotential::PT_CYLSPLINE ||
        configPotential->PotentialType == CPotential::PT_SCALEFREESH ||
        configPotential->PotentialType == CPotential::PT_UNKNOWN ) ) 
        return NULL;   // only may load these types of potential; unknown means try reading coefs for SHE/BSE potential
    const std::string& fileName=configPotential->NbodyFile;
    if(fileName.empty()) {
        my_message(FUNCNAME, "Empty file name");
        return NULL;
    }
    std::ifstream strmText(fileName.c_str(), std::ios::in);
    if(!strmText) {
        my_error(FUNCNAME, "Cannot read from file "+fileName);
        return NULL;   // file is not readable
    }
    std::string buffer;
    std::vector<std::string> fields;
    // check header (in case we are loading Coefs file but densityType is set to something else)
    if(std::getline(strmText, buffer) && buffer.size()<256) {  // to avoid parsing a binary file as a text
        splitString(buffer, "# \t", &fields);
        if( fields[0] == "BSEcoefs" || fields[0] == "BSECcoefs" || fields[0] == "SHEcoefs" || 
            fields[0] == "CylSpline" || fields[0] == "SFcoefs")
            configPotential->DensityType = CPotential::PT_COEFS;
    }
    strmText.seekg(0, std::ios_base::beg);  // back to start
    if(configPotential->PotentialType == CPotential::PT_SPHERICAL &&
        configPotential->DensityType == CPotential::PT_COEFS)
    {   // attempt to read r,M(r) pairs from file
        vectord radii, inmass;
        while(std::getline(strmText, buffer) && !strmText.eof())
        {
            splitString(buffer, "# \t", &fields);
            if(fields.size()>=2 && ((fields[0][0]>='0' && fields[0][0]<='9') || 
                fields[0][0]=='-' || fields[0][0]=='+' || fields[0][0]=='.'))
            {
                double rad=convertTo<double>(fields[0]);
                double mass=convertTo<double>(fields[1]);
                radii.push_back(rad);
                inmass.push_back(mass);
            }
        }
        const CPotentialSpherical* pot=new CPotentialSpherical(radii, inmass);
        CMassModel::MODELSTATUS status=pot->getModel()->getStatus();
        if(status==CMassModel::MS_OK || status==CMassModel::MS_DF_NEGATIVE)  // ok
            return pot;
        else {
            std::string errorstr;
            switch(status)
            {
            case CMassModel::MS_ARRAY_NOT_EQ:  errorstr="Invalid format of input table file (columns of unequal length)"; break;
            case CMassModel::MS_ARRAY_SHORT:   errorstr="Too few points"; break;
            case CMassModel::MS_RADII_NOT_ASC: errorstr="Radii not in ascending order"; break;
            case CMassModel::MS_MASS_NOT_ASC:  errorstr="Mass not increasing with radius"; break;
            default: ;  // do nothing
            }
            my_error(FUNCNAME, "Error loading spherical mass profile from file "+fileName+":\n"+errorstr);
            delete pot;
            return NULL;
        }
    } else if(configPotential->PotentialType != CPotential::PT_NB &&
        configPotential->DensityType == CPotential::PT_COEFS)
    {   // attempt to load coefficients stored in a text file
        bool valid=true;
        if(!std::getline(strmText, buffer)) valid=false;
        splitString(buffer, "# \t", &fields);
        CDensity::POTENTIALTYPE ptype = 
            fields[0] == "BSEcoefs" ? CDensity::PT_BSE :
            fields[0] == "BSECcoefs"? CDensity::PT_BSECOMPACT :
            fields[0] == "SFcoefs"  ? CDensity::PT_SCALEFREESH :
            fields[0] == "SHEcoefs" ? CDensity::PT_SPLINE : 
            fields[0] == "CylSpline"? CDensity::PT_CYLSPLINE : CDensity::PT_UNKNOWN;
        if(valid && ptype==CDensity::PT_CYLSPLINE) 
        {   // read coefs for cylindrical spline potential
#ifdef HAVE_INTERP2D
            if(!std::getline(strmText, buffer)) valid=false;
            splitString(buffer, "# \t", &fields);
            size_t size_R=convertTo<int>(fields[0]);
            if(!std::getline(strmText, buffer)) valid=false;
            splitString(buffer, "# \t", &fields);
            size_t ncoefsAngular=convertTo<int>(fields[0]);
            if(!std::getline(strmText, buffer)) valid=false;
            splitString(buffer, "# \t", &fields);
            size_t size_z=convertTo<int>(fields[0]);
            std::getline(strmText, buffer);  // time, ignored
            if(size_R==0 || size_z==0) valid=false;
            vectord gridR, gridz;
            std::vector<vectord> coefs(2*ncoefsAngular+1);
            while(valid && std::getline(strmText, buffer) && !strmText.eof()) {
                splitString(buffer, "# \t", &fields);
                int m=convertTo<int>(fields[0]);  // m (azimuthal harmonic index)
                if(m<-static_cast<int>(ncoefsAngular) || m>static_cast<int>(ncoefsAngular)) valid=false;
                std::getline(strmText, buffer);  // radii
                if(gridR.size()==0) {  // read values of R only once
                    splitString(buffer, "# \t", &fields);
                    for(size_t i=1; i<fields.size(); i++)
                        gridR.push_back(convertTo<double>(fields[i]));
                    if(gridR.size()!=size_R) valid=false;
                }
                gridz.clear();
                coefs[m+ncoefsAngular].assign(size_R*size_z,0);
                for(size_t iz=0; valid && iz<size_z; iz++)
                {
                    std::getline(strmText, buffer);
                    splitString(buffer, "# \t", &fields);
                    gridz.push_back(convertTo<double>(fields[0]));
                    if(iz>0 && gridz.back()<=gridz[iz-1]) 
                        valid=false;  // the values of z should be in increasing order
                    for(size_t iR=0; iR<size_R; iR++) 
                    {
                        double val=iR+1<fields.size() ? convertTo<double>(fields[iR+1]) : (valid=false, 0);
                        coefs[m+ncoefsAngular][iz*size_R+iR]=val;
                    }
                }
            }
            if(!valid || (configPotential->PotentialType != CPotential::PT_UNKNOWN && ptype!=configPotential->PotentialType))
            {   // load coefs only if potential type is the same as requested, or if request is not specific
                my_error(FUNCNAME,
                    "Error loading potential "+std::string(getPotentialNameByType(configPotential->PotentialType))+
                    " coefs from file "+fileName);
                return NULL;
            }
            const CPotential* pot=new CPotentialCylSpline(gridR, gridz, coefs);
            configPotential->Ncoefs_radial=size_R; 
            configPotential->Ncoefs_angular=ncoefsAngular;
            configPotential->Ncoefs_vertical=(size_z+1)/2;
            configPotential->SymmetryType=pot->symmetry();
            return pot;
#else
            my_error(FUNCNAME, "CylSpline is not supported without interp2d");
            return NULL;
#endif
        }
        else if(valid && ptype!=CDensity::PT_UNKNOWN)
        {   // process file
            if(!std::getline(strmText, buffer)) valid=false;
            splitString(buffer, "# \t", &fields);
            unsigned int ncoefsRadial=convertTo<int>(fields[0]);
            if(!std::getline(strmText, buffer)) valid=false;
            splitString(buffer, "# \t", &fields);
            unsigned int ncoefsAngular=convertTo<int>(fields[0]);
            if(!std::getline(strmText, buffer)) valid=false;
            splitString(buffer, "# \t", &fields);
            double param=convertTo<double>(fields[0]);   // meaning of this parameter depends on potential type
            if(/*ncoefsRadial==0 || ncoefsAngular==0 || */  // zero values are possible, means just a single term in expansion
                (ptype == CDensity::PT_BSE && param<0.5) || 
                (ptype == CDensity::PT_BSECOMPACT && param<=0) || 
                (ptype == CDensity::PT_SCALEFREESH && (param<0 || param>2 || ncoefsRadial!=0)) ||
                (ptype == CDensity::PT_SPLINE && ncoefsRadial<4) ) valid=false;
            std::vector< vectord > coefs;
            std::vector< double > radii;
            if(ncoefsRadial>MAX_NCOEFS_RADIAL) ncoefsRadial=MAX_NCOEFS_RADIAL;
            if(ncoefsAngular>MAX_NCOEFS_ANGULAR) ncoefsAngular=MAX_NCOEFS_ANGULAR;
            while(valid && std::getline(strmText, buffer))  // time, ignored
            {
                std::getline(strmText, buffer);  // comments, ignored
                radii.clear();
                coefs.clear();
                for(unsigned int n=0; valid && n<=ncoefsRadial; n++)
                {
                    std::getline(strmText, buffer);
                    splitString(buffer, "# \t", &fields);
                    radii.push_back(convertTo<double>(fields[0]));
                    if(((ptype == CDensity::PT_BSE || ptype == CDensity::PT_BSECOMPACT) && radii.back()!=n) || 
                        (ptype == CDensity::PT_SPLINE && n>0 && radii.back()<=radii[n-1])) 
                        valid=false;  // for BSE this field is basis function index, for spline the radii should be in increasing order
                    coefs.push_back( vectord() );
                    for(int l=0; l<=static_cast<int>(ncoefsAngular); l++)
                        for(int m=-l; m<=l; m++)
                        {
                            unsigned int fi=1+l*(l+1)+m;
                            coefs.back().push_back( fi<fields.size() ? convertTo<double>(fields[fi]) : 0);
                        }
                }
            }
            if(!valid || (configPotential->PotentialType != CPotential::PT_UNKNOWN && ptype!=configPotential->PotentialType))
            {   // load coefs only if potential type is the same as requested, or if request is not specific
                my_error(FUNCNAME, "Error loading potential "+std::string(getPotentialNameByType(configPotential->PotentialType))+
                    " coefs from file "+fileName);
                return NULL;
            }
            CPotential* pot=NULL;
            switch(ptype)
            {
            case CDensity::PT_BSE: 
                configPotential->Alpha=param;
                pot=new CPotentialBSE(/*Alpha*/param, coefs); 
                break;
            case CDensity::PT_BSECOMPACT: 
                configPotential->Rmax=param;
                pot=new CPotentialBSECompact(/*Rmax*/param, coefs); 
                break;
            case CDensity::PT_SCALEFREESH: 
                configPotential->Gamma=param;
                pot=new CPotentialScaleFreeSH(/*Gamma*/param, coefs[0]); 
                break;
            case CDensity::PT_SPLINE: 
                pot=new CPotentialSpline(radii, coefs); 
                break;
            default:  return NULL;  // shouldn't happen
            }
            configPotential->Ncoefs_radial=ncoefsRadial; 
            configPotential->Ncoefs_angular=ncoefsAngular;
            configPotential->SymmetryType=pot->symmetry();
            return pot;
        } 
        else {   // file did not contain a valid header
            my_error(FUNCNAME, "Error loading potential "+std::string(getPotentialNameByType(configPotential->PotentialType))+
                " coefs from file "+fileName);
            return NULL;
        }
    }   // PT_COEFS
    else if((configPotential->PotentialType == CPotential::PT_SPLINE ||
        //configPotential->PotentialType == CPotential::PT_CYLSPLINE || 
        configPotential->PotentialType == CPotential::PT_BSE ||
        configPotential->PotentialType == CPotential::PT_BSECOMPACT) &&
        (configPotential->DensityType == CPotential::PT_ELLIPSOIDAL ||
         configPotential->DensityType == CPotential::PT_MGE))
    {   // attempt to load ellipsoidal or MGE mass model from a text file
        const CDensity* dens=NULL;
        std::getline(strmText, buffer);
        if(buffer.substr(0, 11)=="Ellipsoidal") {
            vectord radii, inmass, axesradii, axesq, axesp;  // for loading Ellipsoidal mass profile
            while(std::getline(strmText, buffer) && !strmText.eof())
            {
                splitString(buffer, "# \t", &fields);
                if(fields.size()>=2 && ((fields[0][0]>='0' && fields[0][0]<='9') || fields[0][0]=='-' || fields[0][0]=='+'))
                {
                    radii.push_back(convertTo<double>(fields[0]));
                    inmass.push_back(convertTo<double>(fields[1]));
                    if(fields.size()>=4)  // also has anisotropy parameters
                    {
                        axesradii.push_back(radii.back());
                        axesq.push_back(convertTo<double>(fields[2]));
                        axesp.push_back(convertTo<double>(fields[3]));
                    }
                }
            }
            if(radii.size()<=2) {
                my_error(FUNCNAME, "Error loading ellipsoidal mass profile from file "+fileName);
                return NULL;
            }
            dens=new CDensityEllipsoidal(radii, inmass, axesradii, axesq, axesp);
        } else 
        if(buffer.substr(0, 3)=="MGE") {
            CDensityMGE::vectorg data;
            while(std::getline(strmText, buffer) && !strmText.eof())
            {
                splitString(buffer, "# \t", &fields);
                if(fields.size()>=2 && ((fields[0][0]>='0' && fields[0][0]<='9') || fields[0][0]=='-' || fields[0][0]=='+'))
                {
                    data.push_back(CDensityMGE::CGaussian(
                        convertTo<double>(fields[0]), convertTo<double>(fields[1]), 
                        fields.size()>=3 ? convertTo<double>(fields[2]) : 1,
                        fields.size()>=4 ? convertTo<double>(fields[3]) : 1));
                }
            }
            if(data.size()==0) {
                my_error(FUNCNAME, "Error loading MGE mass model from file "+fileName);
                return NULL;
            }
            dens=new CDensityMGE(data);
        } else {
            my_error(FUNCNAME, "Unknown density model in file "+fileName);
            return NULL;
        }
        const CPotential* pot=NULL;
        if(configPotential->PotentialType==CDensity::PT_SPLINE) {
            if(configPotential->splineRMin>0 && configPotential->splineRMax>0) 
            {
                vectord grid;
                createNonuniformGrid(configPotential->Ncoefs_radial+1, configPotential->splineRMin, 
                    configPotential->splineRMax, true, &grid);
                pot = new CPotentialSpline(configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, dens, &grid);
            } else
                pot = new CPotentialSpline(configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, dens);
        }
        else if(configPotential->PotentialType==CDensity::PT_BSE)
            pot = new CPotentialBSE(configPotential->Alpha, configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, dens);
        else if(configPotential->PotentialType==CDensity::PT_BSECOMPACT)
            pot = new CPotentialBSE(configPotential->Rmax, configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, dens);
        delete dens;
        // store coefficients in a text file, later may load this file instead for faster initialization
        if(pot!=NULL)
            writePotential(fileName+coefFileExtension(pot->PotentialType()), pot);
        return pot;
    }   // PT_ELLIPSOIDAL or PT_MGE
    else 
    {   // try to load a point mass set
        CPointMassSet<double> *points=NULL;
        strmText.close();
        CBasicIOSnapshot* snapRead=createIOSnapshotRead(fileName);
        if(snapRead!=NULL)
            snapRead->readSnapshot(points);
        delete snapRead;
        if(points==NULL || points->size()==0)
        {
            my_error(FUNCNAME, "Error loading N-body snapshot for potential "+
                std::string(getPotentialNameByType(configPotential->PotentialType))+" from file "+fileName);
            if(points!=NULL) delete points;
            return NULL;
        }
        const CPotential* pot=createPotentialFromPoints(configPotential, points);
        delete points;
        if(pot!=NULL && configPotential->PotentialType!=CDensity::PT_NB)
        {   // store coefficients in a text file, later may load this file instead for faster initialization
            writePotential(fileName+coefFileExtension(pot->PotentialType()), pot);
        }
        return pot;
    }
    return NULL;
}

/* ------ writing potential expansion coefficients to a text file -------- */

bool writePotential(const std::string &fileName, const CPotential *potential)
{
    if(!potential || !(
        potential->PotentialType()==CDensity::PT_SCALEFREESH || 
        potential->PotentialType()==CDensity::PT_BSE || 
        potential->PotentialType()==CDensity::PT_BSECOMPACT || 
        potential->PotentialType()==CDensity::PT_SPLINE || 
        potential->PotentialType()==CDensity::PT_SPHERICAL || 
        potential->PotentialType()==CDensity::PT_CYLSPLINE ) )
        return false;  // export potential coefs for BSE/Spline/Scale-FreeSH potentials only
    std::ofstream strmText(fileName.c_str(), std::ios::out);
    if(!strmText) 
        return false;  // file not writable
    if(potential->PotentialType()==CDensity::PT_SPHERICAL) {
        vectord radii, mass;
        static_cast<const CPotentialSpherical*>(potential)->getModel()->getMassProfile(&radii, &mass);
        strmText << "r\tM\n";
        for(size_t i=0; i<radii.size(); i++)
            strmText << radii[i] << "\t" << mass[i] << "\n";
        if(!strmText) return false;
        return true;
    }
#ifdef HAVE_INTERP2D
    if(potential->PotentialType()==CDensity::PT_CYLSPLINE) {
        vectord gridR, gridz;
        std::vector<vectord> coefs;
        static_cast<const CPotentialCylSpline*>(potential)->getCoefs(&gridR, &gridz, &coefs);
        int mmax=coefs.size()/2;
        strmText << "CylSpline\t#header\n" << gridR.size() << "\t#size_R\n" << mmax << "\t#m_max\n" <<
            gridz.size() << "\t#size_z\n0\t#time\n" << std::setprecision(16);
        for(int m=0; m<static_cast<int>(coefs.size()); m++) 
            if(coefs[m].size()>0) {
                strmText << (m-mmax) << "\t#m\n#z\\R";
                for(size_t iR=0; iR<gridR.size(); iR++)
                    strmText << "\t" << gridR[iR];
                strmText << "\n";
                for(size_t iz=0; iz<gridz.size(); iz++) {
                    strmText << gridz[iz];
                    for(size_t iR=0; iR<gridR.size(); iR++)
                        strmText << "\t" << coefs[m][iz*gridR.size()+iR];
                    strmText << "\n";
                }
            } 
        if(!strmText) return false;
        return true;
    }
#endif
    vectord indices;
    std::vector< vectord > coefs;
    size_t ncoefsAngular=0;
    switch(potential->PotentialType())
    {
    case CDensity::PT_SCALEFREESH: {
        const CPotentialScaleFreeSH* potSH=static_cast<const CPotentialScaleFreeSH*>(potential);
        indices.assign(1, 0);
        coefs.resize(1);
        potSH->getCoefs(&(coefs[0]));
        ncoefsAngular=potSH->getNcoefs_angular();
        strmText << "SFcoefs\t#header\n" << 0 << "\t#n_radial\n" << ncoefsAngular << "\t#n_angular\n" << 
            potential->getGamma() << "\t#gamma\n0\t#time\n";
        strmText << "#index";
        break; }
    case CDensity::PT_BSE: {
        const CPotentialBSE* potBSE=static_cast<const CPotentialBSE*>(potential);
        indices.resize(potBSE->getNcoefs_radial()+1);
        for(size_t i=0; i<indices.size(); i++) indices[i]=i*1.0;
        potBSE->getCoefs(&coefs);
        assert(coefs.size()==indices.size());
        ncoefsAngular=potBSE->getNcoefs_angular();
        strmText << "BSEcoefs\t#header\n" << potBSE->getNcoefs_radial() << "\t#n_radial\n" << ncoefsAngular << "\t#n_angular\n" << 
            potBSE->getAlpha() <<"\t#alpha\n0\t#time\n";
        strmText << "#index";
        break; }
    case CDensity::PT_BSECOMPACT: {
        const CPotentialBSECompact* potBSE=static_cast<const CPotentialBSECompact*>(potential);
        indices.resize(potBSE->getNcoefs_radial()+1);
        for(size_t i=0; i<indices.size(); i++) indices[i]=i*1.0;
        potBSE->getCoefs(&coefs);
        assert(coefs.size()==indices.size());
        ncoefsAngular=potBSE->getNcoefs_angular();
        strmText << "BSECcoefs\t#header\n" << potBSE->getNcoefs_radial() << "\t#n_radial\n" << ncoefsAngular << "\t#n_angular\n" << 
            potBSE->getRmax() <<"\t#Rmax\n0\t#time\n";
        strmText << "#index";
        break; }
    case CDensity::PT_SPLINE: {
        const CPotentialSpline* potSpline=static_cast<const CPotentialSpline*>(potential);
        potSpline->getCoefs(&indices, &coefs);
        assert(coefs.size()==indices.size());
        assert(indices[0]==0);  // leftmost radius is 0
        coefs[0].resize(1);     // retain only l=0 term for r=0, the rest is supposed to be zero
        ncoefsAngular=potSpline->getNcoefs_angular();
        strmText << "SHEcoefs\t#header\n" << potSpline->getNcoefs_radial() << "\t#n_radial\n" << ncoefsAngular << "\t#n_angular\n" << 
            0 <<"\t#unused\n0\t#time\n";
        strmText << "#radius";
        break; }
    default:
        my_error(FUNCNAME, "wrong potential type");
        return false;   // shouldn't occur, we checked potential type in the beginning
    }
    for(int l=0; l<=static_cast<int>(ncoefsAngular); l++)
        for(int m=-l; m<=l; m++)
            strmText << "\tl="<<l<<",m="<<m;  // header line
    strmText << "\n";
    for(size_t n=0; n<indices.size(); n++)
    {
        strmText << indices[n];
        strmText << "\t" << std::setprecision(14) << coefs[n][0] << std::setprecision(7);   // leading coeft should be high-accuracy at least for spline potential
        for(size_t i=1; i<coefs[n].size(); i++)
            strmText << "\t" << coefs[n][i];
        strmText << "\n";
    }
    if(!strmText) return false;
    return true;
}

// helper function that creates an equivalent spherical mass model for the given density profile; returns NULL on fail
CMassModel* createMassModel(const CDensity* density, int numNodes, double Rmin, double Rmax, const CMassModel* poten)
{
    if(density->PotentialType()==CPotential::PT_SPHERICAL) {
        return new CMassModel(*(static_cast<const CPotentialSpherical*>(density)->getModel()));
    }
    double Mbh=density->Mass(0);
    double totalMass=density->totalMass()-Mbh;
    if(totalMass<0) {
        my_message(FUNCNAME, "Warning, total mass is infinite");
        totalMass=1;   // totalMass<0 should not really happen! means that mass is infinite
    }
    if(Rmin==0) {
        double innerShellMass=totalMass*0.1/numNodes;
        if(Mbh>0) innerShellMass=std::min<double>(innerShellMass, Mbh/10);
        Rmin = std::min<double>(0.1/numNodes, density->getRadiusByMass(innerShellMass+Mbh));
        if(Mbh>0) Rmin=std::min<double>(Rmin, Mbh/10);
    }
    if(Rmax==0) {
        double outerShellMass=totalMass*(1-0.1/numNodes);
        Rmax = density->getRadiusByMass(outerShellMass+Mbh);
    }  // somewhat arbitrary choice for min/max radii, but probably reasonable
    if(Rmin<0 && totalMass<0) Rmin=Rmax/numNodes;    // trying to create a grid for an infinite-mass model.. silly enough but we could try
    if(Rmax<0 || Rmin<0 || Rmax<=Rmin) return NULL;  // some weird error
    vectord radii(numNodes), inmass(numNodes);
    createNonuniformGrid(numNodes, Rmin, Rmax, true, &radii); 
    for(int i=0; i<numNodes; i++)
        inmass[i]=density->Mass(radii[i]);
    // ensure that mass strictly increases with radius
    for(int i=1; i<numNodes; i++)
        if(inmass[i]<=inmass[i-1]) {
            radii.erase(radii.begin()+i);
            inmass.erase(inmass.begin()+i);
            numNodes--;
            i--;
        }
#ifdef DEBUGPRINT
    my_message(FUNCNAME, "Spherical mass model for r=["+convertToString(Rmin)+":"+convertToString(Rmax)+"]");
#endif
    CMassModel* massModel=new CMassModel(radii, inmass, poten);
    if(massModel->getStatus()!=CMassModel::MS_OK && massModel->getStatus()!=CMassModel::MS_DF_NEGATIVE) {
        delete massModel;
        return NULL;
    }
    return massModel;
}

}; // namespace
