#include "massmodelaxi.h"
#include <cmath>
#include <gsl/gsl_math.h>
#include <gsl/gsl_integration.h>
#include "potential.h"
#include "utils.h"
#ifdef DEBUGPRINT
#include <fstream>
#endif

namespace smile{

/// \cond INTERNAL_DOCS
struct CPotentialAxisymmetrizedParam {
    const CDensity* density;
    const CPotential* potential;
    double R,z;
    size_t c;  // index of force component (0-2) or force derivative (3-8)
};

double intAxisymRho(double phi, void* param)
{
    return ((CPotentialAxisymmetrizedParam*)param)->density->Rho(
        ((CPotentialAxisymmetrizedParam*)param)->R*cos(phi),
        ((CPotentialAxisymmetrizedParam*)param)->R*sin(phi),
        ((CPotentialAxisymmetrizedParam*)param)->z);
}

double intAxisymPhi(double phi, void* param)
{
    return ((CPotentialAxisymmetrizedParam*)param)->potential->Phi(
        ((CPotentialAxisymmetrizedParam*)param)->R*cos(phi),
        ((CPotentialAxisymmetrizedParam*)param)->R*sin(phi),
        ((CPotentialAxisymmetrizedParam*)param)->z);
}

double intAxisymForce(double phi, void* param)
{
    double cosphi=cos(phi), sinphi=sin(phi);
    double xyz[N_DIM] = { 
        ((CPotentialAxisymmetrizedParam*)param)->R*cosphi,
        ((CPotentialAxisymmetrizedParam*)param)->R*sinphi,
        ((CPotentialAxisymmetrizedParam*)param)->z };
    double force[N_DIM], forcederiv[N_DIM*(N_DIM+1)/2];
    double *forceDeriv=(((CPotentialAxisymmetrizedParam*)param)->c<N_DIM?NULL:forcederiv);
    ((CPotentialAxisymmetrizedParam*)param)->potential->Force(xyz, 0, force, forceDeriv);
    switch(((CPotentialAxisymmetrizedParam*)param)->c) {
        case 0: return force[0]*cosphi+force[1]*sinphi;
        case 2: return force[2];
        case 3: return forceDeriv[0]*pow_2(cosphi)+forceDeriv[1]*pow_2(sinphi)+2*forceDeriv[3]*cosphi*sinphi;
        case 5: return forceDeriv[2];
        default: return 0;  // don't bother about mixed derivatives..
    }
}
/// \endcond

double CPotentialAxisymmetrized::Rho(double X, double Y, double Z, double t) const
{
    if((density->symmetry() & CDensity::ST_AXISYMMETRIC) == CDensity::ST_AXISYMMETRIC) 
        return density->Rho(X, Y, Z, t);
    gsl_function F;
    CPotentialAxisymmetrizedParam param;
    F.function=&intAxisymRho;
    F.params=&param;
    param.density=density;
    param.R=sqrt(X*X+Y*Y);
    param.z=Z;
    double result, error;
    size_t neval;
    double phimax=(density->symmetry() & ST_PLANESYM)==ST_PLANESYM ? M_PI_2 : 2*M_PI;
    gsl_integration_qng(&F, 0, phimax, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result/phimax;
}

double CPotentialAxisymmetrized::Phi(double X, double Y, double Z, double t) const
{
    if(potential==NULL) return 0;
    if((potential->symmetry() & CDensity::ST_AXISYMMETRIC) == CDensity::ST_AXISYMMETRIC) 
        return potential->Phi(X, Y, Z, t);
    gsl_function F;
    CPotentialAxisymmetrizedParam param;
    F.function=&intAxisymPhi;
    F.params=&param;
    param.potential=potential;
    param.R=sqrt(X*X+Y*Y);
    param.z=Z;
    double result, error;
    size_t neval;
    double phimax=(potential->symmetry() & ST_PLANESYM)==ST_PLANESYM ? M_PI_2 : 2*M_PI;
    gsl_integration_qng(&F, 0, phimax, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    return result/phimax;
}

void CPotentialAxisymmetrized::Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv) const
{
    if(potential==NULL) return;
    if((potential->symmetry() & CDensity::ST_AXISYMMETRIC) == CDensity::ST_AXISYMMETRIC) {
        potential->Force(xyz, t, force, forceDeriv);
        return;
    }
    gsl_function F;
    CPotentialAxisymmetrizedParam param;
    F.function=&intAxisymForce;
    F.params=&param;
    param.potential=potential;
    param.R=sqrt(xyz[0]*xyz[0]+xyz[1]*xyz[1]);
    param.z=xyz[2];
    double result, error;
    size_t neval;
    double phimax=(potential->symmetry() & ST_PLANESYM)==ST_PLANESYM ? M_PI_2 : 2*M_PI;
    param.c=0;
    gsl_integration_qng(&F, 0, phimax, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    force[0]=result/phimax;  // we assume that the y-coordinate is zero
    force[1]=0;
    param.c=2;
    gsl_integration_qng(&F, 0, phimax, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    force[2]=result/phimax;
    if(forceDeriv!=NULL) {
        param.c=3;
        gsl_integration_qng(&F, 0, phimax, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
        forceDeriv[0]=result/phimax;
        param.c=5;
        gsl_integration_qng(&F, 0, phimax, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
        forceDeriv[2]=result/phimax;
    }
}

/// \cond INTERNAL_DOCS
struct CMassIntegrandAxiParam {
    const CDensity* density;
    double R, Z;
};

double massModelAxiRhoZ(double Zscaled, void* param)
{
    double Z=Zscaled/(1-Zscaled);
    return ((CMassIntegrandAxiParam*)param)->density->Rho(((CMassIntegrandAxiParam*)param)->R, 0, Z) / pow_2(1-Zscaled);
}
double massModelAxiRhoR(double Rscaled, void* param)
{
    double R=Rscaled/(1-Rscaled);
    return ((CMassIntegrandAxiParam*)param)->density->Rho(R, 0, ((CMassIntegrandAxiParam*)param)->Z) * R / pow_2(1-Rscaled);
}
double massModelAxiRadialMass(double R, void* param)
{
    gsl_function F;
    F.function=&massModelAxiRhoZ;
    F.params=param;
    ((CMassIntegrandAxiParam*)param)->R=R;
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, 1, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    //gsl_integration_qagiu(&F, 0, 0, EPSREL_DENSITY_INT, 1000, ((CMassIntegrandAxiParam*)param)->ws, &result, &error);
    return 2*M_PI*R*2*result;
}
double massModelAxiVerticalMass(double Z, void* param)
{
    gsl_function F;
    F.function=&massModelAxiRhoR;
    F.params=param;
    ((CMassIntegrandAxiParam*)param)->Z=Z;
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, 1, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
    //gsl_integration_qagiu(&F, 0, 0, EPSREL_DENSITY_INT, 1000, ((CMassIntegrandAxiParam*)param)->ws, &result, &error);
    return 2*M_PI*result*2;
}
/// \endcond

CMassModelAxisymmetric::CMassModelAxisymmetric(const CDensity* density_src, size_t ngrid_R, size_t ngrid_z, bool initSplineMass)
{
    spl_mass_R=spl_mass_z=NULL;
    const CDensity* density = 
        new CPotentialAxisymmetrized(NULL, density_src);  // create temporary rotationally-symmetric density
    double result, error;
    size_t neval;
    gsl_function F;
    CMassIntegrandAxiParam param;
    F.params=&param;
    param.density=density;

    totalMass=density->totalMass();
    if(totalMass<0)
        return;  // weird potential...

    // determine the extent of grid in cylindrical radius
    double innerShellMass=totalMass*0.1/ngrid_R;
    double outerShellMass=totalMass*(1-0.1/ngrid_R);
    double rout = density->getRadiusByMass(outerShellMass);
    double rin = std::min<double>(density->getRadiusByMass(totalMass*0.5)/ngrid_R, 
        density->getRadiusByMass(innerShellMass));  // somewhat arbitrary choice for min/max radii, but probably reasonable
    if(rout<0 || rin<0 || rout<rin) 
        return;  // some weird error
    F.function=&massModelAxiRadialMass;
    do{
        gsl_integration_qng(&F, 0, rin, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
        if(result>totalMass/ngrid_R) rin/=1.6;
    } while(result>totalMass/ngrid_R);
    if(rin>rout/ngrid_R)  rin=rout/ngrid_R;  // avoid creating a grid which is refined towards outer boundary
    createNonuniformGrid(ngrid_R, rin, rout, true, &grid_R); 

    // determine the extent of grid in z
    F.function=&massModelAxiVerticalMass;
    double zin=rin, zout=rout;
    do{
        gsl_integration_qng(&F, 0, zin, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
        if(result>totalMass/ngrid_z) zin/=1.6;
    } while(result>totalMass/ngrid_z);
    if(zin<=0 || zin>zout/ngrid_z) 
        zin=zout/ngrid_z;
    createNonuniformGrid(ngrid_z, zin, zout, true, &grid_z); 
    if(initSplineMass) {
#ifdef DEBUGPRINT
        my_message(FUNCNAME, 
            "Computing mass ["+convertToString(totalMass)+
            "] on a cylindrical grid [R="+convertToString(rin)+":"+convertToString(rout)+
            ", Z="+convertToString(zin)+":"+convertToString(zout)+"]");
#endif 
        // compute mass (integrated in z direction from 0 to infinity) as a function of R
        vectord values_omp(ngrid_R-1);
#pragma omp parallel for
        for(int iR=0; iR<static_cast<int>(ngrid_R)-1; iR++) {
            gsl_function F_omp;
            CMassIntegrandAxiParam param_omp;
            F_omp.function=&massModelAxiRadialMass;
            F_omp.params=&param_omp;
            param_omp.density=density;
            double error_omp;
            size_t neval_omp;
            gsl_integration_qng(&F_omp, grid_R[iR], grid_R[iR+1], 0, EPSREL_DENSITY_INT, 
                &(values_omp[iR]), &error_omp, &neval_omp);
        }
        vectord grid_logR, grid_massR;
        double mass=0;
        for(size_t iR=0; iR<ngrid_R-1; iR++) {
            if(values_omp[iR]>0) {
                mass+=values_omp[iR];
                grid_logR.push_back(log(grid_R[iR+1]));
                grid_massR.push_back(mass);
            }
        }
        // integrate to (almost) infinity to obtain total mass
        gsl_integration_qng(&F, grid_R[ngrid_R-1], grid_R[ngrid_R-1]*100, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
        if(result<=0) result=(grid_massR.back()-grid_massR[grid_massR.size()-2]);
        totalMass=std::max<double>(totalMass, result+mass);
        // compute mass (integrated in R direction from 0 to infinity) as a function of z (=total mass in slice from -z to +z)
        values_omp.resize(ngrid_z-1);
#pragma omp parallel for
        for(int iz=0; iz<static_cast<int>(ngrid_z)-1; iz++) {
            gsl_function F_omp;
            CMassIntegrandAxiParam param_omp;
            F_omp.function=&massModelAxiVerticalMass;
            F_omp.params=&param_omp;
            param_omp.density=density;
            double error_omp;
            size_t neval_omp;
            gsl_integration_qng(&F_omp, grid_z[iz], grid_z[iz+1], 0, EPSREL_DENSITY_INT, 
                &(values_omp[iz]), &error_omp, &neval_omp);
        }
        vectord grid_logz, grid_massz;
        mass=0;
        for(size_t iz=0; iz<ngrid_z-1; iz++) {
            if(values_omp[iz]>0) {
                mass+=values_omp[iz];
                grid_logz.push_back(log(grid_z[iz+1]));
                grid_massz.push_back(mass);
            }
        }
        // integrate to (almost) infinity to obtain total mass
        gsl_integration_qng(&F, grid_z.back(), grid_z.back()*100, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
        if(result<=0) result=(grid_massz.back()-grid_massz[grid_massz.size()-2]);
        totalMass=std::max<double>(totalMass, result+mass);
        // scale M(R) and M(z) using the estimate for total mass, and init corresponding splines
        grid_R.assign(1,0);
        for(size_t iR=0; iR<grid_massR.size(); iR++) {
            grid_R.push_back(exp(grid_logR[iR]));
            grid_massR[iR]=log(grid_massR[iR]/(totalMass-grid_massR[iR]));
        }
        spl_mass_R=gsl_spline_alloc(gsl_interp_cspline, grid_massR.size());
        gsl_spline_init(spl_mass_R, &(grid_logR.front()), &(grid_massR.front()), grid_massR.size());
        grid_z.assign(1,0);
        for(size_t iz=0; iz<grid_massz.size(); iz++) {
            grid_z.push_back(exp(grid_logz[iz]));
            grid_massz[iz]=log(grid_massz[iz]/(totalMass-grid_massz[iz]));
        }
        spl_mass_z=gsl_spline_alloc(gsl_interp_cspline, grid_massz.size());
        gsl_spline_init(spl_mass_z, &(grid_logz.front()), &(grid_massz.front()), grid_massz.size());
        // sanity check
        bool monoR=check_spline_monotonic(spl_mass_R);
        bool monoZ=check_spline_monotonic(spl_mass_z);
        if(!monoR || !monoZ)
            my_message(FUNCNAME,
                std::string("Warning, ")+(monoR?"":"M(R)")+(monoZ?"":monoR?"M(z)":" and M(z)")+
                " is not monotonic, weird behaviour is likely");
    }  // if initSplineMass
#ifdef DEBUGPRINT
    my_message(FUNCNAME,
        "Grid in R from "+convertToString(grid_R[1])+" to "+convertToString(grid_R.back())+
        ", in z from "+convertToString(grid_z[1])+" to "+convertToString(grid_z.back()));
#endif
    delete density;
}

CMassModelAxisymmetric::~CMassModelAxisymmetric()
{
    if(spl_mass_R!=NULL)
        gsl_spline_free(spl_mass_R);
    if(spl_mass_z!=NULL)
        gsl_spline_free(spl_mass_z);
}

double CMassModelAxisymmetric::mass_R(double R) const
{
    if(spl_mass_R==NULL || R<=0) return 0;
    double scaledmass=gsl_spline_evalx(spl_mass_R, log(R), NULL);
    return totalMass/(1+exp(-scaledmass));
}

double CMassModelAxisymmetric::mass_z(double z) const
{
    if(spl_mass_z==NULL || z==0) return 0;
    double scaledmass=gsl_spline_evalx(spl_mass_z, log(fabs(z)), NULL);
    return totalMass/(1+exp(-scaledmass));
}

double CMassModelAxisymmetric::Sigma(double R) const
{
    if(spl_mass_R==NULL || R<=0) return 0;
    double logR=log(R);
    double scaledmass=gsl_spline_evalx(spl_mass_R, logR, NULL);
    double zz=exp(-scaledmass);
    double scaledmder=gsl_spline_eval_derivx(spl_mass_R, logR, NULL);
    return scaledmder * totalMass*zz/pow_2(1+zz) / (2*M_PI*R*R);
}

/// \cond INTERNAL_DOCS
typedef double (CMassModelAxisymmetric::*memberFncAxi)(double x) const;
struct CMassModelAxiFinderParam {
    const CMassModelAxisymmetric* model;
    memberFncAxi memberFunction;
    double value;
};
double massModelAxiFinder(double x, void* param)
{
    memberFncAxi fnc=((CMassModelAxiFinderParam*)param)->memberFunction;
    const CMassModelAxisymmetric& model=*(((CMassModelAxiFinderParam*)param)->model);
    double val= x==1 ? model.getTotalMass() : (model.*fnc)(x/(1-x));
    return val - ((CMassModelAxiFinderParam*)param)->value;
}
/// \endcond

double CMassModelAxisymmetric::find_R_mass(double m) const
{
    if(m>=totalMass || spl_mass_R==NULL)
        return -1;
    gsl_function F;
    CMassModelAxiFinderParam param;
    F.function=&massModelAxiFinder;
    F.params=&param;
    param.model=this;
    param.memberFunction=&CMassModelAxisymmetric::mass_R;
    param.value=m;
    double R1=findRoot(&F, 0, 1, -1);
    if(R1>=0)
        return R1/(1-R1);
    else return -1;
}

double CMassModelAxisymmetric::find_z_mass(double m) const
{
    if(m>=totalMass || spl_mass_z==NULL)
        return -1;
    gsl_function F;
    CMassModelAxiFinderParam param;
    F.function=&massModelAxiFinder;
    F.params=&param;
    param.model=this;
    param.memberFunction=&CMassModelAxisymmetric::mass_z;
    param.value=m;
    double z1=findRoot(&F, 0, 1, -1);
    if(z1>=0)
        return z1/(1-z1);
    else return -1;
}

/// \cond INTERNAL_DOCS
struct CJeansIntegrandAxiParam {
    const CPotential* potential;
    const CDensity* density;
    double R;
    double Zscalednext;
};
double jeansIntegrandAxiZ(double Zscaled, void* param)
{
    double xyz[N_DIM] = { ((CJeansIntegrandAxiParam*)param)->R, 0, Zscaled/(1-Zscaled) };
    double force[N_DIM];
    ((CJeansIntegrandAxiParam*)param)->potential->Force(xyz, 0, force);
    return ((CJeansIntegrandAxiParam*)param)->density->Rho(xyz[0], xyz[1], xyz[2])*(-force[2]) / pow_2(1-Zscaled);
}
double jeansIntegrandAxiZRho(double Zscaled, void* param)
{
    double xyz[N_DIM] = { ((CJeansIntegrandAxiParam*)param)->R, 0, Zscaled/(1-Zscaled) };
    return ((CJeansIntegrandAxiParam*)param)->density->Rho(xyz[0], xyz[1], xyz[2]) * 
        (Zscaled - ((CJeansIntegrandAxiParam*)param)->Zscalednext);
}
/// \endcond

#ifdef HAVE_INTERP2D
CMassModelAxisymmetricJeans::CMassModelAxisymmetricJeans(const CPotential* potential_src, 
    size_t ngrid_R, size_t ngrid_z, double _beta_m, const CDensity* density_src) :
  CMassModelAxisymmetric(density_src==NULL?potential_src:density_src, ngrid_R, ngrid_z, false),
  beta_m(_beta_m)
{
#ifdef DEBUGPRINT
    my_message(FUNCNAME, "Solving Jeans equations in the cylindrical coordinates");
#endif

    ngrid_R=grid_R.size();
    ngrid_z=grid_z.size();  // could be less than required, if corrected in the constructor of CMassModelAxisymmetric
    spl_vmer=spl_vphi=NULL;
    if(ngrid_R==0 || ngrid_z==0) return;
    const CPotential* potential = 
        //(potential_src->symmetry() & CDensity::ST_AXISYMMETRIC) == CDensity::ST_AXISYMMETRIC ? potential_src : 
        new CPotentialAxisymmetrized(potential_src);  // create temporary rotationally-symmetric potential
    const CDensity* density = density_src==NULL ? potential : 
        (density_src->symmetry() & CDensity::ST_AXISYMMETRIC) == CDensity::ST_AXISYMMETRIC ? density_src : 
        new CPotentialAxisymmetrized(NULL, density_src);  // create temporary rotationally-symmetric density

    // determine the velocity dispersion in the meridional plane at the nodes of 2d grid in R,z
    vectord values_vmer2(ngrid_R*ngrid_z);
    grid_R[0]=grid_R[1]*0.01;
    grid_z[0]=grid_z[1]*0.01;  // slightly offset zero to avoid possible singularities
#pragma omp parallel for
    for(int iR=0; iR<static_cast<int>(ngrid_R); iR++) {
        double result, error;
        size_t neval;
        gsl_function F;
        CJeansIntegrandAxiParam param;
        F.params=&param;
        param.potential=potential;
        param.density=density;     // could be the same as potential, but doesn't need to.
        param.R=grid_R[iR];
        double sumresults=0;
        for(int iz=ngrid_z-1; iz>=0; iz--) {
            F.function=&jeansIntegrandAxiZ;
            double Zscaled=grid_z[iz]/(1+grid_z[iz]);
            double Zscaledplus1 = iz==static_cast<int>(ngrid_z)-1 ? 1 : grid_z[iz+1]/(1+grid_z[iz+1]);
            double Zscaledminus1= iz==0 ? 0 : grid_z[iz-1]/(1+grid_z[iz-1]);
            gsl_integration_qng(&F, Zscaled, Zscaledplus1, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
            sumresults+=result;

            F.function=&jeansIntegrandAxiZRho;
            double rhoavg=0;
            param.Zscalednext=Zscaledplus1;
            gsl_integration_qng(&F, Zscaled, Zscaledplus1, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
            rhoavg += result * 2/(Zscaledplus1-Zscaledminus1)/(Zscaled-Zscaledplus1);
            if(iz>0) {
                param.Zscalednext=Zscaledminus1;
                gsl_integration_qng(&F, Zscaledminus1, Zscaled, 0, EPSREL_DENSITY_INT, &result, &error, &neval);
                rhoavg += result * 2/(Zscaledplus1-Zscaledminus1)/(Zscaled-Zscaledminus1);
            }
            //double rho=density->Rho(grid_R[iR], 0, grid_z[iz]);
            double vmax2=std::max<double>(-potential->Phi(grid_R[iR], 0, grid_z[iz]), 0);
            // prevent unphysically large values that could arise if rhoavg is too small
            // this could happen, e.g., when angular expansion in the potential was not sufficiently accurate
            values_vmer2[INDEX_2D(iR, iz, ngrid_R, ngrid_z)] = (rhoavg<=0 || sumresults<=0) ? 0 : 
                std::min<double>(sumresults/rhoavg, vmax2);
        }
    }
    grid_R[0]=grid_z[0]=0;  // set back origin
    spl_vmer=interp2d_spline_alloc(interp2d_bilinear, ngrid_R, ngrid_z);
    if(spl_vmer==NULL) 
        return;  // error occured
    if(interp2d_spline_init(spl_vmer, &(grid_R.front()), &(grid_z.front()), 
        &(values_vmer2.front()), ngrid_R, ngrid_z) != GSL_SUCCESS) 
    {
        interp2d_spline_free(spl_vmer);
        spl_vmer=NULL;
        return;
    }

    // determine the velocity dispersion in the azimuthal direction at the nodes of the same 2d grid
    vectord values_vphi2(ngrid_R*ngrid_z);
    for(size_t iR=0; iR<ngrid_R; iR++) {
        double dR = (iR<ngrid_R-1 ? grid_R[iR+1]-grid_R[iR] : grid_R[iR]-grid_R[iR-1])*0.1;
        double R1 = iR==0 ? grid_R[1]*0.01 : iR==ngrid_R-1 ? grid_R[iR]-dR : grid_R[iR]-dR*0.5;
        double R=grid_R[iR];
        if(iR==0) R=grid_R[1]*0.01;  // avoid possible singularity
        for(size_t iz=0; iz<ngrid_z; iz++) {
            double z=grid_z[iz];
            if(iz==0) z=grid_z[1]*0.01;  // same reason
            double der = (density->Rho(R1+dR, 0, z)*sigma_z(R1+dR, z) - density->Rho(R1, 0, z)*sigma_z(R1, z)) / dR;
            double xyz[N_DIM]={R,0,z};
            double force[N_DIM];
            potential->Force(xyz, 0, force);
            double rho=density->Rho(R, 0, z);
            double val = (sigma_z(R, z) + R/rho*der)/(1-beta_m) + R*-force[0];
            if(val<0 || rho<=0)
                val=0;
            // prevent unphysically large values in case of inaccurate potential
            val=std::min<double>(val, std::max<double>(-1.1*potential->Phi(grid_R[iR], 0, grid_z[iz]), 0));
            values_vphi2[INDEX_2D(iR, iz, ngrid_R, ngrid_z)] = val;
        }
    }
    spl_vphi=interp2d_spline_alloc(interp2d_bilinear, ngrid_R, ngrid_z);
    if(spl_vphi==NULL) 
        return;  // error occured
    if(interp2d_spline_init(spl_vphi, &(grid_R.front()), &(grid_z.front()), 
        &(values_vphi2.front()), ngrid_R, ngrid_z) != GSL_SUCCESS) 
    {
        interp2d_spline_free(spl_vphi);
        spl_vphi=NULL;
        return;
    }

#ifdef DEBUGPRINT
    std::ofstream stm("grid2");
    for(size_t iR=0; iR<ngrid_R; iR++) {
        for(size_t iz=0; iz<ngrid_z; iz++)
            stm << grid_R[iR] << "\t" << grid_z[iz] << "\t" << 
            potential->Phi(grid_R[iR], 0, grid_z[iz]) << "\t" << potential->Rho(grid_R[iR], 0, grid_z[iz]) << "\t" << 
            values_vmer2[INDEX_2D(iR, iz, ngrid_R, ngrid_z)] << "\t" <<
            values_vphi2[INDEX_2D(iR, iz, ngrid_R, ngrid_z)] << "\n";
        stm << "\n";
    }
    stm.close();
    stm.open("gridR");
    stm << "R\tM(R)\tSigma(surf.dens.)\tOmega^2\tkappa^2\tnu^2\tsigma_z\tv_phi\tv_circ\tsigma_z(iso)\n";
    for(size_t iR=1; iR<ngrid_R; iR++) {
        double R=grid_R[iR];
        double xyz[N_DIM]={R,0,0};
        double force[N_DIM], forceDeriv[N_DIM*(N_DIM+1)/2];
        potential->Force(xyz, 0, force, forceDeriv);
        // determine circular, epicyclic and vertical frequencies in the disk plane
        double Omega = -force[0]/R;
        double kappa = 3*Omega - forceDeriv[0];
        double nu    = -forceDeriv[2];
        stm << R << "\t" << mass_R(R) << "\t" << Sigma(R) << "\t" << 
            Omega << "\t" << kappa << "\t" << nu << "\t" << sigma_z(R,0) << "\t" << 
            sqrt(std::max<double>(0, vphisq_av(R,0)-sigma_z(R,0))) << "\t" <<
            sqrt(-force[0]*R) << "\t" << 
            (M_PI_2*pow_2(Sigma(R))/density->Rho(R, 0, 0)) << "\n";
    }
#endif
    //if((potential_src->symmetry() & CDensity::ST_AXISYMMETRIC) != CDensity::ST_AXISYMMETRIC) 
        delete potential;
    if(density_src!=NULL && (density_src->symmetry() & CDensity::ST_AXISYMMETRIC) != CDensity::ST_AXISYMMETRIC) 
        delete density;
}

CMassModelAxisymmetricJeans::~CMassModelAxisymmetricJeans()
{
    if(spl_vmer!=NULL)
        interp2d_spline_free(spl_vmer);
    if(spl_vphi!=NULL)
        interp2d_spline_free(spl_vphi);
}

double CMassModelAxisymmetricJeans::spl_eval(double R, double z, interp2d_spline* spl) const
{
    if(spl==NULL)
        return -1;
    double maxR=spl_vmer->interp_object.xmax;
    double maxz=spl_vmer->interp_object.ymax;
    R=fabs(R); z=fabs(z);
    double factor=1.0;
    if(R>maxR || z>maxz) { // will extrapolate using the asymptotical form  v^2 ~ 1/r
        factor=std::min<double>(maxz/z, maxR/R)*0.9999999;
        R*=factor;
        z*=factor;
    }
    double val=interp2d_spline_eval(spl, R, z, NULL, NULL);
    return val*factor;
    return 0;
}
#endif

}; // namespace
