#include "orbitlib.h"
#include <algorithm>
#include <cassert>
#include "orbitint.h"
#include "orbitan.h"
#include "potential.h"
#include "utils.h"

namespace smile{
#if 0
/// function that rounds double to NUM_DIGITS_ROUND significant digits. Used to initialize ICs (presently unused)
template<typename NumT> double xroundd(NumT x)
{
    if(x==0) return 0;
    return StringVariant(x, NUM_DIGITS_ROUND).toDouble();
}
template<typename NumT> float xroundf(NumT x)
{
    if(x==0) return 0;
    return StringVariant(x, NUM_DIGITS_ROUND).toFloat();
}
initCond(xroundf(_initCond.Pos[0]),xroundf(_initCond.Pos[1]),xroundf(_initCond.Pos[2]),
         xroundf(_initCond.Vel[0]),xroundf(_initCond.Vel[1]),xroundf(_initCond.Vel[2])),
#endif

COrbitLauncher::COrbitLauncher(const CPotential* _potential, const CConfigOrbit& _configOrbit) :
  potential(_potential->clone()), configOrbit(new CConfigOrbit(_configOrbit))   // clone potential and bitwise copy config
{};

COrbitLauncher::~COrbitLauncher()
{
    delete potential;   // it was cloned when creating this object
    delete configOrbit; // same reason
    for(vectorRuntimeFncCreators::const_iterator iter=creators.begin(); iter!=creators.end(); ++iter)
        delete (*iter); // these pointers were not created inside this class, but were "taken over"...
}
template<typename NumT> COrbit* COrbitLauncher::createOrbit(const CPosVelPoint<NumT>& initCond, NumT intTime, NumT maxTime) const
{
    return new COrbit(configOrbit, potential, initCond, &creators, intTime, maxTime);
}
// explicit instantiations
template COrbit* COrbitLauncher::createOrbit(const CPosVelPoint<float >& initCond, float  intTime, float  maxTime) const;
template COrbit* COrbitLauncher::createOrbit(const CPosVelPoint<double>& initCond, double intTime, double maxTime) const;

/* --------------  COrbitDesc class - shortened orbit data ------------- */
template<typename NumT> COrbitDesc::COrbitDesc(
    const CPosVelPoint<NumT> &_initCond, const NumT _intTime, const NumT _maxTime, 
    const double _weight, const double _weightPrior, const vectorInformation* _info) :
  initCond(_initCond), intTime(static_cast<float>(_intTime)), maxTime(static_cast<float>(_maxTime)), 
  weight(_weight), weightPrior(_weightPrior)
{
    state=_info==NULL ? OS_INITIALIZED : OS_DONE;
    orbit=NULL;
    if(_info!=NULL)
        info.assign(_info->begin(), _info->end());
}
template COrbitDesc::COrbitDesc(
    const CPosVelPoint<float> &_initCond, const float _intTime, const float _maxTime, 
    const double _weight, const double _weightPrior, const vectorInformation* _info);
template COrbitDesc::COrbitDesc(
    const CPosVelPoint<double> &_initCond, const double _intTime, const double _maxTime, 
    const double _weight, const double _weightPrior, const vectorInformation* _info);

COrbitDesc::~COrbitDesc()
{
    if(orbit!=NULL)
        delete orbit;
    for(vectorInformation::const_iterator iter=info.begin(); iter!=info.end(); iter++)
        delete (*iter);
}

void COrbitDesc::run(const COrbitLauncher* launcher)
{
    if(state!=OS_INITIALIZED && state!=OS_PREPARING) 
        return;  // shouldn't happen though...
    if(launcher==NULL) 
        return;
    state=OS_RUNNING;
    for(vectorInformation::const_iterator iter=info.begin(); iter!=info.end(); iter++)  // delete all existing data
        delete (*iter);
    info.clear();
    try{
        orbit = launcher->createOrbit(initCond, intTime, maxTime);
        orbit->run();
        intTime=(float)orbit->getIntTime();  // may be longer than desired
        if(intTime<=0)      // indicates a problem in orbit integration (e.g. energy was unbound)
            weightPrior=0;  // exclude this orbit from Schwarzschild model
        // get information from any additional timestep functions that were attached (including orbit analysis, trajectory sampler, etc)
        for(size_t i=0; i<orbit->getInfoNum(); i++)
        {
            const CBasicInformation* inf=orbit->getNewInfo(i);
            if(inf!=NULL) info.push_back(inf);
        }
        delete(orbit); 
        orbit=NULL;
    }
    catch(const std::bad_alloc&) {
        my_error(FUNCNAME, "Not enough memory in orbit integration");
        info.push_back(new COrbitInformation<float>("Failed"));
        // do nothing..
    }
    catch(...) {
        my_error(FUNCNAME, "An exception occured during orbit integration");
        info.push_back(new COrbitInformation<float>("Failed"));
    }
    state=intTime>0 ? OS_DONE : OS_FAILED;
}

void COrbitDesc::halt()
{
    if(state!=OS_PREPARING && state!=OS_RUNNING)
        return;
    if(orbit!=NULL)
    {
        state=OS_NEEDTOTERMINATE;
        orbit->halt();
    }
}

std::string COrbitDesc::toString() const 
{
    std::string strInfo;
    for(size_t i=0; i<getInfoNum(); i++)
        strInfo += getInfo(i)->toString();
    strInfo += "Int.time=" + convertToString(intTime) +
        "\nOrbit weight=" + convertToString(weight, 3) + "\n";
    return strInfo;
}

/* --------------  COrbitLibrary class - array of COrbitDesc ------------- */

COrbitLibrary::~COrbitLibrary() 
{
    for(std::vector<COrbitDesc*>::iterator iter=OrbitList.begin(); iter!=OrbitList.end(); ++iter)
        delete (*iter);
}

size_t COrbitLibrary::numComplete() const
{
    size_t nComplete=0;
    for(std::vector<COrbitDesc*>::const_iterator iter=OrbitList.begin(); iter!=OrbitList.end(); ++iter)
        if( (*iter)->getState()==COrbitDesc::OS_DONE) nComplete++;
    return nComplete;
}

int COrbitLibrary::findUnstartedOrbit() 
{
    for(size_t ind=nextOrbitInd; ind<OrbitList.size(); ind++)
        if( OrbitList[ind]->getState()==COrbitDesc::OS_INITIALIZED)
        {
            OrbitList[ind]->setState(COrbitDesc::OS_PREPARING);
            nextOrbitInd=ind+1;
            return static_cast<int>(ind);
        }
    return -1;
}

void COrbitLibrary::assignInitialConditions(const CPointMassSet<float>* points, const COrbitLauncher* newLauncher, 
    double intTimeInPeriods, double intTimeMaxAdaptive, double intTimeMinAbs)
{
    if(newLauncher!=NULL)
        launcher=newLauncher;
    const CPotential* potential=launcher->getPotential();
    if(!potential) return;
    if(points!=NULL) {  // clear existing orbit list
        for(std::vector<COrbitDesc*>::iterator iter=OrbitList.begin(); iter!=OrbitList.end(); ++iter)
            delete (*iter);
    }
    nextOrbitInd=0;  // reset the counter used in findUnstartedOrbit
    int count = points!=NULL ? points->size() : OrbitList.size();
    OrbitList.resize(count);
#pragma omp parallel for schedule(dynamic, 256)
    for(int ind=0; ind<count; ind++) {
        CPosVelPoint<float> IC = points!=NULL ? points->at(ind).first : OrbitList[ind]->getInitCond();
        float intTime, maxTime;
        double weight, weightPrior;
        if(points==NULL && intTimeInPeriods==0) {  // retain intTime and maxTime from existing orbit list
            intTime=OrbitList[ind]->getIntTime();
            maxTime=OrbitList[ind]->getMaxTime();
        } else {  // use provided values for integration time
            double Torb = computeTimeUnit(potential, IC /*, launcher->getConfig()->Omega*/);
            intTime = static_cast<float>( std::max<double>(intTimeMinAbs, intTimeInPeriods*Torb) );
            maxTime = static_cast<float>( intTimeMaxAdaptive*Torb );
        }
        if(points==NULL) {  // retain weights and weight priors from existing orbit list
            weight=OrbitList[ind]->getWeight();
            weightPrior=OrbitList[ind]->getWeightPrior();
            delete OrbitList[ind];
        } else {
            weight=0;
            weightPrior=points->at(ind).second;
        }
        OrbitList[ind]=new COrbitDesc(IC, intTime, maxTime, weight, weightPrior);
    }
}

void COrbitLibrary::rewind()
{
    nextOrbitInd=0;  // reset the counter used in findUnstartedOrbit
    for(std::vector<COrbitDesc*>::iterator iter=OrbitList.begin(); iter!=OrbitList.end(); ++iter)
        (*iter)->setState(COrbitDesc::OS_INITIALIZED);
}

void COrbitLibrary::removeUnused()
{
    if(OrbitList.size()==0) return;
    std::vector<COrbitDesc*>::iterator iter=OrbitList.end();
    while(iter!=OrbitList.begin()) {
        --iter;
        if((*iter)->getState()!=COrbitDesc::OS_DONE) {
            delete (*iter);
            iter=OrbitList.erase(iter);
        }
    }
}

std::string COrbitLibrary::getOrbitPopulation(const CBasicOrbitFilteringFnc* filter, const CBasicOrbitFilteringFnc* chaos, bool useWeight) const
{
    if(OrbitList.size()==0) return "Error: no orbits";
    std::map<std::string, double> families, familiesr, familiesc;
    double totalweight=0, weightreg=0, weightch=0;
    for(size_t o=0; o<OrbitList.size(); o++)
    {
        double w = useWeight ? OrbitList[o]->getWeight() : 1.0+getRandomNumber()*0.00001;   // add random to avoid duplicates
        const COrbitInformation<float>* infoOrbit=static_cast<const COrbitInformation<float>*>(OrbitList[o]->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
        double acceptFactor=filter!=NULL ? (filter->eval(OrbitList[o])>=0.5 ? 1 : 0) : 1;
        double chaosFactor =chaos!=NULL ? chaos->eval(OrbitList[o]) : 1;
        if(chaosFactor<0) chaosFactor=0;
        if(chaosFactor>1) chaosFactor=1;
        w*=acceptFactor;
        if(w==0 || infoOrbit==NULL) continue;
        std::string family=infoOrbit->getDescription();
        if(family.substr(0, 8)=="chaotic ") family=family.erase(0, 8);
        if(family.find("SAT")!=std::string::npos && family.find("resonance")!=std::string::npos) family="SAT";  else // merge all subgroups like 'SAT 6:6:7' into one
        if(family.find("LAT")!=std::string::npos && family.find("resonance")!=std::string::npos) family="LAT";  else
        if(family.find("SAU")!=std::string::npos && family.find("resonance")!=std::string::npos) family="Saucer";  else
        if(family.find("resonance")!=std::string::npos || family.find("thin")!=std::string::npos) family="resonant box"; else
        if(family.find("box")!=std::string::npos) family="box"; 

        families[family] +=w;
        familiesc[family]+=w*chaosFactor;
        familiesr[family]+=w*(1-chaosFactor);
        totalweight+=w;
        weightreg  +=w*(1-chaosFactor);
        weightch   +=w*chaosFactor;
    }
    std::vector<double> weights;
    for(std::map<std::string, double>::const_iterator iterm=families.begin(); iterm!=families.end(); iterm++) weights.push_back(iterm->second);
    std::sort(weights.begin(), weights.end(), std::greater<double>());
    int nf=0; 
    if(totalweight==0) return "No weights assigned to orbits";
    std::string info = convertToString(weightch/totalweight*100,3)+"% chaotic\n";
    std::vector<double>::const_iterator iterw=weights.begin();
    while(nf<8 && iterw!=weights.end())
    {
        // find key in map
        std::string f;
        for(std::map<std::string, double>::const_iterator iterm=families.begin(); f.empty() && iterm!=families.end(); iterm++) 
            if(iterm->second == *iterw) f=iterm->first;
        info +=  f + ": " + convertToString((familiesr[f]+familiesc[f])/totalweight*100,3) + "% (" + 
            convertToString(familiesc[f]/totalweight*100,3) + "%)\n";
        nf++;
        ++iterw;
    }
    return info;
}

std::string COrbitLibrary::exportNbody(size_t numPoints, const CBasicOrbitFilteringFnc* massRefineFactor, 
    CPointMassSet<float>* *result, COrbitRuntimeTrajSampleCreator* trajSampleCreator)
{
    assert(result);
    size_t maxSamplingPoints=0;
    size_t numOrbitsUsed=0;
    size_t numOrbitsReintegrate=0;
    double sumWeight=0;   // total weight of all orbits (adjusted by refinement factors)
    double totalMass=0;   // total mass of model (sum of orbit weights, unadjusted)
    // mass refinement if necessary
    vectord orbitMassFactor(OrbitList.size(), 1.0);
    if(massRefineFactor!=NULL)
    {
        for(size_t o=0; o<OrbitList.size(); o++)
        {
            orbitMassFactor[o] = massRefineFactor->eval(OrbitList[o]);
            if(orbitMassFactor[o]<=0)
            {
                return("Error, refinement factor is non-positive");
            }
            sumWeight += OrbitList[o]->getWeight() / orbitMassFactor[o];
            totalMass += OrbitList[o]->getWeight();
        }
    }
    else 
    {
        for(size_t o=0; o<OrbitList.size(); o++)
            sumWeight += OrbitList[o]->getWeight();
        totalMass=sumWeight;
    }
    if(sumWeight<=0 || totalMass<=0) 
    {
        return("Error, total orbit weight is zero, need to assign weights first (run Schwarzschild model optimization)");
        // in principle, could use equal weights in this case, but it makes little sense
    }
    if((*result)==NULL) (*result)=new CPointMassSet<float>;
    if(launcher!=NULL && launcher->getPotential()->Mass(0)>0)  // massive black hole at origin
    {
        (*result)->add(CPosVelPoint<float>(), static_cast<float>(launcher->getPotential()->Mass(0)));
        numPoints--;
    }
    double outSumWeight=0;  // accumulator for sum of orbit weights up to current orbit (adjusted by refinement factors)
    size_t outPointIndex=0; // accumulator for sum of points emitted up to current orbit
    double outTotalMass=0;  // accumulator for total mass in points up to current orbit
    // controlling center mass position, net momentum and rotation
    double CX=0, CY=0, CZ=0, VX=0, VY=0, VZ=0, LX=0, LY=0, LZ=0, Lhm=0, Chm=0, Vhm=0;
    for(size_t o=0; o<OrbitList.size(); o++)
    {
        outSumWeight += OrbitList[o]->getWeight() / orbitMassFactor[o];
        size_t newPointIndex = static_cast<size_t>(outSumWeight/sumWeight * numPoints);
        size_t pointsPerOrbit = newPointIndex - outPointIndex;   // the number of sampling points necessary for an orbit
        outPointIndex=newPointIndex;
        if(pointsPerOrbit==0) continue;
        double pointMass = orbitMassFactor[o] * sumWeight / numPoints;  // individual mass of points from this orbit
        const CTrajSampleInformation<float>* trajSampleInfo = 
            static_cast<const CTrajSampleInformation<float>*>(OrbitList[o]->getInfoByType(CBasicInformation::IT_TRAJ_SAMPLE));
        // if there is no trajectory sample stored, at least have one point corresponding to initial conditions
        CPosVelDataFloat onePoint(1, OrbitList[o]->getInitCond());
        const CPosVelDataFloat* trajSample = (trajSampleInfo!=NULL && !trajSampleInfo->getTraj().empty() ? &trajSampleInfo->getTraj() : &onePoint); 
        size_t pointsInSample=trajSample->size();
        maxSamplingPoints = std::max<size_t>(maxSamplingPoints, pointsPerOrbit);
        numOrbitsUsed++;
        if(pointsPerOrbit>pointsInSample && trajSampleCreator!=NULL)
        {   // mark this orbit for reintegration with a greater number of sampling points
            numOrbitsReintegrate++;
            OrbitList[o]->setState(COrbitDesc::OS_INITIALIZED);
            // safety measure to enable sampling >1 points from orbits that have not been integrated at all, but are still required in the solution
            if(OrbitList[o]->getIntTime()<=0) OrbitList[o]->setIntTime(1.0);
            // communicate desired number of points and required timestep to the runtime function used in reintegration
            trajSampleCreator->addOrbitSpecial(OrbitList[o]->getInitCond(), pointsPerOrbit+1, 
                OrbitList[o]->getIntTime()/(pointsPerOrbit*10+1));
            nextOrbitInd=0;  // will restart the integration of orbit library from beginning
        }
        for(size_t p=0; p<pointsPerOrbit; p++)
        {
            outTotalMass+=pointMass;
            // loop through entire sample more than once if number of points necessary is greater than sample size
            CPosVelPoint<float> xv = trajSample->at(p % pointsInSample);
            double lx=pointMass*(xv.Pos[1]*xv.Vel[2]-xv.Pos[2]*xv.Vel[1]);
            double ly=pointMass*(xv.Pos[2]*xv.Vel[0]-xv.Pos[0]*xv.Vel[2]); 
            double lz=pointMass*(xv.Pos[0]*xv.Vel[1]-xv.Pos[1]*xv.Vel[0]);
            // decide whether to flip the orbit, and about which plane.
            int nflip=0;
            double absL = fabs(LX+lx)+fabs(LY+ly)+fabs(LZ+lz) + 
                (fabs(CX+pointMass*xv.Pos[0])+fabs(CY+pointMass*xv.Pos[1])+fabs(CZ+pointMass*xv.Pos[2])) *
                (fabs(VX+pointMass*xv.Vel[0])+fabs(VY+pointMass*xv.Vel[1])+fabs(VZ+pointMass*xv.Vel[2])) / NBEXPORT_XV_FACTOR;
            double absLx= fabs(LX+lx)+fabs(LY-ly)+fabs(LZ-lz) + 
                (fabs(CX-pointMass*xv.Pos[0])+fabs(CY+pointMass*xv.Pos[1])+fabs(CZ+pointMass*xv.Pos[2])) *
                (fabs(VX-pointMass*xv.Vel[0])+fabs(VY+pointMass*xv.Vel[1])+fabs(VZ+pointMass*xv.Vel[2])) / NBEXPORT_XV_FACTOR;
            if(absLx<absL) { absL=absLx; nflip=1; }
            double absLy= fabs(LX-lx)+fabs(LY+ly)+fabs(LZ-lz) + 
                (fabs(CX+pointMass*xv.Pos[0])+fabs(CY-pointMass*xv.Pos[1])+fabs(CZ+pointMass*xv.Pos[2])) *
                (fabs(VX+pointMass*xv.Vel[0])+fabs(VY-pointMass*xv.Vel[1])+fabs(VZ+pointMass*xv.Vel[2])) / NBEXPORT_XV_FACTOR;
            if(absLy<absL) { absL=absLy; nflip=2; }
            double absLz= fabs(LX-lx)+fabs(LY-ly)+fabs(LZ+lz) + 
                (fabs(CX+pointMass*xv.Pos[0])+fabs(CY+pointMass*xv.Pos[1])+fabs(CZ-pointMass*xv.Pos[2])) *
                (fabs(VX+pointMass*xv.Vel[0])+fabs(VY+pointMass*xv.Vel[1])+fabs(VZ-pointMass*xv.Vel[2])) / NBEXPORT_XV_FACTOR;
            if(absLz<absL) { absL=absLz; nflip=3; }
            if(0&&nflip>0)  // enable flipping depending on the symmetry of the model -- unused because it doesn't take into account figure rotation
            { 
                xv.Pos[nflip-1] *= -1;
                xv.Vel[nflip-1] *= -1; 
            }
            (*result)->add(xv, static_cast<float>(pointMass));
            // accumulate total momentum and angular momentum
            LX+=pointMass*(xv.Pos[1]*xv.Vel[2]-xv.Pos[2]*xv.Vel[1]);
            LY+=pointMass*(xv.Pos[2]*xv.Vel[0]-xv.Pos[0]*xv.Vel[2]);
            LZ+=pointMass*(xv.Pos[0]*xv.Vel[1]-xv.Pos[1]*xv.Vel[0]);
            CX+=pointMass*xv.Pos[0];
            CY+=pointMass*xv.Pos[1];
            CZ+=pointMass*xv.Pos[2];
            VX+=pointMass*xv.Vel[0];
            VY+=pointMass*xv.Vel[1];
            VZ+=pointMass*xv.Vel[2];
        }
        if(outTotalMass>0.5*totalMass && Lhm==0 && Vhm==0 && Chm==0)
        {   // store half-mass values
            Chm = sqrt(CX*CX+CY*CY+CZ*CZ)/outTotalMass;
            Vhm = sqrt(VX*VX+VY*VY+VZ*VZ)/outTotalMass;
            Lhm = sqrt(LX*LX+LY*LY+LZ*LZ)/outTotalMass;
        }
    }
    CX/=outTotalMass;
    CY/=outTotalMass;
    CZ/=outTotalMass;
    VX/=outTotalMass;
    VY/=outTotalMass;
    VZ/=outTotalMass;
    LX/=outTotalMass;
    LY/=outTotalMass;
    LZ/=outTotalMass;
    for(size_t i=0; i<(*result)->size(); i++)
    {   // cancel total momentum
        (*result)->at(i).first.Vel[0] -= static_cast<float>(VX);
        (*result)->at(i).first.Vel[1] -= static_cast<float>(VY);
        (*result)->at(i).first.Vel[2] -= static_cast<float>(VZ);
    }
    if(numOrbitsReintegrate>0)  // will be handled in core->exportNbody
        return "Reintegrating "+convertToString(numOrbitsReintegrate)+" orbits..";
    // return textual representation of some statistics
    std::string info = convertToString((*result)->size())+" points created\n" +
        convertToString(outTotalMass) + " total mass\n" + 
        convertToString(numOrbitsUsed)+" orbits used\n" +
        "max "+convertToString(maxSamplingPoints)+" sampling points per orbit\n" + 
        "Avg.X="+convertToString(sqrt(CX*CX+CY*CY+CZ*CZ),3) + " (" + convertToString(Chm,3) + 
        ")\nAvg.V="+convertToString(sqrt(VX*VX+VY*VY+VZ*VZ),3) + " (" + convertToString(Vhm,3) + ") [cancelled]" +
        "\nAvg.L="+convertToString(sqrt(LX*LX+LY*LY+LZ*LZ),3) + " (" + convertToString(Lhm,3) + ")";
    return info;
}

template<typename NumT> void CPointMassSetHandler<NumT>::computeVirialRatio(double* _T, double* _W)
{
    state=PS_RUNNING;
    CPotentialNB treePot(0 /*eps*/, 0.5 /*theta*/, *points);
    double T=0, W=0;
//#pragma omp parallel for reduction(+:T,W)  // commented out because it leads to an obscure crash
    for(int ind=0; ind<static_cast<int>(points->size()); ind++)
    {
        const CPosVelPoint<float>& pt=points->at(ind).first;
        double v2 = pow_2(pt.Vel[0])+pow_2(pt.Vel[1])+pow_2(pt.Vel[2]);
        T += points->at(ind).second * v2/2.0;
        W +=-points->at(ind).second * treePot.Phi(pt.Pos[0], pt.Pos[1], pt.Pos[2])/2.;
        if(state==PS_NEEDTOTERMINATE) ind=static_cast<int>(points->size());
    }
    state=PS_DONE;
    *_T=T; *_W=W;
}
template void CPointMassSetHandler<float>::computeVirialRatio(double *T, double *W);  // explicit instantiation

/* orbit evaluating function for mass refinement in exportNbody */
CBinnedMassRefinementFnc::CBinnedMassRefinementFnc(const smile::COrbitLibrary *orbitlib, size_t numBins, const CBasicOrbitFilteringFnc* _rankFnc) :
  rankFnc(_rankFnc)
{
    size_t numOrbits=orbitlib->size();
    if(numOrbits==0 || numBins==0) return;   // not an error, but no ranking will be performed
    vectord orbitRank(numOrbits);
    for(size_t o=0; o<numOrbits; o++)
        orbitRank[o] = rankFnc->eval(orbitlib->getOrbitDesc(o));
    std::sort(orbitRank.begin(), orbitRank.end());
    binBoundary.resize(numBins+1);
    for(size_t bin=0; bin<=numBins; bin++)
        binBoundary[bin] = orbitRank[ static_cast<size_t>((numOrbits-1) / pow(2.0, 1.0*bin)) ];
}

double CBinnedMassRefinementFnc::eval(const smile::COrbitDesc *orbit) const
{
    if(binBoundary.empty() || orbit==NULL) return 1;
    double orbitRank=rankFnc->eval(orbit);
    size_t numBins=binBoundary.size();
    double factor=0;  // shouldn't remain zero, signals an error
    for(size_t bin=0; bin<numBins; bin++)
        if(orbitRank <= binBoundary[bin]) factor=pow(2.0, numBins-1.0-bin);
    return factor;
}

/* an example of orbit rank evaluation based on total energy */
double CEnergyOrbitFilteringFnc::eval(const smile::COrbitDesc *orbit) const
{
    // the method below is slightly modified for the case of figure rotation:
    // because the Jacobi constant is not monotonically increasing when moving outwards in radius, 
    // we take the initial energy of the orbit instead (which is not constant over time, unfortunately)
    const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(orbit->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
    if(info!=NULL) {
        const CPosVelPoint<float>& point=orbit->getInitCond();
        return info->getEinit() + Omega*(point.Pos[0]*point.Vel[1]-point.Pos[1]*point.Vel[0]);  // compensate for figure rotation if Omega!=0
    }
    else return 0; // shouldn't happen because this function is supposed to be called when exporting Nbody model, which obviously should be done after orbit integration
}

// refine particle mass based on the fraction of time it spends inside the given radius
double CRadiusMassRefinementFnc::eval(const smile::COrbitDesc *orbit) const
{
    double frac=0;
    const CTrajSampleInformation<float>* inf=static_cast<const CTrajSampleInformation<float>*>
        (orbit->getInfoByType(CBasicInformation::IT_TRAJ_SAMPLE));
    if(inf!=NULL) {
        const CPosVelDataFloat& trj=inf->getTraj();
        for(size_t j=0; j<trj.size(); j++)
            if(pow_2(trj[j].Pos[0])+pow_2(trj[j].Pos[1])+pow_2(trj[j].Pos[2]) < pow_2(refinementRadius)) frac+=1;
        if(trj.size()>0) frac/=trj.size();
    }
    return 1./(1./massRatio+frac*(1.-1./massRatio));
}

}  // namespace
