#include "schwarzschild.h"
#include <cmath>
#include <algorithm>
#include <gsl/gsl_math.h>
#include <gsl/gsl_integration.h>
#include <cassert>
#include "potential.h"
#include "orbitlib.h"
#include "orbitint.h"
#include "orbitan.h"
#include "massmodelaxi.h"
#include "utils.h"

namespace smile{

//-------------------------------------------------------------//
// class for recording time spent in each shell during orbit integration, 
// and corresponding radial & tangential velocity dispersion
COrbitRuntimeSchwKinemShell::COrbitRuntimeSchwKinemShell(const COrbit* _orbit, const CSchwDataKinemShell* _model): 
CBasicOrbitRuntimeFnc(_orbit), model(_model)
{
    shellTime.assign(model->numConstraints()+1, 0); 
    shellVr.assign(model->numConstraints()+1, 0); 
    shellVt.assign(model->numConstraints()+1, 0); 
    const CPosVelPoint<double>& IC=orbit->getInitCond();
    shellprev = model->whichShellR(sqrt(pow_2(IC.Pos[0])+pow_2(IC.Pos[1])+pow_2(IC.Pos[2])));
    timeAccuracy = 1e-6*orbit->getTimeUnit()*orbit->getConfig()->samplingInterval;
}

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeSchwKinemShell::Timestep(const double told, const double t, CPosVelPoint<double> *vars)
{
    double r2=pow_2(vars->Pos[0]) + pow_2(vars->Pos[1]) + pow_2(vars->Pos[2]);
    double Lx=vars->Pos[1]*vars->Vel[2]-vars->Pos[2]*vars->Vel[1];
    double Ly=vars->Pos[2]*vars->Vel[0]-vars->Pos[0]*vars->Vel[2];
    double Lz=vars->Pos[0]*vars->Vel[1]-vars->Pos[1]*vars->Vel[0];
    double L=sqrt(pow_2(Lx)+pow_2(Ly)+pow_2(Lz));
    double vr2= pow_2(vars->Pos[0]*vars->Vel[0]+vars->Pos[1]*vars->Vel[1]+vars->Pos[2]*vars->Vel[2])/r2;
    double vt2= pow_2(L)/r2;
    if(r2==0) { vr2=pow_2(vars->Vel[0])+pow_2(vars->Vel[1])+pow_2(vars->Vel[2]); vt2=0; }
    size_t shell=model->whichShellR(sqrt(r2));
    if(shell==shellprev)
    {
        shellTime[shell] += (t-told);
        shellVr[shell] += (t-told)*vr2;
        shellVt[shell] += (t-told)*vt2;
    }
    else  // calculate fractional time spent in shell
        calcShellTime(told, t, shellprev, shell, vr2, vt2);
    shellprev=shell;
    return SR_TERMINATE;
}

void COrbitRuntimeSchwKinemShell::calcShellTime(double tlow, double tupp, size_t slow, size_t supp, double vr2, double vt2)
{
    double tl=tlow, tu=tupp;
    int niter=0;
    double tmid;
    do{
        tmid=(tu+tl)/2;
        size_t smid=model->whichShellR(sqrt(
            pow_2(orbit->getPos(0,tmid)) + pow_2(orbit->getPos(1,tmid)) + pow_2(orbit->getPos(2,tmid))));
        if(smid==slow)
            tl=tmid;
        else if(smid==supp)
            tu=tmid;
        else
        {   // neither of the two shells - go deeper into recursion
            calcShellTime(tlow, tmid, slow, smid, vr2, vt2);
            calcShellTime(tmid, tupp, smid, supp, vr2, vt2);
            return;
        }
        niter++;
    }
    while((tu-tl>timeAccuracy) && (niter<20));
    shellTime[slow] += tmid-tlow;
    shellTime[supp] += tupp-tmid;
    shellVr[slow]   += (tmid-tlow)*vr2;
    shellVr[supp]   += (tupp-tmid)*vr2;
    shellVt[slow]   += (tmid-tlow)*vt2;
    shellVt[supp]   += (tupp-tmid)*vt2;
}

void COrbitRuntimeSchwKinemShell::Finish()
{
    double intTime=orbit->getIntTime();
    if(intTime>0)
    {
        for(size_t s=0; s<shellTime.size(); s++)
        {
            if(shellTime[s]>0)
            {
                shellVr[s] /= shellTime[s];
                shellVt[s] /= shellTime[s];
            }
            shellTime[s] /= intTime;    // normalize to unit time
        }
    }
    shellVr.pop_back();  // eliminate last shell which goes to infty
    shellVt.pop_back(); 
    shellTime.pop_back();
}

//-------------------------------------------------------------//
// class for recording contribution of an orbit to the angular momentum distribution at the given energy
COrbitRuntimeSchwAngMomDist::COrbitRuntimeSchwAngMomDist(const COrbit* _orbit, const CSchwDataAngMomDist* _model): 
  CBasicOrbitRuntimeFnc(_orbit), model(_model), timeStep(orbit->getTimeUnit()*orbit->getConfig()->samplingInterval)
{
    binTime.assign(model->getNumBins(), 0); 
    Einit=totalEnergy(orbit->getPotential(), orbit->getInitCond());
    Lcirc2=findLcirc2(orbit->getPotential(), orbit->getInitCond());  // assume that energy does not change with time, ignore the possibility of Omega!=0
    if(Lcirc2<=0) Lcirc2=gsl_posinf();
}

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeSchwAngMomDist::Timestep(const double told, const double t, CPosVelPoint<double>* /*vars*/)
{
    int i1=(int)(told/timeStep); 
    int i2=(int)(t/timeStep);
    for (int i=i1; i<=i2; i++)
    {
        double tt=timeStep*i;
        if ((tt>=told) && (tt<t))
        {
            double X=orbit->getPos(0, tt);
            double Y=orbit->getPos(1, tt);
            double Z=orbit->getPos(2, tt);
            double VX=orbit->getVel(0, tt);
            double VY=orbit->getVel(1, tt);
            double VZ=orbit->getVel(2, tt);
            double Lrel = std::min<double>(0.999, sqrt((pow_2(Y*VZ-Z*VY) + pow_2(Z*VX-X*VZ) + pow_2(X*VY-Y*VX) )/Lcirc2));
            size_t binIndex=static_cast<size_t>(Lrel*binTime.size());
            binTime[binIndex]+=timeStep;
        }
    }
    return SR_TERMINATE;
}

void COrbitRuntimeSchwAngMomDist::Finish()
{
    double intTime=0;
    for(size_t b=0; b<binTime.size(); b++) intTime+=binTime[b];
    if(intTime>0)
        for(size_t b=0; b<binTime.size(); b++) binTime[b]/=intTime;
    binTime.push_back(Einit);
}

//-------------------------------------------------------------//
// class for recording time spent in each cell during orbit integration for classic Schwarzschild modelling
COrbitRuntimeSchwDensGrid::COrbitRuntimeSchwDensGrid(const COrbit* _orbit, const CSchwDataDensGrid* _model): 
CBasicOrbitRuntimeFnc(_orbit), model(_model)
{
    cellTime.assign(model->numConstraints(), 0);
    const CPosVelPoint<double>& IC=orbit->getInitCond();
    cellprev = model->whichCell(IC.Pos[0],IC.Pos[1],IC.Pos[2]);
    timeAccuracy = 1e-6*orbit->getTimeUnit()*orbit->getConfig()->samplingInterval;
}

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeSchwDensGrid::Timestep(const double told, const double t, CPosVelPoint<double> *vars)
{
    size_t cell=model->whichCell(vars->Pos[0], vars->Pos[1], vars->Pos[2]);
    if(cell==cellprev)
        cellTime[cell] += (t-told);
    else  // calculate fractional time spent in cell
        calcCellTime(told, t, cellprev, cell);
    cellprev=cell;
    return SR_TERMINATE;
}

void COrbitRuntimeSchwDensGrid::calcCellTime(double tlow, double tupp, size_t clow, size_t cupp)
{
    double tl=tlow, tu=tupp;
    int niter=0;
    double tmid;
    do{
        tmid=(tu+tl)/2;
        size_t cmid=model->whichCell(orbit->getPos(0,tmid), orbit->getPos(1,tmid), orbit->getPos(2,tmid));
        if(cmid==clow)
            tl=tmid;
        else if(cmid==cupp)
            tu=tmid;
        else
        {   // neither of the two cells - go deeper into recursion
            calcCellTime(tlow, tmid, clow, cmid);
            calcCellTime(tmid, tupp, cmid, cupp);
            return;
        }
        niter++;
    }
    while((tu-tl>timeAccuracy) && (niter<20));
    cellTime[clow]  += tmid-tlow;
    cellTime[cupp]  += tupp-tmid;
}

void COrbitRuntimeSchwDensGrid::Finish()
{
    double intTime=orbit->getIntTime();
    if(intTime>0)
        for(size_t c=0; c<cellTime.size(); c++)
            cellTime[c] /= intTime;    // normalize to unit time
    cellTime.pop_back();  // eliminate last cell which goes to infty
}

//-------------------------------------------------------------//
// class for recording density data in terms of spherical-harmonic expansion
COrbitRuntimeSchwDensSH::COrbitRuntimeSchwDensSH(const COrbit* _orbit, const CSchwDataDensSH* _model): 
CBasicOrbitRuntimeFnc(_orbit), model(_model)
{
    timeStep=orbit->getTimeUnit()*orbit->getConfig()->samplingInterval;
}

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeSchwDensSH::Timestep(const double told, const double t, CPosVelPoint<double>*)
{
    int i1=(int)(told/timeStep); 
    int i2=(int)(t/timeStep);
    for (int i=i1; i<=i2; i++)
    {
        double tt=timeStep*i;
        if ((tt>=told) && (tt<t))
        {
            traj.add(CPosVelPoint<double>(
                orbit->getPos(0, tt),
                orbit->getPos(1, tt),
                orbit->getPos(2, tt),
                0, 0, 0), timeStep);
        }
    }
    return SR_TERMINATE;
}

void COrbitRuntimeSchwDensSH::Finish()
{
    double intTime=orbit->getIntTime();
    if(intTime>0) 
        for(size_t i=0; i<traj.size(); i++)
            traj[i].second /= intTime;
    model->computeDensityData(traj, &SHcoefs);
    traj.data.clear();
}

//-------------------------------------------------------------//
CBasicShellSchwData::CBasicShellSchwData(const CDensity* density, unsigned int numShells, 
    double innerShellMass, double outerShellMass)
{
    totalMass=density->totalMass();
    createRadialGrid(density, numShells, totalMass, innerShellMass, outerShellMass, &shellRadius);
}

size_t CBasicShellSchwData::whichShellR(double radius) const
{
    size_t shnum=0;
    while(shnum<shellRadius.size() && radius>=shellRadius[shnum]) shnum++;
    return shnum;
}

//-------------------------------------------------------------//
// Schw.data class for handling kinematic information in terms of radial/tangential 
// velocity dispersions in a set of concentric spherical shells
CSchwDataKinemShell::CSchwDataKinemShell(const CDensity* density, unsigned int numShells, 
    double innerShellMass, double outerShellMass, double betaIn, double betaOut) :
  CBasicShellSchwData(density, numShells, innerShellMass, outerShellMass)
{
    beta.resize(numShells);
    for(size_t i=0; i<numShells; i++)
        beta[i]=betaIn + (betaOut-betaIn) * density->Mass(getShellRadius(i)) / totalMass;
}

double CSchwDataKinemShell::getOrbitConstraint(size_t index, const CSchwInformation* info) const
{
    assert(info->schwDataType()==dataType());
    sdNumType shellVt = info->getSchwData(index+2*numConstraints());
    sdNumType shellVr = info->getSchwData(index+1*numConstraints());
    sdNumType shellTime = info->getSchwData(index);
    return (shellVt-2*(1-beta[index])*shellVr) * shellTime;
}

bool CSchwDataKinemShell::isSchwInformationCorrect(const CSchwInformation* info) const
{
    return (info->schwDataType()==dataType() && info->getSchwData().size()==3*numConstraints());
}

//-------------------------------------------------------------//
// Schw.data class for handling angular momentum distribution as a function of energy
CSchwDataAngMomDist::CSchwDataAngMomDist(const CDensity* density, unsigned int numShells, 
    double innerShellMass, double outerShellMass, double betaIn, double betaOut, 
    const CPotential* potential, unsigned int _numBins) :
  CBasicShellSchwData(density, numShells, innerShellMass, outerShellMass), numBins(_numBins)
{
    dataAngularMomentum.resize(numShells*numBins);
    for(size_t s=0; s<numShells; s++) {
        // required value of anisotropy in a given (energy) shell
        double beta=betaIn + (betaOut-betaIn) * density->Mass(getShellRadius(s)) / totalMass;
        for(size_t b=0; b<numBins; b++) {
            dataAngularMomentum[s*numBins+b] = (pow((b+1.0)/numBins, 2*(1-beta))-pow(b*1.0/numBins, 2*(1-beta)));
        }
    }
    shellEnergy.assign(numShells, 0);
    for(unsigned int s=0; s<numShells; s++)
        shellEnergy[s] = potential->Phi(shellRadius[s], 0, 0);
}

double CSchwDataAngMomDist::getOrbitConstraint(size_t index, const CSchwInformation* info) const
{
    assert(info->schwDataType()==dataType());
    double energy = info->getSchwData(numBins);
    size_t shell = whichShellE(energy);
    if(index>=shell*numBins && index<(shell+1)*numBins)
        return info->getSchwData(index-shell*numBins) - dataAngularMomentum[index];
    else
        return 0;
}

size_t CSchwDataAngMomDist::whichShellE(double energy) const
{
    size_t shnum=0;
    while(shnum<shellRadius.size() && energy>=shellEnergy[shnum]) shnum++;
    return shnum;
}

bool CSchwDataAngMomDist::isSchwInformationCorrect(const CSchwInformation* info) const
{
    return (info->schwDataType()==dataType() && info->getSchwData().size()==numBins+1);
}

//-------------------------------------------------------------//
// density data objects

CSchwDataBasicDens::CSchwDataBasicDens(const CDensity* density) :
  totalMass(density->totalMass()) {}

double CSchwDataBasicDens::getOrbitConstraint(size_t index, const CSchwInformation* info) const
{
    assert(info->schwDataType()==dataType());
    return (index==numConstraints()-1 ? 1.0 : info->getSchwData(index)) / getConstraintNormFactor(index);  // last constraint is for total mass
}

bool CSchwDataBasicDens::isSchwInformationCorrect(const CSchwInformation* info) const
{
    return (info->schwDataType()==dataType() && info->getSchwData().size()+1==numConstraints());
}

//-------------------------------------------------------------//
// Class for performing classic 3d-grid-based Schwarzschild density modelling

CSchwDataDensGridClassic::CSchwDataDensGridClassic(const CDensity* density, unsigned int numShells, 
    double innerShellMass, double outerShellMass, unsigned int _linesPerSegment, double _grid_q, double _grid_p): 
  CSchwDataDensGrid(density), 
  linesPerSegment(_linesPerSegment), 
  cellsPerSegment(linesPerSegment*linesPerSegment),
  cellsPerShell(3*cellsPerSegment),
  numCells(numShells*cellsPerShell), 
  grid_x(pow(_grid_p*_grid_q, -1.0/3)), grid_y(_grid_q*grid_x), grid_z(_grid_p*grid_x)
{
    // init angular grid boundaries
    cellBoundary.resize(linesPerSegment+1);
    for(size_t i=0; i<linesPerSegment; i++)
        cellBoundary[i] = tan(M_PI/4/linesPerSegment*i);
    cellBoundary[linesPerSegment] = 1.0;  // exact value
    createRadialGrid(density, numShells, totalMass, innerShellMass, outerShellMass, &shellRadius);
    double massShell=0, prevMassShell=0;
    cellMassNorm.assign(numCells+1, 1.0);
    for(unsigned int s=0; s<numShells; s++) {
        massShell = density->Mass(shellRadius[s]);
        // massShell<=prevMassShell shouldn't normally happen, as it indicates a badly behaved potential; 
        // however, to avoid division by zero we initialize it by some reasonable value
        double shellMassNorm=(massShell>prevMassShell ? massShell-prevMassShell : 0);
        for(unsigned int c=0; c<cellsPerShell; c++)
            cellMassNorm[s*cellsPerShell+c]=shellMassNorm/cellsPerShell;
        prevMassShell=massShell;
    }
    // calculate mass in cells
    cellMass.assign(numCells+1, 0);
    cellMass.back()=totalMass;
    cellMassNorm.back()=totalMass/numCells;
    if(density->PotentialType()==CDensity::PT_NB) {
        // frozen Nbody potential - simply count bodies in each cell
        const CPotentialNB* potentialNB=static_cast<const CPotentialNB*>(density);
        for(size_t ind=0; ind<potentialNB->bodyNum(); ind++)
        {
            CPosPoint<double> pt=potentialNB->bodyPos<double>(ind);
            size_t cell=whichCell(pt.Pos[0], pt.Pos[1], pt.Pos[2]);
            double ptmass=potentialNB->bodyMass(ind);
            if(cell<numCells) cellMass[cell] += ptmass;
        }
        return;
    }
#pragma omp parallel for schedule(dynamic,16)
    for(int c=0; c<static_cast<int>(numCells); c++)
        cellMass[c] = initCellMass(c, density);
}

/// \cond INTERNAL_DOCS
struct CCellMassGridClassicParam {
    const CDensity* density;
    double Xm, Ym, Zm;
};
double intCellMass_r(double r, void* param)
{
    return ((CCellMassGridClassicParam*)param)->density->Rho(
        r*((CCellMassGridClassicParam*)param)->Xm, 
        r*((CCellMassGridClassicParam*)param)->Ym, 
        r*((CCellMassGridClassicParam*)param)->Zm) * r*r;
};
/// \endcond

double CSchwDataDensGridClassic::initCellMass(const size_t cell, const CDensity* density) const
{
    size_t nshell = cell / cellsPerShell;
    if(nshell>=shellRadius.size()) 
        return 0;
    size_t indseg = cell % cellsPerShell;
    size_t nseg = indseg / cellsPerSegment;
    size_t ind1 =(indseg % cellsPerSegment) % linesPerSegment;
    size_t ind2 =(indseg % cellsPerSegment) / linesPerSegment;
    double dr1 = cellBoundary[ind1+1] - cellBoundary[ind1];
    double dr2 = cellBoundary[ind2+1] - cellBoundary[ind2];
    double mass=0;
    gsl_function F;
    CCellMassGridClassicParam param;
    param.density=density;
    F.function=&intCellMass_r;
    F.params=&param;
    double result, error;
    size_t neval;
    for(size_t i1=0; i1<NUM_SUBDIV_ANGULAR; i1++)
        for(size_t i2=0; i2<NUM_SUBDIV_ANGULAR; i2++)
        {
            double r1l = cellBoundary[ind1] + dr1*i1/NUM_SUBDIV_ANGULAR;
            double r1h = cellBoundary[ind1] + dr1*(i1+1)/NUM_SUBDIV_ANGULAR;
            double r2l = cellBoundary[ind2] + dr2*i2/NUM_SUBDIV_ANGULAR;
            double r2h = cellBoundary[ind2] + dr2*(i2+1)/NUM_SUBDIV_ANGULAR;
            double phill = asin(r1l*r2l/sqrt((1+pow_2(r1l))*(1+pow_2(r2l))));
            double philh = asin(r1l*r2h/sqrt((1+pow_2(r1l))*(1+pow_2(r2h))));
            double phihl = asin(r1h*r2l/sqrt((1+pow_2(r1h))*(1+pow_2(r2l))));
            double phihh = asin(r1h*r2h/sqrt((1+pow_2(r1h))*(1+pow_2(r2h))));
            double area = (phill+phihh-philh-phihl);
            double ratio1 = (r1l+r1h)/2;
            double ratio2 = (r2l+r2h)/2;
            double coord0 = 1/sqrt(1 + pow_2(ratio1) + pow_2(ratio2));
            double coord1 = ratio1*coord0;
            double coord2 = ratio2*coord0;
            switch(nseg) {
            case 0:
                param.Xm=coord0;
                param.Ym=coord1;
                param.Zm=coord2;
                break;
            case 1:
                param.Ym=coord0;
                param.Zm=coord1;
                param.Xm=coord2;
                break;
            case 2:
                param.Zm=coord0;
                param.Xm=coord1;
                param.Ym=coord2;
                break;
            }
            param.Xm *= grid_x;
            param.Ym *= grid_y;
            param.Zm *= grid_z;
            gsl_integration_qng(&F, nshell==0 ? 0 : shellRadius[nshell-1], shellRadius[nshell], 0, EPSREL_MASS_INT, &result, &error, &neval);
            mass += result * area * 8;
        }
    return mass;
}

void CSchwDataDensGridClassic::cellCenter(const size_t C, double* X, double* Y, double* Z) const
{
    assert(X); assert(Y); assert(Z);
    size_t nshell = C / cellsPerShell;
    if(nshell>=shellRadius.size()) 
    {
        X=0; Y=0; Z=0;
        return;
    }
    size_t indseg = C % cellsPerShell;
    size_t nseg = indseg / cellsPerSegment;
    size_t ind1 =(indseg % cellsPerSegment) % linesPerSegment;
    size_t ind2 =(indseg % cellsPerSegment) / linesPerSegment;
    double ratio1 = (cellBoundary[ind1] + cellBoundary[ind1+1])/2;
    double ratio2 = (cellBoundary[ind2] + cellBoundary[ind2+1])/2;
    double r = ((nshell>0 ? shellRadius[nshell-1] : 0) + shellRadius[nshell])/2;
    double coord0 = r/sqrt(1 + pow_2(ratio1) + pow_2(ratio2));
    double coord1 = ratio1*coord0;
    double coord2 = ratio2*coord0;
    switch(nseg)
    {
    case 0:
        *X=coord0;
        *Y=coord1;
        *Z=coord2;
        break;
    case 1:
        *Y=coord0;
        *Z=coord1;
        *X=coord2;
        break;
    case 2:
        *Z=coord0;
        *X=coord1;
        *Y=coord2;
        break;
    }
    *X *= grid_x;
    *Y *= grid_y;
    *Z *= grid_z;
    if(whichCell(*X,*Y,*Z) != C)
        my_message(FUNCNAME, "Error in determining cell center, shouldn't happen!");
}

size_t CSchwDataDensGridClassic::whichCell(const double _X, const double _Y, const double _Z) const
{
    double X=fabs(_X/grid_x), Y=fabs(_Y/grid_y), Z=fabs(_Z/grid_z), radius=sqrt(pow_2(X)+pow_2(Y)+pow_2(Z));
    int nseg=-1;
    size_t nshell=0;
    while(nshell<shellRadius.size() && radius>=shellRadius[nshell]) nshell++;
    double coord0=0, coord1=0, coord2=0;
    if((X>=Y)&&(X>Z))
    {
        nseg=0;
        coord0=X;
        coord1=Y;
        coord2=Z;
    } else
    if((Y>=Z)&&(Y>X))
    {
        nseg=1;
        coord0=Y;
        coord1=Z;
        coord2=X;
    } else
    if((Z>=X)&&(Z>=Y))
    {
        nseg=2;
        coord0=Z;
        coord1=X;
        coord2=Y;
    }
    if(nseg==-1)
    {
        my_message(FUNCNAME, "Error in determining cell index, shouldn't happen!");
        return 0;  // should not happen, but who knows?...
    }
    double ratio1 = coord1/coord0;
    double ratio2 = coord2/coord0;
    size_t ind1=0, ind2=0;
    for(size_t i=1; i<linesPerSegment; i++)
    {
        if(ratio1>=cellBoundary[i]) ind1=i;
        if(ratio2>=cellBoundary[i]) ind2=i;
    }
    size_t indseg = ind1 + ind2*linesPerSegment + nseg*cellsPerSegment;
    size_t ncell = nshell*cellsPerShell + indseg;
    return (ncell<numCells)?ncell:numCells;
}

//-------------------------------------------------------------//
CSchwDataDensGridCylindrical::CSchwDataDensGridCylindrical(const CDensity* density, size_t _numR, size_t _numz, size_t _numphi) :
  CSchwDataDensGrid(density), numphi(_numphi), numCells(_numR*_numz*_numphi)
{
    // create axisymmetric mass model to facilitate grid initialization
    const CMassModelAxisymmetric* model=new CMassModelAxisymmetric(density, 100, 100);
    // create uniformly sliced grid in cylindrical radius
    double innerShellMass=1.0/(_numR+1);
    double outerShellMass=1-1.0/(2*_numR+1);
    totalMass=model->getTotalMass();
    if(totalMass<=0) { 
        my_error(FUNCNAME, "Error, density model does not seem to have finite mass"); 
        totalMass=1; 
    }
    createNonuniformGrid(_numR, innerShellMass, outerShellMass, false, &gridR);  // grid in Mass(R)
    for(size_t i=0; i<_numR; i++) {
        double val=model->find_R_mass(gridR[i]*totalMass);
        if(val<=0) {  // too bad, can't find the root
            if(i==0) val=density->getRadiusByMass(totalMass*innerShellMass);
            else {  // arbitrary shift grid inward
                val=gridR[i-1];
                gridR[i-1]=(gridR[i-1]+(i>1?gridR[i-2]:0))/2;
            }
        }
        gridR[i]=val;
    }
    // create uniformly sliced grid in cylindrical radius
    innerShellMass=1.0/(_numz+1);
    outerShellMass=1-1.0/(2*_numz+1);
    createNonuniformGrid(_numz, innerShellMass, outerShellMass, false, &gridz);  // grid in Mass(z)
    for(size_t i=0; i<_numz; i++) {
        double val=model->find_z_mass(gridz[i]*totalMass);
        if(val<=0) {  // too bad, can't find the root
            if(i==0) val=density->getRadiusByMass(totalMass*innerShellMass);
            else {  // arbitrary shift grid inward
                val=gridz[i-1];
                gridz[i-1]=(gridz[i-1]+(i>1?gridz[i-2]:0))/2;
            }
        }
        gridz[i]=val;
    }
#ifdef DEBUGPRINT
    my_message(FUNCNAME, 
        "Axisym.grid: R="+convertToString(gridR.front())+"-"+convertToString(gridR.back())+
        ", Z="+convertToString(gridz.front())+"-"+convertToString(gridz.back()));
#endif
    delete model;
    cellMass.assign(numCells+1, 0);
    cellMassNorm.assign(numCells+1, totalMass/numCells);  // if the density is separable in R,z coordinates, then all cells should have the same prior on mass
    cellMass.back()=totalMass;  // last constraint is for total mass
    // compute mass in each cell
#pragma omp parallel for schedule(dynamic,16)
    for(int cell=0; cell<static_cast<int>(numCells); cell++)
        cellMass[cell]=initCellMass(cell, density);
}

/// \cond INTERNAL_DOCS
struct CCellMassCylindricalParam {
    const CDensity* density;
    double R, z;
    double phi_min, phi_max, z_min, z_max;
};
double intCellMass_phi(double phi, void* param)
{
    double R=((CCellMassCylindricalParam*)param)->R;
    double z=((CCellMassCylindricalParam*)param)->z;
    return ((CCellMassCylindricalParam*)param)->density->Rho(R*cos(phi), R*sin(phi), z);
};
double intCellMass_z(double z, void* param)
{
    if( (((CCellMassCylindricalParam*)param)->density->symmetry() & CDensity::ST_AXISYMMETRIC) == CDensity::ST_AXISYMMETRIC)
        return ((CCellMassCylindricalParam*)param)->density->Rho(((CCellMassCylindricalParam*)param)->R, 0, z) *
            ( ((CCellMassCylindricalParam*)param)->phi_max-((CCellMassCylindricalParam*)param)->phi_min );
    else {
        ((CCellMassCylindricalParam*)param)->z=z;
        gsl_function F;
        F.function=&intCellMass_phi;
        F.params=param;
        double result, error;
        size_t neval;
        gsl_integration_qng(&F, ((CCellMassCylindricalParam*)param)->phi_min, ((CCellMassCylindricalParam*)param)->phi_max, 
            0, EPSREL_MASS_INT, &result, &error, &neval);
        return result;
    }
};
double intCellMass_R(double R, void* param)
{
    ((CCellMassCylindricalParam*)param)->R=R;
    gsl_function F;
    F.function=&intCellMass_z;
    F.params=param;
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, ((CCellMassCylindricalParam*)param)->z_min, ((CCellMassCylindricalParam*)param)->z_max, 
        0, EPSREL_MASS_INT, &result, &error, &neval);
    return result*R;
};
/// \endcond

double CSchwDataDensGridCylindrical::initCellMass(const size_t cell, const CDensity* density) const
{
    size_t indR, indz, indphi;
    if(!getCoefDecomposition(cell, &indR, &indz, &indphi))
        return 0;
    CCellMassCylindricalParam param;
    gsl_function F;
    F.function=&intCellMass_R;
    F.params=&param;
    param.density=density;
    param.z_min=indz==0 ? 0 : gridz[indz-1];
    param.z_max=gridz[indz];
    double result, error;
    size_t neval;
    param.phi_min=indphi*M_PI_2/numphi;
    param.phi_max=(indphi+1)*M_PI_2/numphi;
    gsl_integration_qng(&F, indR==0?0:gridR[indR-1], gridR[indR], 0, EPSREL_MASS_INT, &result, &error, &neval);
    return result*8;
}

size_t CSchwDataDensGridCylindrical::whichCell(const double _X, const double _Y, const double _Z) const
{
    double X=fabs(_X), Y=fabs(_Y), Z=fabs(_Z);
    double R=sqrt(pow_2(X)+pow_2(Y));
    double phi = numphi>1 ? atan2(Y, X) : 0;
    size_t indR=0;
    while(indR<gridR.size() && R>gridR[indR]) indR++;
    size_t indz=0;
    while(indz<gridz.size() && Z>gridz[indz]) indz++;
    if(indR>=gridR.size() || indz>=gridz.size())
        return numCells;
    size_t indphi=static_cast<size_t>(phi/M_PI_2 * numphi);
    if(indphi>=numphi) indphi=numphi-1;
    return indR*numphi*gridz.size() + indz*numphi + indphi;
}

bool CSchwDataDensGridCylindrical::getCoefDecomposition(const size_t C, size_t* ind_R, size_t* ind_z, size_t* ind_phi) const
{
    assert(ind_R); assert(ind_z); assert(ind_phi);
    *ind_R=C/(numphi*gridz.size());
    if(*ind_R>=gridR.size()) {
        *ind_phi=*ind_z=0;
        return false;
    }
    *ind_z=(C-(*ind_R)*numphi*gridz.size())/numphi;
    *ind_phi=C%numphi;
    return true;
}

void CSchwDataDensGridCylindrical::cellCenter(const size_t C, double* X, double* Y, double* Z) const
{
    assert(X); assert(Y); assert(Z);
    size_t indR, indz, indphi;
    if(!getCoefDecomposition(C, &indR, &indz, &indphi)) {
        *X = *Y = *Z = -1;
        return;
    }
    double phi=(indphi+0.5)*M_PI_2/numphi;
    *Z = (gridz[indz]+ (indz>0 ? gridz[indz-1] : 0))/2;
    double R = (gridR[indR]+ (indR>0 ? gridR[indR-1] : 0))/2;
    *X = R*cos(phi);
    *Y = R*sin(phi);
    if(whichCell(*X,*Y,*Z) != C)
        my_message(FUNCNAME, "Error in determining cell center, shouldn't happen!");
}

//-------------------------------------------------------------//
CSchwDataDensSHMesh::CSchwDataDensSHMesh(const CDensity* density, size_t numShells, double innerShellMass, double outerShellMass, size_t _Nang) :
  CSchwDataDensSH(density, _Nang), numMesh(numShells)
{
    std::vector< vectord > tmpCoefs;
    createRadialGrid(density, numShells, totalMass, innerShellMass, outerShellMass, &meshRadii);
    meshRadii.insert(meshRadii.begin(), 0);  // additional grid node inserted at origin
    const CDensity* dens = density;
    if(density->PotentialType()==CPotential::PT_COMPOSITE && static_cast<const CPotentialComposite*>(density)->size()==1)
        dens=static_cast<const CPotentialComposite*>(density)->component(0);
    if(dens->PotentialType()==CPotential::PT_SPLINE)   // hack to speed up initialization
        static_cast<const CPotentialSpline*>(dens)->getCoefs(&meshRadii, &tmpCoefs, false);
    else
    {
        CPotentialSpline pot(numMesh, Nang, dens, &meshRadii);
        pot.getCoefs(&meshRadii, &tmpCoefs, true);
    }
    assert(meshRadii.size()==numMesh+1 && tmpCoefs.size()==numMesh+1);
    // take only even l,m>=0 coefficients
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    coefConstraints.assign(1 + numMesh * nAngCoefs + 1, 0);  // one coef for 0th node, nAngCoefs coefs for each of other nodes, and the last one for total mass
    coefConstraints[0] = tmpCoefs[0][0];
    for(size_t radInd=0; radInd<numMesh; radInd++)
        for(size_t l=0; l<=Nang/2; l++)
            for(size_t m=0; m<=l; m++)
            {
                size_t tmpInd=2*l*(2*l+1)+2*m;
                coefConstraints[1 + radInd*nAngCoefs + l*(l+1)/2 + m] = 
                    tmpInd<tmpCoefs[radInd+1].size() ? tmpCoefs[radInd+1][tmpInd] : 0;  // may happen because of hack above, i.e. tmpcoefs array may have less items that required by the model
            }
    coefConstraints.back()=totalMass;
}

void CSchwDataDensSHMesh::computeDensityData(const smile::CPointMassSet<double> &traj, vectord* data) const
{
    assert(data);
    std::vector< vectord > tmpCoefs;
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    data->assign(1 + numMesh * nAngCoefs, 0);
    if(traj.size()<=1) return;  // zerozero
    CPotentialSpline pot(Nang, traj, &meshRadii, &tmpCoefs);
    if(tmpCoefs.size()==0)  // signals error
        return;   // return array of the necessary length filled with zeroes
    assert(tmpCoefs.size()==numMesh+1);
    // take only even l,m>=0 coefficients
    data->at(0) = tmpCoefs[0][0];
    for(size_t radInd=0; radInd<numMesh; radInd++)
        for(size_t l=0; l<=Nang/2; l++)
            for(size_t m=0; m<=l; m++)
            {
                data->at(1 + radInd*nAngCoefs + l*(l+1)/2 + m) = tmpCoefs[radInd+1][2*l*(2*l+1)+2*m];
            }
}

double CSchwDataDensSHMesh::getConstraintNormFactor(size_t index) const
{
    if(index==coefConstraints.size()-1) 
        return totalMass/(coefConstraints.size()-1);  // weight factor for total mass
    if(index==0) return 
        coefConstraints[0]/(coefConstraints.size()-1);
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    size_t n=(index-1)/nAngCoefs;
    size_t lm=(index-1)%nAngCoefs;
    int l=static_cast<int>(sqrt(2.*lm-floor(sqrt(2.*lm))));
    double norm=0;
    for(int m=0; m<=l; m++)
        norm+=pow_2(coefConstraints[1 + n*nAngCoefs + l*(l+1)/2 + m]);
    if(norm>1e-20) 
        return sqrt(norm) * pow_2(l+1.)/(coefConstraints.size()-1);   // use mean-square coef value for all m and given l
    else if(fabs(coefConstraints[1+n*nAngCoefs])>1e-10) 
        return fabs(coefConstraints[1+n*nAngCoefs])/(coefConstraints.size()-1);  // use m=0 coef
    else return 1;  // shouldn't arrive here?
}

void CSchwDataDensSHMesh::getCoefDecomposition(size_t index, double *n, int *l, int *m) const
{
    if(index==0 || index>=numConstraints()) { *l=0; *m=0; *n=0; return; }
    index--;
    size_t indRad= index/getNumCoefsAngularExpansion();
    size_t indAng= index%getNumCoefsAngularExpansion();
    *l=static_cast<int>(sqrt(2.*indAng-floor(sqrt(2.*indAng))));
    *m=static_cast<int>(indAng-(*l)*(*l+1)/2)*2;
    *l*=2;
    *n=indRad+1<meshRadii.size() ? meshRadii[indRad+1] : -1;
}

//-------------------------------------------------------------//
CSchwDataDensSHBSE::CSchwDataDensSHBSE(const CDensity* density, size_t _Nang, size_t _Nrad, double _Alpha) :
  CSchwDataDensSH(density, _Nang), Nrad(_Nrad)
{
    std::vector< vectord > tmpCoefs;

    const CDensity* dens = density;
    if(density->PotentialType()==CPotential::PT_COMPOSITE && static_cast<const CPotentialComposite*>(density)->size()==1)
        dens=static_cast<const CPotentialComposite*>(density)->component(0);
    if(dens->PotentialType()==CPotential::PT_BSE && 
        (_Alpha==0 || static_cast<const CPotentialBSE*>(dens)->getAlpha()==_Alpha))
    {  // HACK: to speed up initialization, assign coefficients from existing density model instead of computing costly integrals
        static_cast<const CPotentialBSE*>(dens)->getCoefs(&tmpCoefs);
        Alpha=static_cast<const CPotentialBSE*>(dens)->getAlpha();
        // the number of coefficients in potential may be greater or less than needed in S.M.
    } else {
        CPotentialBSE pot(_Alpha, Nrad, Nang, dens);
        Alpha=pot.getAlpha();   // if _Alpha==0, it means autodetect; here we get this autodetected value (or the original supplied value if it was >=0.5)
        pot.getCoefs(&tmpCoefs);
        assert(tmpCoefs.size()==Nrad+1);
    }
    // take only even l,m>=0 coefficients
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    coefConstraints.assign((Nrad+1) * nAngCoefs + 1, 0);  // Nang*(Nang+1)/2 coefs for each radial basis function, plus one for total mass
    for(size_t radInd=0; radInd<=Nrad; radInd++)
        for(size_t l=0; l<=Nang/2; l++)
            for(size_t m=0; m<=l; m++)
            {
                size_t tmpInd=2*l*(2*l+1)+2*m;
                coefConstraints[radInd*nAngCoefs + l*(l+1)/2 + m] = 
                    radInd<tmpCoefs.size() && tmpInd<tmpCoefs[radInd].size() ? tmpCoefs[radInd][tmpInd] : 0;  // may happen because of hack above, i.e. tmpcoefs array may have less items that required by the model
            }
    coefConstraints.back()=totalMass;
}

void CSchwDataDensSHBSE::computeDensityData(const smile::CPointMassSet<double> &traj, vectord *data) const
{
    assert(data);
    std::vector< vectord > tmpCoefs;
    size_t nAngCoefs = getNumCoefsAngularExpansion();
    data->assign((Nrad+1) * nAngCoefs, 0);
    if(traj.size()<=1) return;  // zerozero
    CPotentialBSE pot(Alpha, Nrad, Nang, traj, CPotential::ST_TRIAXIAL, false);
    pot.getCoefs(&tmpCoefs);
    assert(tmpCoefs.size()==Nrad+1);
    // take only even l,m>=0 coefficients
    for(size_t radInd=0; radInd<=Nrad; radInd++)
        for(size_t l=0; l<=Nang/2; l++)
            for(size_t m=0; m<=l; m++)
            {
                data->at(radInd*nAngCoefs + l*(l+1)/2 + m) = tmpCoefs[radInd][2*l*(2*l+1)+2*m];
            }
}

/** this function is necessary because the values of coefficients may span many orders of magnitude,
    especially for high alpha. 
    To ensure a more uniform scaling of penalties for constraint violation, the scaling function is introduced. 
    The expression below is based on a number of `test cases' and is very approximate (within a few orders of mag.)
*/
double CSchwDataDensSHBSE::getConstraintNormFactor(size_t index) const
{
    if(index==coefConstraints.size()-1) 
        return totalMass / (coefConstraints.size()-1);  // weight factor for total mass
    size_t n=index/getNumCoefsAngularExpansion();
    size_t lm=index%getNumCoefsAngularExpansion();
    int l=static_cast<int>(sqrt(2.*lm-floor(sqrt(2.*lm))));
#if 0
    double norm=0;
    for(int m=0; m<=l; m++)
        norm+=pow_2(densityData[n*Nang*(Nang+1)/2 + l*(l+1)/2 + m]);
    if(norm>1e-20) return sqrt(norm);
#endif
    // some arbitrary, not very good heuristical approximation. A better one is desirable.
    double coef_n0=pow(n/(Alpha+1)+1, -4-2*log(Alpha)/log(2.)) * pow_2(n+1.);
    if(l==0) return coef_n0/(coefConstraints.size()-1);
    double A=2.8*Alpha-0.6;
    double B=-0.1*Alpha-0.024;
    double C=-0.01*Alpha-0.05;
    double D=-0.11*Alpha*Alpha+0.07*Alpha-1.9;
    return coef_n0*exp(A*l + B*n + C*l*n + D) * pow_2(l+1.);
}

void CSchwDataDensSHBSE::getCoefDecomposition(size_t index, double *n, int *l, int *m) const
{
    if(index>=numConstraints()) { *l=0; *m=0; *n=0; return; }
    size_t indRad= index/getNumCoefsAngularExpansion();
    size_t indAng= index%getNumCoefsAngularExpansion();
    *l=static_cast<int>(sqrt(2.*indAng-floor(sqrt(2.*indAng))));
    *m=static_cast<int>(indAng-(*l)*(*l+1)/2)*2;
    *l*=2;
    *n=static_cast<double>(indRad);
}

//-------------------------------------------------------------//
// driver class for quadratic Schwarzschild modelling of density and shell-kinematic data 
// (the only variant implemented so far)
CSchwModelQuadOpt::CSchwModelQuadOpt(COrbitLibrary* _orbits, 
    const vectorSchwData &_data, 
    const CBasicQuadraticOptimizationSolver<smNumType>* _solver,
    const CBasicOrbitFilteringFnc* _orbitPenaltyFnc,
    double _lambda, double _penaltyConsLin, double _penaltyConsQuad, 
    double _maxWeight) :
  CBasicSchwModel(_orbits, _data), 
  solver(_solver), orbitPenaltyFnc(_orbitPenaltyFnc),
  lambda(_lambda), penaltyConsLin(_penaltyConsLin), penaltyConsQuad(_penaltyConsQuad), maxWeight(_maxWeight), 
  numConstr(0), correct(true)
{
    assert(orbits);
    // count total number of constraints
    for(vectorSchwData::iterator iter=data.begin(); iter!=data.end(); ++iter)
        numConstr += (*iter)->numConstraints();
    // should check that all orbits have necessary kind of data with right array sizes...
    for(size_t ind_orb=0; correct && ind_orb<orbits->size(); ind_orb++)
    {
        std::vector<bool> orbHasData(data.size(), false);   // record what kinds of orbit information we have
        for(size_t ind_inf=0; ind_inf<orbits->getOrbitDesc(ind_orb)->getInfoNum(); ind_inf++)
        {
            const CBasicInformation* info=orbits->getOrbitDesc(ind_orb)->getInfo(ind_inf);
            if(info->infoType()==CBasicInformation::IT_SCHW) {
                const CSchwInformation* infoSchw=static_cast<const CSchwInformation* >(info);
                CBasicSchwData::SCHWDATATYPE dtype=infoSchw->schwDataType();
                // find a corresponding schw.data type from what we have in the model
                for(size_t ind_data=0; ind_data<data.size(); ind_data++)
                    if(data[ind_data]->dataType() == dtype) {
                        if(data[ind_data]->isSchwInformationCorrect(infoSchw))
                            orbHasData[ind_data]=true;
                        else {
                            my_error(FUNCNAME,
                                "Orbit #"+convertToString(ind_orb)+" isn't compatible with Schwarzschild data "+
                                data[ind_data]->dataName()+" (size="+convertToString(data[ind_data]->numConstraints())+")");
                            correct=false;
                        }
                    }
            }
        }
        // check that orbit has all necessary data
        for(size_t ind_data=0; correct && ind_data<data.size(); ind_data++)
            if(!orbHasData[ind_data]) {
                my_error(FUNCNAME,
                    "Orbit library isn't compatible with Schwarzschild data: "
                    "orbit #"+convertToString(ind_orb)+" doesn't have "+data[ind_data]->dataName()+" data");
                correct=false;
            }
    }
    // store the total number of constraints in all data objects, 
    // without the total mass constraint or additional tolerance constraints in the case penaltyCons<0
    numConstrData=numConstr;
    // if have constraint penalties, then the linear system has additional `slack' variables, two per constraint 
    numVar = orbits->size() + (penaltyConsLin!=0 || penaltyConsQuad!=0 ? 2*numConstrData : 0);
    // if constraints must be satisfied to within a given tolerance, then the linear system has 
    // additional block of equations, one per `normal' constraint
    if(penaltyConsLin<0) numConstr*=2;
    // initialize rhs
    rhs.assign(numConstr, 0);        // the rhs of equations
    size_t ind_cons=0;
    for(vectorSchwData::iterator iter=data.begin(); iter!=data.end(); iter++)
    {
        for(size_t c=0; c<(*iter)->numConstraints(); c++, ind_cons++)
            rhs[ind_cons] = static_cast<smNumType>((*iter)->getConstraint(c));
    }
    if(penaltyConsLin<0)  // add the block of equations for constraint tolerance intervals
        for(size_t c=0; c<ind_cons; c++)
            rhs[ind_cons+c] = static_cast<smNumType>(fabs(penaltyConsLin*rhs[c]));
}

smNumType CSchwModelQuadOpt::getLinearMatrixElem(size_t ind_orb, size_t ind_cons) const
{
    if(!correct || ind_cons>=numConstr || ind_orb>=numVar)
        return 0;   // shouldn't happen?
    if(penaltyConsLin!=0 && ind_orb>=orbits->size()) {  // return coef for additional slack variable
        ind_orb -= orbits->size();
        if(penaltyConsLin<0 && ind_cons>=numConstrData) {  // additional block of equations
            ind_cons -= numConstrData;
            if(ind_orb==ind_cons || ind_orb==ind_cons+numConstrData) return +1.0; else return 0;
        }
        if(ind_orb==ind_cons) return +1.0; else
        if(ind_orb==ind_cons+numConstrData) return -1.0; else return 0;
    } else {  // find which Schw.data does the constraint correspond to
        vectorSchwData::const_iterator iter=data.begin();   // the corresponding data object
        size_t ind_cons_this=ind_cons;                      // index of constraint in this object
        while(iter!=data.end() && ind_cons_this >= (*iter)->numConstraints()) {
            ind_cons_this -= (*iter)->numConstraints(); 
            iter++; 
        }
        if(iter==data.end()) 
            return 0;   // may only happen in the case penaltyCons<0 and ind_orb<numOrbits; this block of linear matrix is zero
        const CBasicSchwData::SCHWDATATYPE dtype=(*iter)->dataType();
        const COrbitDesc* orb=orbits->getOrbitDesc(ind_orb);
        for(size_t i=0; i<orb->getInfoNum(); i++)
        {   // find info of corresponding Schw.data type
            const CBasicInformation* info=orb->getInfo(i);
            if(info->infoType()==CBasicInformation::IT_SCHW &&
                static_cast<const CSchwInformation*>(info)->schwDataType()==dtype)
            {   // found corresponding information
                const CSchwInformation* info_s=static_cast<const CSchwInformation*>(info);
                return static_cast<smNumType>( (*iter)->getOrbitConstraint(ind_cons_this, info_s) );
            }
        }
        return 0;   // coefficient not found, shouldn't raise an error?
    }
}

std::string CSchwModelQuadOpt::performModelling()
{
    if(!correct) 
        return "Error, model is not consistent with orbit data";
    if(orbits==NULL || orbits->size()==0 || solver==NULL || data.size()==0) 
        return "Error, empty solver, model or orbit library!";
    // internal arrays used in optimization procedure
    CMatrixDiagonal<smNumType> *objFuncQuad=NULL;
    CMatrixBasicSchwModel* matrixInterface=new CMatrixBasicSchwModel(this);
    int result = 0;
    try{
        std::vector<smNumType> objFuncLinear(numVar);    // linear terms in the objective function
        std::vector<smNumType> orbitWeight(numVar);      // the solution to be obtained
        std::vector<smNumType> maxOrbitWeight(numVar);   // upper limits on orbit weights if necessary
        if(lambda>0 || penaltyConsQuad>0)                // quadratic terms in the objective function
            objFuncQuad = new CMatrixDiagonal<smNumType>(numVar);
        // 1st stage: preparation
        //my_message(FUNCNAME, "Preparing model");
        double sumPriors=0;  // typically is equivalent to the total expected model mass
        for(size_t o=0; o<orbits->size(); o++)
            sumPriors+=orbits->getOrbitDesc(o)->getWeightPrior();
        for(size_t o=0; o<orbits->size(); o++)
        {
            if(orbitPenaltyFnc!=NULL && sumPriors>0)
                objFuncLinear[o] = static_cast<smNumType>(orbitPenaltyFnc->eval(orbits->getOrbitDesc(o))/sumPriors);
            orbitWeight[o]=static_cast<smNumType>(orbits->getOrbitDesc(o)->getWeight());  // current orbit weights, may be used to warm-start the solver
            if(lambda>0) {
                double weightPrior=orbits->getOrbitDesc(o)->getWeightPrior();
                if(weightPrior*sumPriors>0)
                    objFuncQuad->set(o, static_cast<smNumType>(lambda/(weightPrior*sumPriors)));
            }
            maxOrbitWeight[o]=static_cast<smNumType>(maxWeight);
        }
        if(penaltyConsLin>0 || penaltyConsQuad>0)   // add penalty for slack variables
            for(size_t c=0; c<2*numConstrData; c++) {
                objFuncLinear[c+orbits->size()] = static_cast<smNumType>(penaltyConsLin);
                if(penaltyConsQuad>0)
                    objFuncQuad->set(c+orbits->size(), static_cast<smNumType>(penaltyConsQuad));  // add a quadratic weight factor as well, for a more uniform distribution of infeasible cells
            }

        // 2nd stage: optimization itself
        my_message(FUNCNAME, "Calling solver");
        result = solver->callSolver(matrixInterface, rhs, maxOrbitWeight, objFuncLinear, objFuncQuad, &orbitWeight);
        if(objFuncQuad!=NULL) delete objFuncQuad;
        delete matrixInterface;
        if(result<0)   // error in solver
            return solver->errorDescription(result);

        // 3rd stage: post-processing
        my_message(FUNCNAME, "Computing statistics");
        for(size_t o=0; o<orbits->size(); o++)
        {
            double weightPrior=orbits->getOrbitDesc(o)->getWeightPrior();
            if(weightPrior>0 && orbitWeight[o] < weightPrior*SCHW_MIN_ORBIT_WEIGHT_USED) 
                orbitWeight[o]=0;  // eliminate unimportant orbits
            orbits->getOrbitDesc(o)->setWeight(orbitWeight[o]);
        }
        return getStatistics();  // on success return statistical information about solution
    }
    catch(const std::bad_alloc&) {
        if(objFuncQuad!=NULL) delete objFuncQuad;
        delete matrixInterface;
        return "Not enough memory in preparing model";
    }
    catch(...) {
        if(objFuncQuad!=NULL) delete objFuncQuad;
        delete matrixInterface;
        return "Unknown error in Schwarzschild modelling";
    }
}

std::string CSchwModelQuadOpt::getStatistics() const
{
    if(!correct) 
        return "Error, model is not consistent with orbit data";
    if(orbits==NULL || orbits->size()==0) 
        return "Error in getStatistics: no orbits!";
    size_t numOrbitsNonzero=0;     // number of orbits with non-zero weight
    size_t numCoefsInfeasible=0;   // number of coefficients that did not meet requirement values
    double totalDensityPenalty=0;  // sum of densityPenalty for all infeasible coefs
    double entropy=0;
    double totalWeight=0;
    std::vector<double> weights(orbits->size(), 0);
    for(size_t o=0; o<orbits->size(); o++)
    {
        double orbitWeight=orbits->getOrbitDesc(o)->getWeight();
        totalWeight+=orbitWeight;
        if(orbitWeight>0) numOrbitsNonzero++;
        weights[o]=orbitWeight;
    }
    std::sort(weights.begin(), weights.end(), std::greater<double>());
    double avgWeight=totalWeight/orbits->size();
    double sumHighWeights=0;
    size_t numOrbitsHalfMass=0;
    if(avgWeight>0) {
        for(size_t o=0; o<orbits->size(); o++)  // compute entropy
        {
            double orbitWeight=orbits->getOrbitDesc(o)->getWeight();
            double weightPrior=orbits->getOrbitDesc(o)->getWeightPrior();
            if(weightPrior==0) weightPrior = avgWeight;
            if(orbitWeight>0) entropy -= orbitWeight*log(orbitWeight/weightPrior);
            sumHighWeights+=weights[o];
            if(sumHighWeights>=totalWeight*0.5 && numOrbitsHalfMass==0) numOrbitsHalfMass=o;
        }
        entropy /= totalWeight;
    }
    double maxToAvgWeight = weights.front()/avgWeight;
    double fracOrbitsHalfMass = numOrbitsHalfMass*1.0/orbits->size();

    for(size_t c=0; c<numConstr; c++)
    {
        double deviation=0;
        size_t no=0, noz=0;
        calcCoefDeviation(c, &deviation, &no, &noz);
        if(deviation!=0) 
        {
            numCoefsInfeasible++;
            totalDensityPenalty+=fabs(deviation);
        }
    }
    return
        convertToString(numOrbitsNonzero)+" orbits out of "+convertToString(orbits->size())+" with nonzero weights\n" +
        (totalWeight>0 ?   "Total weight: " + convertToString(totalWeight) + "\n" +
        convertToString(fracOrbitsHalfMass*100, 3) + "% orbits make up 50% total mass" +
        (fracOrbitsHalfMass>0.2 ? " (good)" : fracOrbitsHalfMass>0.1 ? " (fair)" : " (poor)") + "\n" +
        "entropy: " + convertToString(entropy, 3) + 
        (entropy>=-0.5 ? " (good)" : entropy>=-1.0 ? " (fair)" : " (poor)") + "\n" +
        "Max/avg weight: " + convertToString(maxToAvgWeight, 3) + 
        (maxToAvgWeight<=10 ? " (good)" : maxToAvgWeight<=50 ? " (fair)" : " (poor)") + "\n" : "" ) +
        (numCoefsInfeasible>0 ? "WARNING: " +
        convertToString(numCoefsInfeasible)+" coefs out of "+convertToString(numConstr)+" infeasible\n" +
        convertToString(totalDensityPenalty, 3)+" total difference" : "Model is feasible");
}

void CSchwModelQuadOpt::calcCoefDeviation(size_t coefIndex, double* deviation, size_t* numOrbits, size_t* numOrbitsNonzero) const
{
    assert(deviation);
    assert(numOrbits);
    assert(numOrbitsNonzero);
    *deviation=0;
    *numOrbits=0;
    *numOrbitsNonzero=0;
    if(!correct || coefIndex>numConstr) return;
    for(size_t o=0; o<orbits->size(); o++) {
        double weight=orbits->getOrbitDesc(o)->getWeight();
        double value =getLinearMatrixElem(o, coefIndex);
        if(value!=0) {
            (*numOrbits)++;
            if(weight!=0) {
                (*numOrbitsNonzero)++;
                *deviation += value * weight;
            }
        }
    }
    *deviation -= rhs[coefIndex];
    if(fabs(*deviation) <= fabs(rhs[coefIndex]==0 ? 1.0 : rhs[coefIndex])*SCHW_MAX_CELLMASS_REL_DIFFERENCE)
        *deviation=0;
}

double CSchwModelQuadOpt::calcCoefComposition(size_t coefIndex, const CBasicOrbitFilteringFnc* weightFnc) const
{
    if(!correct || coefIndex>numConstr || weightFnc==NULL) return 0;
    double sumW=0;
    for(size_t o=0; o<orbits->size(); o++)
        sumW += getLinearMatrixElem(o, coefIndex) * orbits->getOrbitDesc(o)->getWeight() * 
            weightFnc->eval(orbits->getOrbitDesc(o));
    return sumW;
}

//-------------------------------------------------------------//
double CChaosOrbitFilteringFnc::eval(const smile::COrbitDesc *orbit) const
{
    if(orbit==NULL) return 0;
    const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>(orbit->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
    if(info!=NULL) 
    {
        double wf=log10(info->getlfccdiff())-log10(chaoticMinFreqDiff) + 0.5;   // `soft' transition from reg to chaos, based on freq.diff.rate
        if(wf<0) wf=0; if(wf>1) wf=1;
        if(info->getlambda()>chaoticMinLambda) wf=1;
        return chaoticWeightFactor*wf;
    }
    else return 0;
}

double CCentrophilicOrbitFilteringFnc::eval(const smile::COrbitDesc *orbit) const
{
    if(orbit==NULL) return 0;
    const CPericenterInformation<float>* info=static_cast<const CPericenterInformation<float>*>(orbit->getInfoByType(CBasicInformation::IT_PERICENTER));
    if(info!=NULL) {
        float fitIntercept, fitSlope, fitScatter, fitSignificance, Lcirc2;
        info->getParams(&fitIntercept, &fitSlope, &fitScatter, &fitSignificance, &Lcirc2);
        if(fitIntercept==0) return weightFactor; else return 0;
    }
    else return 0;
}

double CAngMomOrbitFilteringFnc::eval(const smile::COrbitDesc *orbit) const
{
    if(orbit==NULL) return 0;
    const CPericenterInformation<float>* info_p=static_cast<const CPericenterInformation<float>*>
        (orbit->getInfoByType(CBasicInformation::IT_PERICENTER));
    if(info_p==NULL) return 0;
    float fitIntercept, fitSlope, fitScatter, fitSignificance, Lcirc2;
    info_p->getParams(&fitIntercept, &fitSlope, &fitScatter, &fitSignificance, &Lcirc2);
    const COrbitInformation<float>* info=static_cast<const COrbitInformation<float>*>
        (orbit->getInfoByType(CBasicInformation::IT_TRAJ_ANALYSIS));
    if(info!=NULL)
    {
        double Lztrue=info->getLavg(2)+Omega*(pow_2(info->getinertia(0))+pow_2(info->getinertia(1)));
        double norm=sqrt(Lcirc2+0.1*pow_2(Lztrue));
        return weightFactor*Lztrue/norm;   // relative magnitude of L_z (normalized by ang.mom.of circular orbit)
    }
    else return 0;
}

/** Correspondence between Schw.model names and types **/
const char* getSchwDataNameByType(CBasicSchwData::SCHWDATATYPE dataType)
{
    switch(dataType) {
    case CBasicSchwData::SD_DENS_GRIDCLASSIC:     return CSchwDataDensGridClassic::myName();
    case CBasicSchwData::SD_DENS_GRIDCYLINDRICAL: return CSchwDataDensGridCylindrical::myName();
    case CBasicSchwData::SD_DENS_SHMESH:          return CSchwDataDensSHMesh::myName();
    case CBasicSchwData::SD_DENS_SHBSE:           return CSchwDataDensSHBSE::myName();
    case CBasicSchwData::SD_KINEM_SHELL:          return CSchwDataKinemShell::myName();
    case CBasicSchwData::SD_KINEM_ANGMOMDIST:     return CSchwDataAngMomDist::myName();
    default: return "Error";
    }
};

CBasicSchwData::SCHWDATATYPE getSchwDataTypeByName(const std::string& dataName)
{
    if(strings_equal(dataName, CSchwDataDensGridClassic    ::myName())) return CBasicSchwData::SD_DENS_GRIDCLASSIC;
    if(strings_equal(dataName, CSchwDataDensGridCylindrical::myName())) return CBasicSchwData::SD_DENS_GRIDCYLINDRICAL;
    if(strings_equal(dataName, CSchwDataDensSHMesh         ::myName())) return CBasicSchwData::SD_DENS_SHMESH;
    if(strings_equal(dataName, CSchwDataDensSHBSE          ::myName())) return CBasicSchwData::SD_DENS_SHBSE;
    if(strings_equal(dataName, CSchwDataKinemShell         ::myName())) return CBasicSchwData::SD_KINEM_SHELL;
    if(strings_equal(dataName, CSchwDataAngMomDist         ::myName())) return CBasicSchwData::SD_KINEM_ANGMOMDIST;
    return CBasicSchwData::SD_UNKNOWN;  // default
};

CBasicOrbitRuntimeFncCreator* createSchwCreator(const CBasicSchwData* schwData)
{
    switch(schwData->dataType()) {
        case CBasicSchwData::SD_KINEM_SHELL:
            return new COrbitRuntimeSchwKinemShellCreator(static_cast<const CSchwDataKinemShell*>(schwData));
        case CBasicSchwData::SD_KINEM_ANGMOMDIST:
            return new COrbitRuntimeSchwAngMomDistCreator(static_cast<const CSchwDataAngMomDist*>(schwData));
        case CBasicSchwData::SD_DENS_GRIDCLASSIC:
        case CBasicSchwData::SD_DENS_GRIDCYLINDRICAL:
            return new COrbitRuntimeSchwDensGridCreator(static_cast<const CSchwDataDensGrid*>(schwData));
        case CBasicSchwData::SD_DENS_SHBSE:
        case CBasicSchwData::SD_DENS_SHMESH:
            return new COrbitRuntimeSchwDensSHCreator(static_cast<const CSchwDataDensSH*>(schwData));
        default:
            return NULL;  // unknown class
    }
};

void createRadialGrid(const smile::CDensity *density, size_t numShells, double totalMass, double innerShellMass, double outerShellMass, smile::vectord *grid)
{
    assert(grid);
    if(outerShellMass==0) outerShellMass=1.0-1.0/(numShells+1);
    if(innerShellMass==0) innerShellMass=outerShellMass/numShells;
    if(totalMass<=0) { 
        my_error(FUNCNAME, "Error, density model does not seem to have finite mass"); 
        totalMass=1; 
    }
    if(numShells<=0) {
        my_error(FUNCNAME, "Error, number of grid shells is zero");
        return;
    }
    vectord gridtmp;
    // first create a (possibly non-uniform) grid in the enclosed mass (normalized to total mass)
    createNonuniformGrid(numShells, innerShellMass, outerShellMass, false, &gridtmp);
    // now convert the grid in mass to the grid in radius
    for(size_t s=0; s<numShells; s++)
        gridtmp[s] *= totalMass;
    density->getRadiiByMass(gridtmp, grid);
    if(grid->at(0)<0) {
        my_error(FUNCNAME, "Error determining grid in mass");
        return;
    }
}

}  // namespace
