#include "utils.h"
#include <iostream>
#include <cmath>
#include <cassert>
#include <ctime>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_version.h>

namespace smile{

inline double pow_2(const double x) { return x*x; }

/* -------- error reporting routines ------- */

/// remove common prefix if it is the namespace "smile::", and remove function signature from GCC __PRETTY_FUNCTION__
std::string undecorateFunction(std::string origin)
{
#ifdef __GNUC__
    // parse the full function signature returned by __PRETTY_FUNCTION__
    std::string::size_type ind=origin.find('(');
    if(ind!=std::string::npos) origin.erase(ind);
    ind=origin.rfind(' ');
    if(ind!=std::string::npos) origin.erase(0, ind+1);
#endif
    if(origin.substr(0, 7)=="smile::") origin.erase(0, 7);
    std::string::size_type indc=origin.find(':');
    if(indc!=std::string::npos && indc+1==origin.size()/2 && origin[indc+1]==':') {
        if(origin.substr(0, indc)==origin.substr(indc+2)) origin.replace(indc+2, std::string::npos, "Ctor");
        if(origin[indc+2]=='~' && origin.substr(0, indc)==origin.substr(indc+3)) origin.replace(indc+2, std::string::npos, "Dtor");
    }
    return origin;
}

/// default routine that dumps text messages to stderr
void my_stderr_show_message(const std::string &origin, const std::string &message)
{
    std::cerr << ((origin.empty() ? "" : "{"+undecorateFunction(origin)+"} ") + message+"\n"); 
}

/// global variable to the routine that displays errors (if it is NULL then nothing is done)
show_message_type* my_error_ptr   = &my_stderr_show_message;

/// global variable to the routine that shows information messages 
/// (emitted by various time-consuming routines to display progress), if NULL then no messaging is done
show_message_type* my_message_ptr = &my_stderr_show_message;

// replacement for GSL error handler that doesn't crash program
void gsl_verbose_error_handler (const char *reason, const char *file, int line, int gsl_errno)
{
    if( // list error codes that are non-critical and don't need to be reported
        gsl_errno == GSL_ETOL || 
        gsl_errno == GSL_EROUND || 
        gsl_errno == GSL_ESING || 
        gsl_errno == GSL_EDIVERGE ||
        //gsl_errno == GSL_EOVRFLW || 
        //gsl_errno == GSL_EUNDRFLW || 
        0) return;  // do nothing
    my_stderr_show_message("GSL", "error "+convertToString(gsl_errno)+" in "+std::string(file)+" at line "+
        convertToString(line)+": "+std::string(reason));
}

/* ----------- string/number conversion and parsing routines ----------------- */
//  Pretty-print - convert float (and integer) numbers to string of fixed width.
//  Uses some sophisticated techniques to fit the number into a string of exactly the given length.
std::string pp(double num, unsigned int width)
{
    std::string result, tmp;
    if(num==0) { 
        for(int i=0; i<static_cast<int>(width)-1; i++) result+=' ';
        result+='0';
        return result;
    }
    unsigned int sign=num<0;
    double mag=log10(fabs(num));
    std::ostringstream stream;
    if(num!=num || num/2==num || num+0!=num)
    {
        if(width>=4) stream << std::setw(width) << num;
        else stream << "#ERR";
    }
    else if(width<=2+sign)  // display int if possible
    {
        if(mag<0) stream << (sign?"-":"+") << 0;
        else if(mag>=2-sign) stream << (sign?"-":"+") << "!";
        else stream << (int)floor(num+0.5);
    }
    else if(mag>=0 && mag+sign<width && mag<6)  // try fixed-point for |x|>=1
    {
        stream << std::setw(width) << std::setprecision(width-1-sign) << num;
        if(stream.str().find('e')!=std::string::npos) { 
            //std::string x=stream.str();
            //size_t e=x.find('e');
            stream.str(""); 
            stream << (int)floor(num+0.5); 
        }
    }
    else if(mag<0 && -mag+sign<width-2 && mag>=-4) // try fixed-point for |x|<1
    {
        stream << std::setw(width) << std::setprecision(width-1-sign+(int)floor(mag)) << num;
    }
    else
    {
        std::ostringstream strexp;
        int expon=static_cast<int>(floor(mag));
        strexp << std::setiosflags(std::ios_base::showpos) << expon;
        std::string expstr=strexp.str();
        size_t w=(width-expstr.size()-1);
        double mant=num*pow(10.0, -expon);
        if(w<sign)  // no luck with exp-format -- try fixed 
        {
            stream << (sign?"-":"+") << (mag<0 ? "0" : "!");
        }
        else 
        {
            if(w==sign) 
                stream << (sign?"-":""); // skip mantissa
            else if(w<=2+sign)
            { 
                int mantint=(int)floor(fabs(mant)+0.5);
                if(mantint>=10) mantint=9;  // hack
                stream << (sign?"-":"") << mantint;
            }
            else
                stream << std::setprecision(w-1-sign) << mant;
            stream << "e" << expstr;
        }
    }
    result=stream.str();
    // padding if necessary (add spaces after string)
    while(result.length()<static_cast<size_t>(width))
        result+=" ";
    if(result.length()>static_cast<size_t>(width))  // cut tail if necessary (no warning given!)
        result=result.substr(0,width);
    return result;
}

void splitString(const std::string& src, const std::string& delim, std::vector<std::string> *result)
{
    result->clear();
    std::string str(src);
    std::string::size_type indx=str.find_first_not_of(delim);
    if(indx==std::string::npos) 
    {
        result->push_back("");   // ensure that result contains at least one element
        return;
    }
    if(indx>0)  // remove unnecessary delimiters at the beginning
        str=str.erase(0, indx);
    while(!str.empty())
    {
        indx=str.find_first_of(delim);
        if(indx==std::string::npos) indx=str.size();
        result->push_back(str.substr(0, indx));
        str=str.erase(0, indx);
        indx=str.find_first_not_of(delim);
        if(indx==std::string::npos) return;
        str=str.erase(0, indx);
    }
}

bool ends_with_str(const std::string& str, const std::string& end)
{
    return end.size()<=str.size() && str.find(end, str.size()-end.size())!=str.npos;
}

bool strings_equal(const std::string& str1, const std::string& str2)
{
    std::string::size_type len=str1.size();
    if(len!=str2.size()) return false;
    for(std::string::size_type i=0; i<len; i++)
        if(tolower(str1[i]) != tolower(str2[i])) return false;
    return true;
}

bool strings_equal(const std::string& str1, const char* str2)
{
    if(str2==NULL) return false;
    for(std::string::size_type i=0; i<str1.size(); i++)
        if(str2[i]==0 || tolower(str1[i]) != tolower(str2[i])) return false;
    return str2[str1.size()]==0;  // ensure that the 2nd string length is the same as the 1st
}

/* --------- random numbers -------- */
// global random number generator
gsl_rng* randgen=gsl_rng_alloc(gsl_rng_default);

// randomly init random number generator
void randomize()
{
    gsl_rng_set(randgen, (unsigned int)time( NULL ));
}

// convenience function to generate a random number using global generator
double getRandomNumber()
{
    return gsl_rng_uniform(randgen);
}

// generate 2 random numbers with normal distribution, using Box-Muller approach
void getNormalRandomNumbers(double* num1, double* num2)
{
    double p1=getRandomNumber();
    double p2=getRandomNumber();
    if(p1>0&&p1<=1) p1=sqrt(-2*log(p1));
    *num1=p1*sin(2*M_PI*p2);
    *num2=p1*cos(2*M_PI*p2);
}

// generate random 2d or 3d vector with unit length
void getRandomUnitVector(double* x, double* y, double* z)
{
    double costh=(z==NULL)? 0 : (getRandomNumber()*2-1);
    double sinth=sqrt(1-pow_2(costh));
    double phi=getRandomNumber()*2*M_PI;
    *x = sinth*cos(phi);
    *y = sinth*sin(phi);
    if(z!=NULL) *z=costh;
}

// generate random 3d rotation matrix (using Arvo(1992)'s algo)
void getRandomRotationMatrix(double* m)
{
    double theta=getRandomNumber()*2*M_PI;
    double phi=getRandomNumber()*2*M_PI;
    double mu=getRandomNumber()*2;
    double nu=std::sqrt(mu);
    double vx=sin(phi)*nu;
    double vy=cos(phi)*nu;
    double vz=sqrt(2-mu);
    double st=sin(theta);
    double ct=cos(theta);
    double sx=vx*ct-vy*st;
    double sy=vx*st+vy*ct;
    m[0]=vx*sx-ct;
    m[1]=vx*sy-st;
    m[2]=vx*vz;
    m[3]=vy*sx+st;
    m[4]=vy*sy-ct;
    m[5]=vy*vz;
    m[6]=vz*sx;
    m[7]=vz*sy;
    m[8]=1-mu;
}

/* --------- wrappers for GSL equation solvers and minimizers ------------- */
// ordinary 1d equation solver
double findRoot(gsl_function *F, double xlower, double xupper, double xdefault, double reltoler)
{
    if(F->function(xlower, F->params) * F->function(xupper, F->params) > 0)
        return xdefault;   // endpoints do not enclose root
    gsl_root_fsolver *solv = gsl_root_fsolver_alloc(gsl_root_fsolver_brent);
    gsl_root_fsolver_set(solv, F, xlower, xupper);
    int status=0, iter=0;
    double xroot=xdefault;
    do{
        iter++;
        status = gsl_root_fsolver_iterate(solv);
        if(status!=GSL_SUCCESS) break;
        xroot  = gsl_root_fsolver_root(solv);
        xlower = gsl_root_fsolver_x_lower(solv);
        xupper = gsl_root_fsolver_x_upper(solv);
        status = gsl_root_test_interval (xlower, xupper, 0, reltoler);
    }
    while(status == GSL_CONTINUE && iter < 50);
    gsl_root_fsolver_free(solv);
    return xroot;
}

// 1d minimizer
double findMin(gsl_function *F, double xlower, double xupper, double xinit, double reltoler)
{
    gsl_min_fminimizer *minser = gsl_min_fminimizer_alloc(gsl_min_fminimizer_brent);
    double xroot=xinit;
    if(gsl_min_fminimizer_set(minser, F, xinit, xlower, xupper) == GSL_SUCCESS) 
    {
        int status=0, iter=0;
        do{
            iter++;
            status = gsl_min_fminimizer_iterate (minser);
            if(status!=GSL_SUCCESS) break;
            xroot  = gsl_min_fminimizer_x_minimum (minser);
            xlower = gsl_min_fminimizer_x_lower (minser);
            xupper = gsl_min_fminimizer_x_upper (minser);
            status = gsl_root_test_interval (xlower, xupper, 0, reltoler);
        }
        while (status == GSL_CONTINUE && iter < 50);
    }
    gsl_min_fminimizer_free(minser);
    return xroot;
}

// 1d minimizer without prior knowledge of minimum location
double findMinGuess(gsl_function *F, double xlower, double xupper, double reltoler)
{
    double xinit =(xlower+xupper)/2;
    double ylower=F->function(xlower, F->params);
    double yupper=F->function(xupper, F->params);
    double yinit =F->function(xinit, F->params);
    int iter=0;
    while( (yinit>ylower || yinit>yupper) && iter<50 && gsl_root_test_interval(xlower, xupper, 0, reltoler)==GSL_CONTINUE)
    {
        if(yinit<ylower) {
            xlower=xinit;
            ylower=yinit;
        } else 
        if(yinit<yupper) {
            xupper=xinit;
            yupper=yinit;
        } else {  // pathological case - initial guess was higher than both ends
            double xmin1=findMinGuess(F, xlower, xinit, reltoler);
            double xmin2=findMinGuess(F, xinit, xupper, reltoler);
            double ymin1=F->function(xmin1, F->params);
            double ymin2=F->function(xmin2, F->params);
            return ymin1<ymin2 ? xmin1 : xmin2;
        }
        xinit=(xlower+xupper)/2;
        yinit=F->function(xinit, F->params);
        iter++;
    }
    if(yinit<=ylower && yinit<=yupper)
        return findMin(F, xlower, xupper, xinit, reltoler);  // normal min-search
    else if(ylower<yinit) return xlower;
    else return xupper;    // otherwise return one of endpoints
}

/* ----------- generation of unequally spaced grids --------------- */
// Creation of grid with exponentially increasing cells
/// \cond INTERNAL_DOCS
static double findA(double A, void* params)
{
    double dynrange = ((double*)params)[0];
    double nnodes   = ((double*)params)[1];
    if(A==0) return nnodes-dynrange;
    return (exp(A*nnodes)-1)/(exp(A)-1) - dynrange;
}
/// \endcond
void createNonuniformGrid(size_t nnodes, double xmin, double xmax, bool zeroelem, std::vector<double> *grid)
{   // create grid so that x_k = B*(exp(A*k)-1)
    assert(grid);
    double A, B, dynrange=xmax/xmin;
    grid->resize(nnodes);
    int indexstart=zeroelem?1:0;
    if(zeroelem) { grid->at(0)=0; nnodes--; }
    if(gsl_fcmp(static_cast<double>(nnodes), dynrange, 1e-6)==0)  // no need for non-uniform grid
    {
        for(size_t i=0; i<nnodes; i++)
            grid->at(i+indexstart) = xmin+(xmax-xmin)*i/(nnodes-1);
        return;
    }
    // solve for A:  dynrange = (exp(A*nnodes)-1)/(exp(A)-1)
    double params[2] = {dynrange, static_cast<double>(nnodes)};
    gsl_function F;
    F.function=&findA;
    F.params=&params;
    // first localize the root coarsely, to avoid overflows in root solver
    double Amin=0, Amax=0;
    double step=1; while(step>10./nnodes) step/=2;
    if(dynrange>nnodes) {
        while(Amax<10 && findA(Amax, &params)<=0) Amax+=step;
        Amin=Amax-step;
    } else {
        while(Amin>-10 && findA(Amin, &params)>=0) Amin-=step;
        Amax=Amin+step;
    }
    A = findRoot(&F, Amin, Amax);
    B = xmin / (exp(A)-1);
    for(size_t i=0; i<nnodes; i++)
        grid->at(i+indexstart) = B*(exp(A*(i+1))-1);
    grid->at(nnodes-1+indexstart)=xmax;
}

/// creation of a grid with minimum guaranteed number of input points per bin
static void makegrid(std::vector<double>::iterator begin, std::vector<double>::iterator end, double startval, double endval)
{
    double step=(endval-startval)/(end-begin-1);
    while(begin!=end){
        *begin=startval;
        startval+=step;
        begin++;
    }
    *(end-1)=endval;  // exact value
}
void createAlmostUniformGrid(const std::vector<double> &srcpoints, size_t minbin, size_t *gridsize, std::vector<double> *grid)
{
    if(grid==NULL || gridsize==NULL || srcpoints.size()==0) return;  // invalid call
    *gridsize = std::max<size_t>(2, std::min<size_t>(*gridsize, static_cast<size_t>(srcpoints.size()/minbin)));
    grid->resize(*gridsize);
    std::vector<double>::iterator gridbegin=grid->begin(), gridend=grid->end();
    std::vector<double>::const_iterator srcbegin=srcpoints.begin(), srcend=srcpoints.end();
    std::vector<double>::const_iterator srciter;
    std::vector<double>::iterator griditer;
    bool ok=true, directionBackward=false;
    int numChangesDirection=0;
    do{
        makegrid(gridbegin, gridend, *srcbegin, *(srcend-1));
        ok=true; 
        // find the index of bin with the largest number of points
        int largestbin=-1;
        size_t maxptperbin=0;
        //static_cast<int>((gridend-gridbegin)/2+gridbegin-grid->begin());
        for(srciter=srcbegin, griditer=gridbegin; griditer!=gridend-1; griditer++) {
            size_t ptperbin=0;
            while(srciter+ptperbin!=srcend && *(srciter+ptperbin) < *(griditer+1)) 
                ptperbin++;
            if(ptperbin>maxptperbin) {
                maxptperbin=ptperbin;
                largestbin=griditer-grid->begin();
            }
            srciter+=ptperbin;
        }
        //size_t maxptperbin=0;
        // check that all bins contain at least minbin srcpoints
        if(!directionBackward) {  // forward scan
            srciter = srcbegin;
            griditer = gridbegin;
            while(ok && griditer!=gridend-1) {
                size_t ptperbin=0;
                while(srciter+ptperbin!=srcend && *(srciter+ptperbin) < *(griditer+1)) 
                    ptperbin++;
                if(ptperbin>=minbin)  // ok, move to the next one
                {
                    /*if(ptperbin>maxptperbin) {
                    maxptperbin=ptperbin;
                    largestbin=static_cast<int>(griditer-grid->begin());
                    }*/
                    griditer++;
                    srciter+=ptperbin;
                } else {  // assign minbin points and decrease the available grid interval from the front
                    if(griditer-grid->begin() < largestbin) { 
                        // bad bin is closer to the grid front; move gridbegin forward
                        while(ptperbin<minbin && srciter+ptperbin!=srcend) 
                            ptperbin++;
                        if(srciter+ptperbin==srcend)
                            directionBackward=true; // oops, hit the end of array..
                        else {
                            srcbegin=srciter+ptperbin;
                            gridbegin=griditer+1;
                        }
                    } else {
                        directionBackward=true;
                    }   // will restart scanning from the end of the grid
                    ok=false;
                }
            }
        } else {  // backward scan
            srciter = srcend-1;
            griditer = gridend-1;
            while(ok && griditer!=gridbegin) {
                size_t ptperbin=0;
                while(srciter+1-ptperbin!=srcbegin && *(srciter-ptperbin) >= *(griditer-1))
                    ptperbin++;
                if(ptperbin>=minbin)  // ok, move to the previous one
                {
                    /*if(ptperbin>maxptperbin) {
                    maxptperbin=ptperbin;
                    largestbin=static_cast<int>(griditer-grid->begin());
                    }*/
                    griditer--;
                    if(srciter+1-ptperbin==srcbegin)
                        srciter=srcbegin;
                    else
                        srciter-=ptperbin;
                } else {  // assign minbin points and decrease the available grid interval from the back
                    if(griditer-grid->begin() <= largestbin) { 
                        // bad bin is closer to the grid front; reset direction to forward
                        directionBackward=false;
                        numChangesDirection++;
                        if(numChangesDirection>10) {
                            my_message(FUNCNAME, "grid creation seems not to converge?");
                            return;  // don't run forever but would not fulfill the minbin condition
                        }
                    } else {
                        // move gridend backward
                        while(ptperbin<minbin && srciter-ptperbin!=srcbegin) 
                            ptperbin++;
                        if(srciter-ptperbin==srcbegin) {
                            directionBackward=false;
                            numChangesDirection++;
                            if(numChangesDirection>10) {
                                my_message(FUNCNAME, "grid creation seems not to converge?");
                                return;  // don't run forever but would not fulfill the minbin condition
                            }
                        } else {
                            srcend=srciter-ptperbin+1;
                            gridend=griditer;
                        }
                    }
                    ok=false;
                }
            }
        }
    } while(!ok);
}

/* ----------- spline functions --------------- */

// clamped cubic spline, with given values of first derivative at endpoints and linear extrapolation beyond the definition range
// Hacking into GSL spline machinery by defining a new spline type which differs from natural cubic spline only in initialization
/// \cond INTERNAL_DOCS
typedef struct
{
  double * c;
  double * g;
  double * diag;
  double * offdiag;
} cspline_state_t;

// 1st derivatives at left- and rightmost points are given in ya[n] & ya[n+1]
// if one wants to have a 'natural' spline boundary condition then pass NaN as the value of derivative.
int ccspline_init (void * vstate, const double xa[], const double ya[], size_t size)
{
  cspline_state_t *state = (cspline_state_t *) vstate;

  size_t i;
  size_t num_points = size;
  size_t max_index = num_points - 1;  /* Engeln-Mullges + Uhlig "n" */
  size_t sys_size = max_index - 1;    /* linear system is sys_size x sys_size */

  double der1=ya[size], der2=ya[size+1];  // numbers must be supplied!
  for (i = 0; i < sys_size; i++)
    {
      const double h_i   = xa[i + 1] - xa[i];
      const double h_ip1 = xa[i + 2] - xa[i + 1];
      const double ydiff_i   = ya[i + 1] - ya[i];
      const double ydiff_ip1 = ya[i + 2] - ya[i + 1];
      const double g_i = (h_i != 0.0) ? 1.0 / h_i : 0.0;
      const double g_ip1 = (h_ip1 != 0.0) ? 1.0 / h_ip1 : 0.0;
      state->offdiag[i] = h_ip1;
      state->diag[i] = 2.0 * (h_ip1 + h_i);
      state->g[i] = 3.0 * (ydiff_ip1 * g_ip1 -  ydiff_i * g_i);
      if(i == 0 && !gsl_isnan(der1)) {
        state->diag[i] = 1.5 * h_i + 2.0 * h_ip1;
        state->g[i] = 3.0 * (ydiff_ip1 * g_ip1 - 1.5 * ydiff_i * g_i + 0.5 * der1);
      }
      if(i == sys_size-1 && !gsl_isnan(der2)) {
        state->diag[i] = 1.5 * h_ip1 + 2.0 * h_i;
        state->g[i] = 3.0 * (1.5 * ydiff_ip1 * g_ip1 - 0.5 * der2 - ydiff_i * g_i);
      }
    }

  if (sys_size == 1)
    {
      state->c[1] = state->g[0] / state->diag[0];
      return GSL_SUCCESS;
    }
  else
    {
      gsl_vector_view g_vec = gsl_vector_view_array(state->g, sys_size);
      gsl_vector_view diag_vec = gsl_vector_view_array(state->diag, sys_size);
      gsl_vector_view offdiag_vec = gsl_vector_view_array(state->offdiag, sys_size - 1);
      gsl_vector_view solution_vec = gsl_vector_view_array ((state->c) + 1, sys_size);
 
      int status = gsl_linalg_solve_symm_tridiag(&diag_vec.vector, &offdiag_vec.vector, &g_vec.vector, &solution_vec.vector);
      if(!gsl_isnan(der1)) 
          state->c[0] = ( 3.0*(ya[1]-ya[0])/(xa[1]>xa[0] ? xa[1]-xa[0] : 1) - 3.0*der1 - state->c[1]*(xa[1]-xa[0]) )*0.5/(xa[1]>xa[0] ? xa[1]-xa[0] : 1);
      else state->c[0]=0.0;
      if(!gsl_isnan(der2)) 
          state->c[max_index] = -( 3*(ya[max_index]-ya[max_index-1])/(xa[max_index]-xa[max_index-1]) - 3*der2 + state->c[max_index-1]*(xa[max_index]-xa[max_index-1]) )*0.5/(xa[max_index]-xa[max_index-1]);
      else state->c[max_index]=0.0;
      return status;
    }
}
/// \endcond

// modified functions for spline evaluation that extrapolate last segment linearly beyond the definition region
double gsl_spline_evalx(const gsl_spline* spline, double x, gsl_interp_accel * a) 
{
    const cspline_state_t *state = (const cspline_state_t *)spline->interp->state; 
    const size_t size=spline->interp->size;
    if(x<spline->interp->xmin) {
        double dx  =  spline->x[1]-spline->x[0];
        double der = (spline->y[1]-spline->y[0])/dx - dx*(state->c[1]+2*state->c[0])/3.0;
        return spline->y[0] + der*(x-spline->interp->xmin);
    }
    else if(x>spline->interp->xmax) {
        double dx  =  spline->x[size-1]-spline->x[size-2];
        double der = (spline->y[size-1]-spline->y[size-2])/dx + dx*(state->c[size-2]+2*state->c[size-1])/3.0;
        return spline->y[size-1] + der*(x-spline->interp->xmax);
    }
    else // use standard method
        return gsl_spline_eval(spline, x, a);
}

double gsl_spline_eval_derivx(const gsl_spline* spline, double x, gsl_interp_accel * a) 
{
    const size_t size=spline->interp->size;
    const cspline_state_t *state = (const cspline_state_t *)spline->interp->state; 
    if(x<spline->interp->xmin) {
        double dx = spline->x[1]-spline->x[0];
        return (spline->y[1]-spline->y[0])/dx - dx*(state->c[1]+2*state->c[0])/3.0;
    }
    else if(x>spline->interp->xmax) {
        double dx = spline->x[size-1]-spline->x[size-2];
        return (spline->y[size-1]-spline->y[size-2])/dx + dx*(state->c[size-2]+2*state->c[size-1])/3.0;
    }
    else // use standard method
        return gsl_spline_eval_deriv(spline, x, a);
}

double gsl_spline_eval_deriv2x(const gsl_spline* spline, double x, gsl_interp_accel * a) 
{
    if(x<spline->interp->xmin || x>spline->interp->xmax) 
        return 0;
    else // use standard method
        return gsl_spline_eval_deriv2(spline, x, a);
}

// convenience function that evaluates spline value, derivative and 2nd derivative at once (faster than doing it separately)
void cspline_eval_all(const gsl_spline* spline, double x, double* val, double* der, double* der2)
{   // assuming that the spline is a cubic spline and x is within the definition region (this must be ensured by the calling routine)
    size_t index = 0;
    size_t indhi = spline->size-1;
    while(indhi > index + 1) {    // binary search to determine the spline segment that contains x
        size_t i = (indhi + index)/2;
        if(spline->x[i] > x)
            indhi = i;
        else
            index = i;
    }
    double x_hi = spline->x[index + 1];
    double x_lo = spline->x[index];
    double dx = x_hi - x_lo;
    double y_lo = spline->y[index];
    double y_hi = spline->y[index + 1];
    double dy = y_hi - y_lo;
    double delx = x - x_lo;
    double c_i   = static_cast<const cspline_state_t*>(spline->interp->state)->c[index];
    double c_ip1 = static_cast<const cspline_state_t*>(spline->interp->state)->c[index+1];
    double b_i = (dy / dx) - dx * (c_ip1 + 2.0 * c_i) / 3.0;
    double d_i = (c_ip1 - c_i) / (3.0 * dx);
    if(val!=NULL)  *val = y_lo + delx * (b_i + delx * (c_i + delx * d_i));
    if(der!=NULL)  *der = b_i + delx * (2.0 * c_i + 3.0 * d_i * delx);
    if(der2!=NULL) *der2= 2.0 * c_i + 6.0 * d_i * delx;
}

bool check_spline_monotonic(const gsl_spline* spline)
{
    bool ismonotonic=true;
    for(size_t index=0; ismonotonic && index<spline->interp->size-1; index++) {
        double dx = spline->x[index + 1] - spline->x[index];
        double dy = spline->y[index + 1] - spline->y[index];
        double c_i   = static_cast<const cspline_state_t*>(spline->interp->state)->c[index];
        double c_ip1 = static_cast<const cspline_state_t*>(spline->interp->state)->c[index+1];
        double a = dx * (c_ip1 - c_i);
        double b = 2 * dx * c_i;
        double c = (dy / dx) - dx * (c_ip1 + 2.0 * c_i) / 3.0;
        // derivative is  a * chi^2 + b * chi + c,  with 0<=chi<=1 on the given interval.
        double D = b*b-4*a*c;
        if(D>=0) { // need to check roots
            double chi1 = (-b-sqrt(D))/(2*a);
            double chi2 = (-b+sqrt(D))/(2*a);
            if( (chi1>=0 && chi1<=1) || (chi2>=0 && chi2<=1) )
                ismonotonic=false;    // there is a root ( y'=0 ) somewhere on the given interval
        }  // otherwise there are no roots
    }
    return ismonotonic;
}

/// \cond INTERNAL_DOCS
const gsl_interp_type ccspline_type = 
{
  "ccspline", 
  3,
  gsl_interp_cspline->alloc,
  &ccspline_init,
  gsl_interp_cspline->eval,
  gsl_interp_cspline->eval_deriv, 
  gsl_interp_cspline->eval_deriv2, 
  gsl_interp_cspline->eval_integ,
  gsl_interp_cspline->free
};
const gsl_interp_type * gsl_interp_ccspline = &ccspline_type;
/// \endcond

// emulate "copy constructor" for non-class structure as gsl_spline
gsl_spline* copyspline(const gsl_spline* src)
{
    if(src==NULL) 
        return NULL;  // wouldn't hurt to call it with an empty input
    // tricky part: for ccspline we pretend that it is a natural cubic spline, 
    // initialize it in a standard way, then copy internal data manually
    // (since endpoint derivatives are nowhere stored in the x,y data itself)
    gsl_spline* dst=gsl_spline_alloc(
        src->interp->type==gsl_interp_ccspline ? gsl_interp_cspline : src->interp->type, 
        src->size);
    gsl_spline_init(dst, src->x, src->y, src->size);
    if(src->interp->type==gsl_interp_ccspline) {
        dst->interp->type=gsl_interp_ccspline;
        cspline_state_t *dststate = (cspline_state_t *) dst->interp->state;
        cspline_state_t *srcstate = (cspline_state_t *) src->interp->state;
        for(size_t i=0; i<src->size; i++) {
            dststate->diag[i]=srcstate->diag[i];
            dststate->offdiag[i]=srcstate->offdiag[i];
            dststate->c[i]=srcstate->c[i];
            dststate->g[i]=srcstate->g[i];
        }
    }
    return dst;
}

/* ------------- class for performing penalized spline approximation --------------- */

CSplineApprox::CSplineApprox(const std::vector<double> &_xvalues, const std::vector<double> &_knots) :
  numDataPoints(_xvalues.size()), numKnots(_knots.size()), numBasisFnc(numKnots+2)
{
    knots=xvalues=yvalues=singValues=weightCoefs=zRHS=bsplineValues=NULL;
    bsplineMatrix=LMatrix=MMatrix=NULL;
    bsplineWorkspace=NULL;

    // first check for validity of input range
    mystatus=AS_INVRANGE;
    if(numDataPoints<=2 || numKnots<=2) return;
    for(size_t k=1; k<numKnots; k++)
        if(_knots[k-1]>=_knots[k]) return;  // knots must be in ascending order
    xvalues = gsl_vector_alloc(numDataPoints);
    knots = gsl_vector_alloc(numKnots);
    if(xvalues==NULL || knots==NULL) { 
        mystatus=AS_NOMEM;
        return;
    }
    for(size_t i=0; i<numDataPoints; i++)
        gsl_vector_set(xvalues, i, _xvalues[i]);
    for(size_t k=0; k<numKnots; k++)
        gsl_vector_set(knots, k, _knots[k]);
    if(gsl_vector_min(xvalues) < _knots.front() || gsl_vector_max(xvalues) > _knots.back()) 
        return;  // source data points must lie within spline definition region

    // next allocate b-splines and other matrices
    mystatus=AS_NOMEM;
    bsplineWorkspace = gsl_bspline_alloc(4, numKnots);
    bsplineValues= gsl_vector_alloc(numKnots+2);
    bsplineDerivWorkspace = gsl_bspline_deriv_alloc(4);
    bsplineDerivValues = gsl_matrix_alloc(numKnots+2, 3);
    bsplineMatrix= gsl_matrix_alloc(numDataPoints, numBasisFnc);  // matrix C_ip -- this is the largest chunk of memory to be used
    yvalues      = gsl_vector_alloc(numDataPoints);
    LMatrix      = gsl_matrix_alloc(numBasisFnc, numBasisFnc);    // lower triangular matrix L obtained by Cholesky decomposition of matrix A = C^T C
    weightCoefs  = gsl_vector_calloc(numBasisFnc);                // weight coefficients at basis functions, which are the unknowns in the linear system
    zRHS         = gsl_vector_alloc(numBasisFnc);                 // z = C^T y, RHS of the linear system
    MTz          = gsl_vector_alloc(numBasisFnc);
    tempv        = gsl_vector_alloc(numBasisFnc);
    if(bsplineWorkspace==NULL || bsplineValues==NULL || bsplineDerivWorkspace==NULL || 
        bsplineDerivValues==NULL || yvalues==NULL || bsplineMatrix==NULL || LMatrix==NULL || 
        weightCoefs==NULL || zRHS==NULL || MTz==NULL || tempv==NULL) 
        return;  // memory deallocation is done in the destructor
    ynorm2=gsl_nan();  // to indicate that no y-values have been loaded yet

    // initialize b-spline matrix C 
    gsl_bspline_knots(knots, bsplineWorkspace);
    for(size_t i=0; i<numDataPoints; i++) {
        gsl_bspline_eval(_xvalues[i], bsplineValues, bsplineWorkspace);
        for(size_t p=0; p<numBasisFnc; p++)
            gsl_matrix_set(bsplineMatrix, i, p, gsl_vector_get(bsplineValues, p));
    }

    // pre-compute matrix L
    gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, bsplineMatrix, bsplineMatrix, 0, LMatrix);
    if(gsl_linalg_cholesky_decomp(LMatrix) != GSL_SUCCESS) {
        mystatus=AS_SINGULAR;
        return;
    } 

    mystatus=AS_OK;
}

CSplineApprox::~CSplineApprox()
{
    gsl_bspline_free(bsplineWorkspace);
    gsl_bspline_deriv_free(bsplineDerivWorkspace);
    gsl_vector_free(bsplineValues);
    gsl_matrix_free(bsplineDerivValues);
    gsl_vector_free(xvalues);
    gsl_vector_free(yvalues);
    gsl_vector_free(knots);
    gsl_vector_free(weightCoefs);
    gsl_vector_free(zRHS);
    gsl_vector_free(MTz);
    gsl_vector_free(tempv);
    gsl_matrix_free(bsplineMatrix);
    gsl_matrix_free(LMatrix);
    if(MMatrix!=NULL) { // actually no need to check, it is done inside gsl as well; just to stress that this matrix may remain unallocated even in the normal course of events.
        gsl_matrix_free(MMatrix);
        gsl_vector_free(singValues);
    }
}

CSplineApprox::APPROXSTATUS CSplineApprox::loadyvalues(const std::vector<double> &_yvalues)
{
    if(mystatus != AS_OK && mystatus != AS_SINGULAR) return mystatus;  // if matrix is singular, still allow to load data, but fit will be performed with an alternative algorithm
    if(_yvalues.size() != numDataPoints) return AS_INVRANGE;
    ynorm2=0;
    for(size_t i=0; i<numDataPoints; i++)
    {
        gsl_vector_set(yvalues, i, _yvalues[i]);
        ynorm2 += pow_2(_yvalues[i]);
    }
    if(mystatus == AS_SINGULAR) return mystatus;
    // precompute z = C^T y
    gsl_blas_dgemv(CblasTrans, 1, bsplineMatrix, yvalues, 0, zRHS);
    return AS_OK;
}

/// convenience function returning values from band matrix or zero if indexes are outside the band
double getVal(const gsl_matrix* deriv, size_t row, size_t col)
{
    if(row<col || row>=col+3) return 0; 
    else return gsl_matrix_get(deriv, row-col, col);
}
void CSplineApprox::initRoughnessMatrix()
{
    if(MMatrix != NULL) return;   // don't do the work twice
    // init matrix with roughness penalty (integrals of product of second derivatives of basis functions)
    MMatrix = gsl_matrix_calloc(numBasisFnc, numBasisFnc);   // matrix R_pq
    singValues = gsl_vector_alloc(numBasisFnc);   // vector S
    gsl_matrix* tempm  = gsl_matrix_alloc(numBasisFnc, numBasisFnc);
    gsl_matrix* derivs = gsl_matrix_calloc(3, numKnots);
    if(MMatrix==NULL || singValues==NULL || tempm==NULL || derivs==NULL) { 
        mystatus=AS_NOMEM; return; }
    for(size_t k=0; k<numKnots; k++)
    {
        size_t istart, iend;
#if GSL_MAJOR_VERSION >= 2
        gsl_bspline_deriv_eval_nonzero(gsl_vector_get(knots, k), 2, bsplineDerivWorkspace->dB, &istart, &iend, bsplineWorkspace);
#else
        gsl_bspline_deriv_eval_nonzero(gsl_vector_get(knots, k), 2, bsplineDerivWorkspace->dB, &istart, &iend, bsplineWorkspace, bsplineDerivWorkspace);
#endif
        for(size_t b=0; b<3; b++)
            gsl_matrix_set(derivs, b, k, gsl_matrix_get(bsplineDerivWorkspace->dB, b+k-istart, 2));
    }
    for(size_t p=0; p<numBasisFnc; p++)
    {
        size_t kmin = p>3 ? p-3 : 0;
        size_t kmax = std::min<size_t>(p+3,knots->size-1);
        for(size_t q=p; q<std::min<size_t>(p+4,numBasisFnc); q++)
        {
            double result=0;
            for(size_t k=kmin; k<kmax; k++)
            {
                double x0 = gsl_vector_get(knots, k);
                double x1 = gsl_vector_get(knots, k+1);
                double Gp = getVal(derivs,p,k)*x1 - getVal(derivs,p,k+1)*x0;
                double Hp = getVal(derivs,p,k+1)  - getVal(derivs,p,k);
                double Gq = getVal(derivs,q,k)*x1 - getVal(derivs,q,k+1)*x0;
                double Hq = getVal(derivs,q,k+1)  - getVal(derivs,q,k);
                result += (Hp*Hq*(pow(x1,3.0)-pow(x0,3.0))/3.0 + (Gp*Hq+Gq*Hp)*(pow_2(x1)-pow_2(x0))/2.0 + Gp*Gq*(x1-x0)) / pow_2(x1-x0);
            }
            gsl_matrix_set(MMatrix, p, q, result);
            gsl_matrix_set(MMatrix, q, p, result);  // it is symmetric
        }
    }

    // now transform the roughness matrix R into more suitable form (so far MMatrix contains R)
    // obtain Q = L^{-1} R L^{-T}, where R is the roughness penalty matrix (store Q instead of R)
    gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1, LMatrix, MMatrix);
    gsl_blas_dtrsm(CblasRight, CblasLower, CblasTrans, CblasNonUnit, 1, LMatrix, MMatrix);   // now MMatrix contains Q = L^{-1} R L^(-T}

    // next decompose this Q via singular value decomposition: Q = U * diag(SV) * V^T
    gsl_linalg_SV_decomp(MMatrix, tempm, singValues, tempv);   // now MMatrix contains U, and workm contains V^T
    // Because Q was symmetric and positive definite, we expect that U=V, but don't actually check it.
    gsl_vector_set(singValues, numBasisFnc-2, 0);   // the smallest two singular values must be zero; set explicitly to avoid roundoff error
    gsl_vector_set(singValues, numBasisFnc-1, 0);

    // precompute M = L^{-T} U  which is used in computing basis weight coefs.
    gsl_blas_dtrsm(CblasLeft, CblasLower, CblasTrans, CblasNonUnit, 1, LMatrix, MMatrix);   // now M is finally in place
    // now the weight coefs for any lambda are given by  w = M (I + lambda*diag(singValues))^{-1} M^T  z

    gsl_matrix_free(tempm);
    gsl_matrix_free(derivs);
}

/// \cond INTERNAL_DOCS
// auxiliary structure used to pass data to non-member functions
struct CFittingParam {
    const size_t numDataPoints;
    const size_t numBasisFnc;
    const gsl_matrix* LMatrix;    // lower triangular matrix from Cholesky decomposition of basis matrix
    const gsl_vector* zRHS;       // RHS of linear system
    const gsl_matrix* MMatrix;    // matrix and
    const gsl_vector* singValues; // vector of decomposition of roughness matrix
    const gsl_vector* MTz;        // precomputed M^T z
    double ynorm2;                // |y|^2
    double targetAIC;             // value of AIC for root-finder
    gsl_vector* temp;             // storage for internal calculations
    gsl_vector* weightCoefs;      // the solution of linear system
    CFittingParam(const size_t _numDataPoints, const size_t _numBasisFnc) : 
      numDataPoints(_numDataPoints), numBasisFnc(_numBasisFnc) {};
};

// non-member function to obtain solution of linear system with lambda>0
void computeFitSolution(double lambda, CFittingParam* par)
{
    if(lambda==0) // simple case, no need to use roughness penalty matrix
        gsl_linalg_cholesky_solve(par->LMatrix, par->zRHS, par->weightCoefs);
    else {
        for(size_t p=0; p<par->numBasisFnc; p++)
            gsl_vector_set(par->temp, p, gsl_vector_get(par->MTz, p)/(1+lambda*gsl_vector_get(par->singValues, p)));
        gsl_blas_dgemv(CblasNoTrans, 1, par->MMatrix, par->temp, 0, par->weightCoefs);
    }
}

void computeRMSandEDF(double lambda, CFittingParam* par, double* rmserror, double* edf)
{
    gsl_vector_memcpy(par->temp, par->weightCoefs);
    gsl_blas_dtrmv(CblasLower, CblasTrans, CblasNonUnit, par->LMatrix, par->temp);
    double wTz;
    gsl_blas_ddot(par->weightCoefs, par->zRHS, &wTz);
    *rmserror = (par->ynorm2 - 2*wTz + pow_2(gsl_blas_dnrm2(par->temp))) / par->numDataPoints;
    // equivalent degrees of freedom
    *edf = 0;
    if(!gsl_finite(lambda))
        *edf = 2;
    else if(lambda>0) 
        for(size_t c=0; c<par->numBasisFnc; c++)
            *edf += 1/(1+lambda*gsl_vector_get(par->singValues, c));
    else
        *edf = static_cast<double>(par->numBasisFnc);
}

double computeAIC(double lambda, void* params)
{
    CFittingParam* par = static_cast<CFittingParam*>(params);
    double rmserror, edf;
    computeFitSolution(lambda, par);
    computeRMSandEDF(lambda, par, &rmserror, &edf);
    double AIC = log(rmserror) + 2*edf/(par->numDataPoints-edf-1);
    return AIC - par->targetAIC;
}
/// \endcond

void CSplineApprox::computeYvalues(std::vector<double> *splineValues) const
{
    splineValues->assign(numKnots+2, 0);
    for(size_t k=1; k<numKnots-1; k++)  // loop over interior nodes
    {
        gsl_bspline_eval(gsl_vector_get(knots, k), bsplineValues, bsplineWorkspace);
        double value=0;
        for(size_t p=0; p<numKnots+2; p++)
            value += gsl_vector_get(bsplineValues, p) * gsl_vector_get(weightCoefs, p);
        splineValues->at(k) = value;
    }
    for(size_t k=0; k<numKnots; k+=numKnots-1)  // two endpoints
    {
#if GSL_MAJOR_VERSION >= 2
        gsl_bspline_deriv_eval(gsl_vector_get(knots, k), 1, bsplineDerivValues, bsplineWorkspace);
#else
        gsl_bspline_deriv_eval(gsl_vector_get(knots, k), 1, bsplineDerivValues, bsplineWorkspace, bsplineDerivWorkspace);
#endif
        double value=0, deriv=0;
        for(size_t p=0; p<numKnots+2; p++)
        {
            value += gsl_matrix_get(bsplineDerivValues, p, 0) * gsl_vector_get(weightCoefs, p);
            deriv += gsl_matrix_get(bsplineDerivValues, p, 1) * gsl_vector_get(weightCoefs, p);
        }
        splineValues->at(k) = value;
        splineValues->at(k==0 ? numKnots : numKnots+1) = deriv;   // store endpoint derivatives in two additional elements of Yvalues vector
    }
}

CSplineApprox::APPROXSTATUS CSplineApprox::fitData(const double lambda, std::vector<double> *splineValues, double *rmserror, double* edf)
{
    if(splineValues==NULL) return AS_NODATA;   // should pass a valid pointer
    splineValues->clear();
    if(gsl_isnan(ynorm2)) return AS_NODATA;    // wasn't initialized properly or data wasn't loaded yet
    if(mystatus == AS_SINGULAR)  // in this case just pick up nearest y-values for each x-knot
        return fitDataSingular(splineValues, rmserror, edf);
    if(mystatus != AS_OK) return mystatus;
    CFittingParam par(numDataPoints, numBasisFnc);
    par.LMatrix=LMatrix;
    par.zRHS=zRHS;
    par.weightCoefs=weightCoefs;
    par.ynorm2=ynorm2;
    par.temp=tempv;
    if(lambda!=0) {
        initRoughnessMatrix();   // perform smoothed fit
        par.MTz = MTz;
        par.MMatrix=MMatrix;
        par.singValues=singValues;
        gsl_blas_dgemv(CblasTrans, 1, MMatrix, zRHS, 0, MTz);
    }
    computeFitSolution(lambda, &par);
    computeYvalues(splineValues);
    if(rmserror!=NULL || edf!=NULL)
    {
        double _rmserror, _edf;
        computeRMSandEDF(lambda, &par, &_rmserror, &_edf);
        if(rmserror!=NULL) *rmserror=_rmserror;
        if(edf!=NULL) *edf=_edf;
    }
    return AS_OK;
}

CSplineApprox::APPROXSTATUS CSplineApprox::fitDataOptimal(std::vector<double> *splineValues, double *rmserror, double* edf, double* lambda)
{
    if(splineValues==NULL) return AS_NODATA;   // should pass a valid pointer
    splineValues->clear();
    if(gsl_isnan(ynorm2)) return AS_NODATA;    // wasn't initialized properly or data wasn't loaded yet
    if(mystatus == AS_SINGULAR) {  // in this case just pick up nearest y-values for each x-knot
        if(lambda!=NULL) *lambda=0;
        return fitDataSingular(splineValues, rmserror, edf);
    }
    if(mystatus != AS_OK) return mystatus;
    CFittingParam par(numDataPoints, numBasisFnc);
    par.LMatrix=LMatrix;
    par.zRHS=zRHS;
    par.weightCoefs=weightCoefs;
    par.ynorm2=ynorm2;
    par.temp=tempv;
    initRoughnessMatrix();   // perform smoothed fit
    par.MTz = MTz;
    par.MMatrix=MMatrix;
    par.singValues=singValues;
    gsl_blas_dgemv(CblasTrans, 1, MMatrix, zRHS, 0, MTz);  // precompute M^T z
    gsl_function F;
    F.function=&computeAIC;
    F.params=&par;
    par.targetAIC=0;
    double lamopt=1, lamlow=0, lamupp;
    double AIC0 = computeAIC(lamlow, &par), AIC;
    while(lamopt>1e-15 && (AIC=computeAIC(lamopt, &par)) > AIC0) lamopt/=8;  // find initial guess for minimum
    lamupp=lamopt*2;
    while(lamupp < 1e10 && computeAIC(lamupp, &par) < AIC) lamupp*=8;   // safely initialize minimizer to enclose minimum
    lamopt = findMin(&F, lamlow, lamupp, lamopt, 0.1);
    // lambda is the optimal value of smoothing parameter
    computeYvalues(splineValues);
    if(rmserror!=NULL || edf!=NULL)
    {
        double _rmserror, _edf;
        computeRMSandEDF(lamopt, &par, &_rmserror, &_edf);
        if(rmserror!=NULL) *rmserror=_rmserror;
        if(edf!=NULL) *edf=_edf;
    }
    if(lambda!=NULL) *lambda=lamopt;
    return AS_OK;
}

CSplineApprox::APPROXSTATUS CSplineApprox::fitDataOversmooth(const double deltaAIC, std::vector<double> *splineValues, double *rmserror, double* edf, double *lambda)
{
    if(deltaAIC==0) {
        if(lambda!=NULL) *lambda=0;  // indicates no smoothing
        return fitData(0, splineValues, rmserror, edf);
    }
    if(splineValues==NULL) return AS_NODATA;   // should pass a valid pointer
    splineValues->clear();
    if(gsl_isnan(ynorm2)) return AS_NODATA;    // wasn't initialized properly or data wasn't loaded yet
    if(mystatus == AS_SINGULAR) {  // in this case just pick up nearest y-values for each x-knot
        if(lambda!=NULL) *lambda=0;
        return fitDataSingular(splineValues, rmserror, edf);
    }
    if(mystatus != AS_OK) return mystatus;
    CFittingParam par(numDataPoints, numBasisFnc);
    par.LMatrix=LMatrix;
    par.zRHS=zRHS;
    par.weightCoefs=weightCoefs;
    par.ynorm2=ynorm2;
    par.temp=tempv;
    initRoughnessMatrix();   // perform smoothed fit
    par.MTz = MTz;
    par.MMatrix=MMatrix;
    par.singValues=singValues;
    gsl_blas_dgemv(CblasTrans, 1, MMatrix, zRHS, 0, MTz);  // precompute M^T z
    gsl_function F;
    F.function=&computeAIC;
    F.params=&par;
    par.targetAIC=0;
    double lamopt=0, lamlow=0, lamupp=1;
    double AIC0 = computeAIC(lamlow, &par);
    par.targetAIC=AIC0+deltaAIC;   // allow for somewhat higher AIC value, to smooth more than minimum necessary amount
    while(lamupp < 1e10 && computeAIC(lamupp, &par) < 0) lamupp*=8;   // safely initialize minimizer to enclose the root
    if(lamupp < 1e10) 
        lamopt=findRoot(&F, lamlow, lamupp, gsl_posinf());
    else 
        lamopt=gsl_posinf();  // basically means fitting with a linear regression
    // lambda is the optimal value of smoothing parameter
    computeYvalues(splineValues);
    if(rmserror!=NULL || edf!=NULL)
    {
        double _rmserror, _edf;
        computeRMSandEDF(lamopt, &par, &_rmserror, &_edf);
        if(rmserror!=NULL) *rmserror=_rmserror;
        if(edf!=NULL) *edf=_edf;
    }
    if(lambda!=NULL) *lambda=lamopt;
    return AS_OK;
}

CSplineApprox::APPROXSTATUS CSplineApprox::fitDataSingular(std::vector<double> *splineValues, double *rmserror, double* edf)
{
    gsl_matrix* covarMatrix  = gsl_matrix_alloc(numBasisFnc, numBasisFnc);
    gsl_multifit_linear_workspace* fitWorkspace = gsl_multifit_linear_alloc(numDataPoints, numBasisFnc);
    double chisq;
    if(covarMatrix==NULL || fitWorkspace==NULL) 
        return AS_NOMEM;
#if GSL_MAJOR_VERSION >= 2
    gsl_multifit_linear(bsplineMatrix, yvalues, weightCoefs, covarMatrix, &chisq, fitWorkspace);
#else
    size_t rank;
    gsl_multifit_linear_svd(bsplineMatrix, yvalues, 1e-8, &rank, weightCoefs, covarMatrix, &chisq, fitWorkspace);
#endif
    computeYvalues(splineValues);
    if(rmserror!=NULL) *rmserror = chisq/numDataPoints;
    if(edf!=NULL) *edf=static_cast<double>(numBasisFnc);
    gsl_multifit_linear_free(fitWorkspace);
    gsl_matrix_free(covarMatrix);
    return AS_SINGULAR;
}

CSplineApprox::APPROXSTATUS CSplineApprox::computeRegressionAtPoints(const std::vector<double> &xpoints, std::vector<double> *ypoints) const
{
    if(mystatus!=AS_OK && mystatus!=AS_SINGULAR) return mystatus;
    if(ypoints==NULL) return AS_NODATA;   // should pass a valid pointer
    ypoints->assign(xpoints.size(), gsl_nan());  // default value for nodes outside the definition range
    for(size_t i=0; i<xpoints.size(); i++)  // loop over interior nodes
        if(xpoints[i]>=gsl_vector_get(knots, 0) && xpoints[i]<=gsl_vector_get(knots, numKnots-1))
        {
            gsl_bspline_eval(xpoints[i], bsplineValues, bsplineWorkspace);
            double value=0;
            for(size_t p=0; p<numKnots+2; p++)
                value += gsl_vector_get(bsplineValues, p) * gsl_vector_get(weightCoefs, p);
            ypoints->at(i) = value;
        }
    return AS_OK;
}

}  // namespace
