#include <Python.h>
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#include <numpy/arrayobject.h>
#include <fstream>
#include "configfile.h"
#include "managepotential.h"
#include "potential.h"
/// \cond INTERNAL_DOCS

typedef struct {
    PyObject_HEAD
    const smile::CPotential* pot;
} SmilePotObject;

/*static PyObject* SmilePot_new(PyTypeObject *type, PyObject*, PyObject*)
{
    SmilePotObject *self = (SmilePotObject*)type->tp_alloc(type, 0);
    if(self) self->pot=NULL;
    printf("Allocated a new SmilePot object\n");
    return (PyObject*)self;
}*/

static void SmilePot_dealloc(SmilePotObject* self)
{
    if(self->pot) {
#ifdef DEBUGPRINT
        printf("Deleted an instance of %s potential\n", self->pot->PotentialName().c_str());
#endif
        delete self->pot;
    }
#ifdef DEBUGPRINT
    else printf("Deleted empty SmilePot\n");
#endif
    self->ob_type->tp_free((PyObject*)self);
}

bool endsWith(const std::string& str, const std::string& end)
{
    return end.size() <= str.size() && str.find(end, str.size()-end.size())!=str.npos;
}

int listOrArraySize(PyObject* a)
{
    if(a && PyList_Check(a))
        return static_cast<int>(PyList_Size(a));
    if(a && PyArray_Check((PyArrayObject *)a) &&
        PyArray_ISFLOAT(const_cast<const PyArrayObject*>((PyArrayObject*)a)) &&
        PyArray_NDIM(const_cast<const PyArrayObject*>((PyArrayObject*)a))==1 )
        return static_cast<int>(*PyArray_DIMS((PyArrayObject*)a));
    return -1;
}

double listOrArrayElem(PyObject* a, int index)
{
    if(PyList_Check(a))
        return PyFloat_AsDouble(PyList_GetItem(a, index));
    else {
        npy_intp i=static_cast<npy_intp>(index);
        void* data=PyArray_GetPtr((PyArrayObject*)a, &i);
        if(PyArray_TYPE((PyArrayObject*)a)==NPY_FLOAT) return *((float*)data);
        if(PyArray_TYPE((PyArrayObject*)a)==NPY_DOUBLE) return *((double*)data);
        PyErr_SetString(PyExc_ValueError, "Unknown data type in array");
        return 0;
    }
}

const char* class_docstring = 
    "SmilePot is a class that represents a wide range of gravitational potentials\n"
    "There are a number of possible named arguments for the constructor:\n"
    "    file='string' - the name of ini file, potential coefficients file, or N-body snapshot file\n"
    "    type='string' - potential type, like 'Dehnen', 'Ferrers', or potential expansion type, like 'BSE', 'Spline'\n"
    "    density='string' - density model for potential expansion, like 'Dehnen', 'Plummer', 'MGE'\n"
    "    symmetry='string' - assumed symmetry for potential expansion constructed from an N-body snapshot\n"
    "    pointx,pointy,pointz,pointm=[list or array] - four parameters that provide particle coordinates and masses to construct a potential expansion (instead of an external file)\n"
    "    mass,scalerad,scalerad2,gamma,sersicindex,q,p=float - parameters of analytic potential or density models\n"
    "    ncoefs_radial,ncoefs_angular,ncoefs_vertical=int - parameters controlling the order of potential expansion or its grid dimensions\n"
    "    alpha,rmax=float - shape parameter and radial extent for BSE and BSECompact expansions, correspondingly\n"
    "    splinermin,splinermax,splinezmin,splinezmax=float - grid extent for Spline and CylSpline expansions\n"
    "    splinesmoothfactor=float - amount of smoothing applied for Spline potential constructed from an Nbody snapshot\n"
    "    treecodeeps,treecodetheta=float - parameters of tree-code Nbody potential\n"
    "A full description of the arguments may be found in readme_smilepot.pdf\n"
    "Required parameters are either type or file (or both)";
static int SmilePot_init(PyObject* self, PyObject* args, PyObject* namedargs)
{
    smile::CConfigPotential cfg=smile::defaultConfigPotential;
    static const char* keywords[] = {"file","type","density","symmetry",
        "mass","scalerad","scalerad2","q","p","gamma","sersicindex",
        "ncoefs_radial","ncoefs_angular","ncoefs_vertical","alpha","rmax",
        "splinesmoothfactor","splinermin","splinermax","splinezmin","splinezmax",
        "treecodeeps","treecodetheta", "pointx","pointy","pointz","pointm",NULL};
    const char* file="";
    const char* type="";
    const char* density="";
    const char* symmetry="";
    PyObject* pointx=NULL;
    PyObject* pointy=NULL;
    PyObject* pointz=NULL;
    PyObject* pointm=NULL;
    if(!PyArg_ParseTupleAndKeywords(args, namedargs, "|ssssdddddddiiidddddddddOOOO", const_cast<char **>(keywords),
        &file, &type, &density, &symmetry,
        &(cfg.mass), &(cfg.scalerad), &(cfg.scalerad2), &(cfg.q), &(cfg.p), &(cfg.Gamma), &(cfg.sersicIndex),
        &(cfg.Ncoefs_radial), &(cfg.Ncoefs_angular), &(cfg.Ncoefs_vertical), &(cfg.Alpha), &(cfg.Rmax),
        &(cfg.splineSmoothFactor), &(cfg.splineRMin), &(cfg.splineRMax), &(cfg.splineZMin), &(cfg.splineZMax),
        &(cfg.treecodeEps), &(cfg.treecodeTheta), &pointx, &pointy, &pointz, &pointm)) {
        PyErr_SetString(PyExc_ValueError, "Incorrect parameters passed to the SmilePot constructor");
        return -1;
    }
    if(file[0]!=0) {
        std::string filename(file);   // could be an ini file, coefs file, or Nbody file
        std::ifstream strm(file);  // check its existence
        if(!strm) {
            PyErr_SetString(PyExc_ValueError, ("Cannot read from "+filename).c_str());
            return -1;
        }
        strm.close();
        if(endsWith(filename, ".ini")) {  // supposedly ini file
            smile::CConfigFile iniFile(filename);
            smile::readConfigPotential(iniFile, &cfg);  // there's no way we can tell that this file actually was an ini file...
        } else if(endsWith(filename, ".coef_bse")) {
            cfg.PotentialType=smile::CDensity::PT_BSE;
            cfg.DensityType=smile::CDensity::PT_COEFS;
            cfg.NbodyFile=filename;
        } else if(endsWith(filename, ".coef_bsec")) {
            cfg.PotentialType=smile::CDensity::PT_BSECOMPACT;
            cfg.DensityType=smile::CDensity::PT_COEFS;
            cfg.NbodyFile=filename;
        } else if(endsWith(filename, ".coef_spl")) {
            cfg.PotentialType=smile::CDensity::PT_SPLINE;
            cfg.DensityType=smile::CDensity::PT_COEFS;
            cfg.NbodyFile=filename;
        } else if(endsWith(filename, ".coef_cyl")) {
            cfg.PotentialType=smile::CDensity::PT_CYLSPLINE;
            cfg.DensityType=smile::CDensity::PT_COEFS;
            cfg.NbodyFile=filename;
        } else if(endsWith(filename, ".mass")) {
            cfg.PotentialType=smile::CDensity::PT_SPHERICAL;
            cfg.DensityType=smile::CDensity::PT_COEFS;
            cfg.NbodyFile=filename;
        } else {  // not an ini or coefs file -- must be a file with Nbody snapshot
            cfg.DensityType=smile::CDensity::PT_NB;  // PotentialType is left unspecified
            cfg.NbodyFile=filename;
        }
    } else  // densitytype may be set by a parameter passed to the constructor
        cfg.DensityType=smile::getDensityTypeByName(density);
    if(type[0]!=0) {
        cfg.PotentialType=smile::getPotentialTypeByName(type);
    }
    if(cfg.PotentialType==smile::CPotential::PT_UNKNOWN) {
        if(type[0]==0)
            PyErr_SetString(PyExc_ValueError, "Should provide type='...' or file='...' parameter");
        else
            PyErr_SetString(PyExc_ValueError, "Incorrect type='...' parameter");
        return -1;
    }
    cfg.SymmetryType=smile::getSymmetryTypeByName(symmetry);
    const smile::CPotential* pot=NULL;
    if(pointx!=NULL || pointy!=NULL || pointz!=NULL || pointm!=NULL) {  // a list of particles was provided
        if(pointx==NULL || pointy==NULL || pointz==NULL || pointm==NULL) {
            PyErr_SetString(PyExc_ValueError, "Should provide four parameters: pointx, pointy, pointz, pointm");
            return -1;
        }
        if( file[0]!=0 ) {
            PyErr_SetString(PyExc_ValueError, "Cannot provide both points and filename");
            return -1;
        }
        if( cfg.PotentialType!=smile::CDensity::PT_BSE &&
            cfg.PotentialType!=smile::CDensity::PT_BSECOMPACT &&
            cfg.PotentialType!=smile::CDensity::PT_SPLINE &&
            cfg.PotentialType!=smile::CDensity::PT_CYLSPLINE &&
            cfg.PotentialType!=smile::CDensity::PT_SPHERICAL &&
            cfg.PotentialType!=smile::CDensity::PT_NB ) {
            PyErr_SetString(PyExc_ValueError, "Potential should be of an expansion type");
            return -1;
        }
        // check type
        int numpt=0;
        if( ((numpt=listOrArraySize(pointx))<=0) || (numpt!=listOrArraySize(pointy)) ||
            (numpt!=listOrArraySize(pointz))     || (numpt!=listOrArraySize(pointm)) ) {
            PyErr_SetString(PyExc_ValueError, "Four parameters must be lists or arrays of equal length: pointx, pointy, pointz, pointm");
            return -1;
        }
        smile::CPointMassSet<double> points;
        for(int i=0; i<numpt; i++) {
            double x=listOrArrayElem(pointx, i);
            double y=listOrArrayElem(pointy, i);
            double z=listOrArrayElem(pointz, i);
            double m=listOrArrayElem(pointm, i);
            points.data.push_back(smile::CPointMassSet<double>::ElemType(smile::CPosVelPoint<double>(x,y,z,0,0,0), m));
        }
        if(PyErr_Occurred()) {  // numerical conversion error
            return -1;
        }
        pot=smile::createPotentialFromPoints(&cfg, &points);
    } else {
        pot=smile::createPotential(&cfg);  // attempt to create potential from configuration parameters
    }
    if(!pot) {
        PyErr_SetString(PyExc_ValueError, "Incorrect parameters for the potential");
        return -1;
    }
#ifdef DEBUGPRINT
    printf("Created an instance of %s potential\n", pot->PotentialName().c_str());
#endif
    if(((SmilePotObject*)self)->pot) {
#ifdef DEBUGPRINT
        printf("Deleted previous instance of %s potential\n", ((SmilePotObject*)self)->pot->PotentialName().c_str());
#endif
        delete ((SmilePotObject*)self)->pot;
    }
    ((SmilePotObject*)self)->pot=pot;
    return 0;
}

bool SmilePot_isCorrect(PyObject* self)
{
    if(self==NULL) {
        PyErr_SetString(PyExc_ValueError, "Should be called as method of SmilePot object");
        return false;
    }
    if(((SmilePotObject*)self)->pot==NULL) {
        PyErr_SetString(PyExc_ValueError, "Potential is not initialized properly");
        return false;
    }
    return true;
}

static PyObject* SmilePot_potential(PyObject* self, PyObject* args)
{
    double x=0, y=0, z=0;
    if(!SmilePot_isCorrect(self) || !PyArg_ParseTuple(args, "ddd", &x, &y, &z))
        return NULL;
    return Py_BuildValue("d", ((SmilePotObject*)self)->pot->Phi(x, y, z));
}

static PyObject* SmilePot_density(PyObject* self, PyObject* args)
{
    double x=0, y=0, z=0;
    if(!SmilePot_isCorrect(self) || !PyArg_ParseTuple(args, "ddd", &x, &y, &z))
        return NULL;
    return Py_BuildValue("d", ((SmilePotObject*)self)->pot->Rho(x, y, z));
}

static PyObject* SmilePot_force(PyObject* self, PyObject* args)
{
    double xyz[3], force[3];
    if(!SmilePot_isCorrect(self) || !PyArg_ParseTuple(args, "ddd", &xyz[0], &xyz[1], &xyz[2]))
        return NULL;
    ((SmilePotObject*)self)->pot->Force(xyz, 0, force);
    return Py_BuildValue("ddd", force[0], force[1], force[2]);
}

static PyObject* SmilePot_force_deriv(PyObject* self, PyObject* args)
{
    double xyz[3], force[3], deriv[6];
    if(!SmilePot_isCorrect(self) || !PyArg_ParseTuple(args, "ddd", &xyz[0], &xyz[1], &xyz[2]))
        return NULL;
    ((SmilePotObject*)self)->pot->Force(xyz, 0, force, deriv);
    return Py_BuildValue("(ddd)(dddddd)", force[0], force[1], force[2],
        deriv[0], deriv[1], deriv[2], deriv[3], deriv[4], deriv[5]);
}

static PyObject* SmilePot_name(PyObject* self)
{
    if(!SmilePot_isCorrect(self))
        return NULL;
    return Py_BuildValue("s", ((SmilePotObject*)self)->pot->PotentialName());
}

static PyObject* SmilePot_export(PyObject* self, PyObject* args)
{
    const char* filename=NULL;
    if(!SmilePot_isCorrect(self) || !PyArg_ParseTuple(args, "s", &filename))
        return NULL;
    smile::CDensity::POTENTIALTYPE t=((SmilePotObject*)self)->pot->PotentialType();
    if (t!=smile::CDensity::PT_BSE &&
        t!=smile::CDensity::PT_BSECOMPACT &&
        t!=smile::CDensity::PT_SPLINE &&
        t!=smile::CDensity::PT_CYLSPLINE &&
        t!=smile::CDensity::PT_SPHERICAL) {
        PyErr_SetString(PyExc_ValueError, "Potential is not of an expansion type");
        return NULL;
    }
    bool result=smile::writePotential(filename, ((SmilePotObject*)self)->pot);
    if(result) {
        Py_INCREF(Py_None);
        return Py_None;
    } else {
        PyErr_SetString(PyExc_ValueError, "Cannot write file");
        return NULL;
    }
}

static PyMethodDef SmilePot_methods[] = {
    { "name", (PyCFunction)SmilePot_name, METH_NOARGS, 
      "Return the name of the potential\nNo arguments\nReturns: string" },
    { "potential", SmilePot_potential, METH_VARARGS, 
      "Compute potential at given point\nArguments: x,y,z (float)\nReturns: float" },
    { "density", SmilePot_density, METH_VARARGS, 
      "Compute density at given point\nArguments: x,y,z (float)\nReturns: float" },
    { "force", SmilePot_force, METH_VARARGS, 
      "Compute force at given point\nArguments: x,y,z (float)\nReturns: float[3] - x,y,z components of force" },
    { "force_deriv", SmilePot_force_deriv, METH_VARARGS, 
      "Compute force and its derivatives at given point\nArguments: x,y,z (float)\nReturns: (float[3],float[6]) - "
      "x,y,z components of force, and the matrix of force derivatives stored as dFx/dx,dFy/dy,dFz/dz,dFx/dy,dFy/dz,dFz/dx" },
    { "export", (PyCFunction)SmilePot_export, METH_VARARGS, 
      "Export potential expansion coefficients to a text file\nArguments: filename (string)\nReturns: none" },
    { NULL, NULL, 0, NULL }
};

static PyTypeObject SmilePotType = {
    PyObject_HEAD_INIT(NULL)
    0, "py_smilepot.SmilePot",
    sizeof(SmilePotObject), 0, (destructor)SmilePot_dealloc,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, SmilePot_name, 0, 0, 0,
    Py_TPFLAGS_DEFAULT, class_docstring, 
    0, 0, 0, 0, 0, 0, SmilePot_methods, 0, 0, 0, 0, 0, 0, 0,
    SmilePot_init, 0, /*SmilePot_new*/
};

static PyMethodDef py_smilepot_methods[] = {
    {NULL}
};

PyMODINIT_FUNC
initpy_smilepot(void)
{
    SmilePotType.tp_new = PyType_GenericNew;
    if (PyType_Ready(&SmilePotType) < 0) return;
    PyObject* mod = Py_InitModule("py_smilepot", py_smilepot_methods);
    if(!mod) return;
    Py_INCREF(&SmilePotType);
    PyModule_AddObject(mod, "SmilePot", (PyObject *)&SmilePotType);
    import_array();
}
/// \endcond