#include "smilepot.h"
#include <gsl/gsl_errno.h>
#include "configfile.h"
#include "managepotential.h"
#include "potential.h"
#include "utils.h"
extern "C" {
    // prevent bailing out on arithmetic errors: set up a new error handler upon library initialization
    gsl_error_handler_t* setupErrorHandler() {
        //smile::my_message("smilepot","setup GSL error handler");
        return gsl_set_error_handler(&(smile::gsl_verbose_error_handler));
    }
    gsl_error_handler_t* dummy_var = setupErrorHandler();

    const smilePotential* smilepot_create(const char *filename)
    {
        if(filename==NULL || filename[0]==0)
            return NULL;
        smile::CConfigPotential cfg=smile::defaultConfigPotential;
        if(smile::ends_with_str(filename, ".ini")) {
            smile::CConfigFile cfgFile(filename);
            smile::readConfigPotential(cfgFile, &cfg);
        } else {
            cfg.PotentialType=smile::coefFileType(filename);
            if(cfg.PotentialType==smile::CDensity::PT_UNKNOWN)  // none of the known file extensions
                return NULL;
            cfg.DensityType=smile::CDensity::PT_COEFS;
            cfg.NbodyFile=std::string(filename);
        }
        return smile::createPotential(&cfg);  // NULL on error
    }

    const smilePotential* smilepot_create_params(const int numParams, const char* const* params)
    {
        if(numParams<=0 || params==NULL || params[0]==NULL)
            return NULL;
        smile::CConfigPotential cfg=smile::defaultConfigPotential;
        smile::CConfigFile cfgFile("Potential", numParams, params);
        smile::readConfigPotential(cfgFile, &cfg);
        return smile::createPotential(&cfg);  // NULL on error
    }

    const smilePotential* smilepot_create_from_particles(const int numParams, const char* const* params,
        const int numPoints, const double x[], const double y[], const double z[], const double mass[])
    {
        if( numParams<=0 || params==NULL || params[0]==NULL ||
            numPoints<=0 || x==NULL || y==NULL || z==NULL || mass==NULL)
            return NULL;
        smile::CConfigPotential cfg=smile::defaultConfigPotential;
        smile::CConfigFile cfgFile("Potential", numParams, params);
        smile::readConfigPotential(cfgFile, &cfg);
        smile::CPointMassSet<double> points;
        points.data.resize(numPoints);
        for(int i=0; i<numPoints; i++) {
            points[i].first.Pos[0]=x[i];
            points[i].first.Pos[1]=y[i];
            points[i].first.Pos[2]=z[i];
            points[i].second=mass[i];
        }
        return smile::createPotentialFromPoints(&cfg, &points);
    }

    const smilePotential* smilepot_copy(const smilePotential* pot)
    {
        if(pot==NULL) return NULL;
        return pot->clone();
    }

    void smilepot_delete(const smilePotential* pot)
    {
        if(pot!=NULL) delete pot;
    }

    const char* smilepot_name(const smilePotential* pot)
    {
        if(pot==NULL) return "";
        return pot->PotentialName();
    }

    double smilepot_density(const smilePotential* pot, const double xyz[3])
    {
        if(pot==NULL) return 0;
        return pot->Rho(xyz[0], xyz[1], xyz[2]);
    }

    double smilepot_potential(const smilePotential* pot, const double xyz[3])
    {
        if(pot==NULL) return 0;
        return pot->Phi(xyz[0], xyz[1], xyz[2]);
    }

    void smilepot_force(const smilePotential* pot, const double xyz[3], double force[])
    {
        force[0]=force[1]=force[2]=0;
        if(pot==NULL) return;
        pot->Force(xyz, 0, force, NULL);
    }

    void smilepot_force_deriv(const smilePotential* pot, const double xyz[3], double force[], double forceDeriv[])
    {
        force[0]=force[1]=force[2]=0;
        forceDeriv[0]=forceDeriv[1]=forceDeriv[2]=forceDeriv[3]=forceDeriv[4]=forceDeriv[5]=0;
        if(pot==NULL) return;
        pot->Force(xyz, 0, force, forceDeriv);
    }

    int smilepot_export(const smilePotential* pot, const char* filename)
    {
        if(pot==NULL) return 0;
        return smile::writePotential(filename, pot);
    }

}  // extern
