/** \file    snaporbits.cpp
    \brief   Tool: Perform orbit analysis on trajectories of particles from an N-body snapshot
    \author  EV
    \date    2012-2015

This program demonstrates how to use SMILE orbit analysis machinery on trajectories from an N-body 
simulation. It loads NEMO snapshot file containing particles at multiple moments of time and 
treats trajectories of each particle as a separate orbit, performing the same kind of orbit 
analysis as for loading an orbit from a text file in SMILE GUI. 

To conduct orbit analysis, one needs to provide the potential. It may be done either by specifying 
a potential coefs file or simply by using the first snapshot in the input file as the source data 
for potential approximation (using BSE or Spline potential expansions). 
In this latter case, several additional parameters may be given for initializing this potential 
approximation (number of radial and angular coefficients, symmetry and potential type), but they 
are not very important, since the potential is only used to compute the orbit energy and the 
natural orbital period, which serves as the unit of frequency. The potential coefficients are 
stored in a text file to simplify latter use (for instance, when loading the orbit library file 
to SMILE, one may provide this text file for the potential model). In addition, a point mass 
(central black hole) may be added to the potential. Beware that initializing the potential from 
the 1st snapshot itself or from another snapshot file will only account for the central point mass 
if it is precisely at origin; otherwise the potential will not be initialized properly. 
On the other hand, when loading the potential coefs from a text file, this additional central mass 
may need to be provided manually since it is not stored in the coefs file.

Once the entire snapshot file has been read, the orbit analysis is performed on each particle 
trajectory, and the resulting orbit library is saved in the output text file. 

Note that all particles at all times are loaded to memory; to reduce memory demand, one may need 
to filter the input snapshot by standard NEMO tools such as 
`snapmask input_file - select=0:1000 | snaporbits - out=output_library pot=input_file`
(note that in this case we must provide the original input file containing all particles that 
contribute to the potential, not only the subset of particles to be processed).

**/
#include <vector>
#include <iostream>
#include <io_nemo.h>
#include <gsl/gsl_errno.h>
#include "configfile.h"
#include "iodata.h"
#include "managepotential.h"
#include "orbitan.h"
#include "orbitint.h"
#include "orbitlib.h"
#include "potential.h"
#include "utils.h"

int main(int argc,char *argv[])
{
  if(argc<=1) {
    std::cerr << 
    "snaporbits: perform orbit analysis for orbits from Nbody snapshots\n"
    "Usage: snaporbits in=... out=... [optional potential parameters as in smilepot library]\n"
    "VERSION 1.1  Jan 2015 EV\n";
    return 0;
  }
  gsl_set_error_handler(&smile::gsl_verbose_error_handler);
  smile::CConfigPotential cfg=smile::defaultConfigPotential;
  smile::CConfigFile cfgFile("Potential", argc, argv);
  smile::readConfigPotential(cfgFile, &cfg);  // parse command-line arguments into a virtual ini file
  std::string infile  = cfgFile.getValueString("Potential/in");  // two additional parameters beyond those defining the potential
  std::string outfile = cfgFile.getValueString("Potential/out");
  if(infile.empty()) {
    std::cerr << "Must provide input snapshot file in=...\n";
    return 0;
  }
  if(outfile.empty()) {
    std::cerr << "Must provide output orbit library file out=...\n";
    return 0;
  }
  if(smile::ends_with_str(cfg.NbodyFile, ".ini")) {  // a real ini file was provided
    smile::CConfigFile cfgFile1(cfg.NbodyFile);
    smile::readConfigPotential(cfgFile1, &cfg);
  }
  if(cfg.DensityType==smile::CDensity::PT_UNKNOWN)  // not specified -- take from input snapshot
    cfg.DensityType=smile::CDensity::PT_NB;
  if(cfg.PotentialType==smile::CDensity::PT_UNKNOWN)
    cfg.PotentialType=smile::CDensity::PT_SPLINE;
  size_t Nbody=0;
  float * pos=NULL, * vel=NULL, * mass=NULL;
  float * time=NULL, currenttime=0;
  int   * nbody=NULL;
  std::vector< smile::CPosVelDataFloat > trj;
  std::vector< float > masses;
  int numTimesteps=0;
  const smile::CPotential* potential=NULL;
  try {
    while (io_nemo((char*)infile.c_str(),const_cast<char*>("float,n,t,x,v,m,read"),&nbody,&time,&pos,&vel,&mass)) {
      currenttime=*time;
      if(numTimesteps>0 && static_cast<size_t>(*nbody)!=Nbody) {
        std::cerr<<"Error, number of particles is different between snapshots at time="<<currenttime<<std::endl;
        return 0;
      }
      if(pos==NULL || vel==NULL) {
        std::cerr<<"Error, no position/velocity data in snapshot"<<std::endl;
        return 0;
      }
      Nbody = *nbody;
      if(trj.size()==0) {
        trj.resize(Nbody);
        masses.assign(Nbody,0);
        if(mass!=NULL) 
          for(size_t o=0; o<Nbody; o++)
            masses[o]=mass[o];
      }
      for(size_t o=0; o<Nbody; o++)
        trj[o].push_back(smile::CPosVelPoint<float>(pos[o*3],pos[o*3+1],pos[o*3+2],vel[o*3],vel[o*3+1],vel[o*3+2]));
      if(numTimesteps==0) { // first snapshot -- try to init potential
        std::cerr << "Creating potential"<<std::endl;
        potential=smile::createPotential(&cfg);
        if(potential==NULL) {
          smile::CPointMassSet<float> points;
          for(size_t o=0; o<Nbody; o++)
            points.add(smile::CPosVelPoint<float>(pos[o*3],pos[o*3+1],pos[o*3+2],vel[o*3],vel[o*3+1],vel[o*3+2]), mass[o]);
          potential=smile::createPotentialFromPoints(&cfg, &points);
        }
        if(potential==NULL) {
          std::cerr << "Invalid parameters for creating potential"<<std::endl;
          return 0;
        }
        if(cfg.Mbh!=0) {  // need to create a composite potential with black hole
          std::vector<const smile::CPotential*> components(1, potential);
          potential=new smile::CPotentialComposite(components, cfg.Mbh);
          delete components[0];
        }
        std::cerr << "Loading orbits"<<std::endl;
      }
      numTimesteps++;
    }
  }
  catch(const std::bad_alloc&)
  {
    std::cerr << "Error, not enough memory at time=" << currenttime << std::endl;
    return 0;
  }
  if(numTimesteps<=1) {
    std::cerr << "Error, only a single snapshot time is not enough"<<std::endl;
    return 0;
  }
  std::cerr << "Performing orbit analysis"<<std::endl;
  double timeStep = currenttime / (numTimesteps-1);
  try{
    smile::CConfigOrbit configOrbit=smile::defaultConfigOrbit;   // several parameters used for orbit integration [unused] and classification
    configOrbit.freqStrictOrder=false; // actually the only parameter needed to be initialized
    smile::COrbitLibrary* orbits=new smile::COrbitLibrary();
    for(size_t o=0; o<Nbody; o++)
    {
      smile::COrbit orb(&configOrbit, potential, trj[o], timeStep);
      smile::vectorInformation vi;
      for(size_t ni=0; ni<orb.getInfoNum(); ni++)
        vi.push_back(orb.getNewInfo(ni));
      orbits->addOrbitDesc(new smile::COrbitDesc(
        trj[o][0], currenttime, /*maxtime*/0.0f, /*weight*/masses[o], masses[o], &vi));
      if(o%1000==0) { std::cerr << (o*100/Nbody) << "% done\r"; }
    }
    std::cerr << "            \r";
    smile::CFileOutputText out(outfile);
    out.writeOrbitLibrary(orbits, false);
    delete orbits;
    if(!out.ok()) std::cerr << "Error writing orbit library to "<<outfile<< std::endl;
  }
  catch(const std::bad_alloc&)
  {
    std::cerr << "Error, not enough memory" << std::endl;
    return 0;
  }
}
